Package: SpectriPy
Title: Enhancing Cross-Language Mass Spectrometry Data Analysis with R
        and Python
Version: 1.0.0
Description: The SpectriPy package allows integration of Python-based MS
    analysis code with the Spectra package. Spectra objects can be converted
    into Python MS data structures. In addition, SpectriPy integrates
    and wraps the similarity scoring and processing/filtering functions
    from the Python matchms package into R.
Authors@R: c(person(given = "Michael", family = "Witting",
                    role = c("aut"),
                    email = "michael.witting@helmholtz-muenchen.de",
                    comment = c(ORCID = "0000-0002-1462-4426")),
             person(given = "Johannes", family = "Rainer",
                    email = "Johannes.Rainer@eurac.edu",
                    role = c("aut", "cre"),
                    comment = c(ORCID = "0000-0002-6977-7147")),
             person(given = "Carolin", family = "Huber",
                    email = "Carolin-elisabeth.huber@ufz.de",
                    role = c("aut"),
                    comment = c(ORCID = "0000-0002-9355-8948")),
             person(given = "Helge", family = "Hecht",
                    email = "helge.hecht@recetox.muni.cz",
                    role = c("ctb"),
                    comment = c(ORCID = "0000-0001-6744-996X")),
             person(given = "Marilyn", family = "De Graeve",
                    email = "marilyn.degraeve@eurac.edu",
                    role = c("aut"),
                    comment = c(ORCID = "0000-0001-6916-401X")),
             person(given = "Wout", family = "Bittremieux",
                    email = "wout.bittremieux@uantwerpen.be",
                    role = c("aut"),
                    comment = c(ORCID = "0000-0002-3105-1359")),
             person(given = "Thomas", family = "Naake",
                    email = "thomasnaake@googlemail.com",
                    role = c("aut"),
                    comment = c(ORCID = "0000-0001-7917-5580")),
             person(given = "Victor", family = "Chrone",
                    email = "vchrone@bmb.sdu.dk",
                    role = c("ctb"),
                    comment = c(ORCID = "0009-0007-2121-4066")),
             person(given = "Matthias", family = "Anagho-Mattanovich",
                    email = "matthias.mattanovich@sund.ku.dk",
                    role = c("ctb"),
                    comment = c(ORCID = "0000-0001-7561-7898")),
             person(given = "Pierre", family = "Marchal",
                    email = "pierre.marchal@unibe.ch",
                    role = c("ctb"),
                    comment = c(ORCID = "0009-0006-6567-6257")),
             person(given = "Philippine", family = "Louail",
                    email = "philippine.louail@eurac.edu",
                    role = c("ctb"),
                    comment = c(ORCID = "0009-0007-5429-6846"))
             )
Depends: R (>= 4.4.0), reticulate (>= 1.42.0)
Imports: Spectra (>= 1.19.9), IRanges, S4Vectors, MsCoreUtils,
        ProtGenerics, methods
Suggests: testthat, quarto, MsBackendMgf, MsDataHub, mzR, knitr,
        BiocStyle
License: Artistic-2.0
BugReports: https://github.com/RforMassSpectrometry/SpectriPy/issues
URL: https://github.com/RforMassSpectrometry/SpectriPy
biocViews: Infrastructure, Metabolomics, MassSpectrometry, Proteomics
Encoding: UTF-8
SystemRequirements: python (>= 3.12), pandoc, quarto
VignetteBuilder: quarto
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
Collate: 'conversion.R' 'MsBackendPython.R' 'compareSpectriPy.R'
        'filterSpectriPy.R' 'zzz.R'
git_url: https://git.bioconductor.org/packages/SpectriPy
git_branch: RELEASE_3_22
git_last_commit: 017658a
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 06:50:09 UTC; biocbuild
Author: Michael Witting [aut] (ORCID: <https://orcid.org/0000-0002-1462-4426>),
  Johannes Rainer [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6977-7147>),
  Carolin Huber [aut] (ORCID: <https://orcid.org/0000-0002-9355-8948>),
  Helge Hecht [ctb] (ORCID: <https://orcid.org/0000-0001-6744-996X>),
  Marilyn De Graeve [aut] (ORCID:
    <https://orcid.org/0000-0001-6916-401X>),
  Wout Bittremieux [aut] (ORCID: <https://orcid.org/0000-0002-3105-1359>),
  Thomas Naake [aut] (ORCID: <https://orcid.org/0000-0001-7917-5580>),
  Victor Chrone [ctb] (ORCID: <https://orcid.org/0009-0007-2121-4066>),
  Matthias Anagho-Mattanovich [ctb] (ORCID:
    <https://orcid.org/0000-0001-7561-7898>),
  Pierre Marchal [ctb] (ORCID: <https://orcid.org/0009-0006-6567-6257>),
  Philippine Louail [ctb] (ORCID:
    <https://orcid.org/0009-0007-5429-6846>)
Maintainer: Johannes Rainer <Johannes.Rainer@eurac.edu>

# Generated by roxygen2: do not edit by hand

S3method(r_to_py,Spectra)
export(CosineGreedy)
export(CosineHungarian)
export(ModifiedCosine)
export(MsBackendPy)
export(NeutralLossesCosine)
export(compareSpectriPy)
export(defaultSpectraVariableMapping)
export(filterSpectriPy)
export(normalize_intensities)
export(pyspec_to_rspec)
export(reindex)
export(remove_peaks_around_precursor_mz)
export(rspec_to_pyspec)
export(select_by_intensity)
export(select_by_mz)
export(setSpectraVariableMapping)
exportMethods("spectraData<-")
exportMethods("spectraVariableMapping<-")
exportMethods(compareSpectriPy)
exportMethods(filterSpectriPy)
exportMethods(spectraVariableMapping)
importFrom(IRanges,NumericList)
importFrom(MsCoreUtils,rbindFill)
importFrom(MsCoreUtils,vapply1l)
importFrom(S4Vectors,DataFrame)
importFrom(Spectra,MsBackendMemory)
importFrom(Spectra,coreSpectraVariables)
importFrom(Spectra,fillCoreSpectraVariables)
importFrom(Spectra,spectrapply)
importFrom(methods,"slot<-")
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(reticulate,import)
importFrom(reticulate,iterate)
importFrom(reticulate,np_array)
importFrom(reticulate,py)
importFrom(reticulate,py_available)
importFrom(reticulate,py_config)
importFrom(reticulate,py_del_attr)
importFrom(reticulate,py_dict)
importFrom(reticulate,py_eval)
importFrom(reticulate,py_get_attr)
importFrom(reticulate,py_has_attr)
importFrom(reticulate,py_require)
importFrom(reticulate,py_run_string)
importFrom(reticulate,py_set_attr)
importFrom(reticulate,py_to_r)
importFrom(reticulate,r_to_py)
importMethodsFrom(ProtGenerics,"intensity<-")
importMethodsFrom(ProtGenerics,"mz<-")
importMethodsFrom(ProtGenerics,"peaksData<-")
importMethodsFrom(ProtGenerics,"spectraData<-")
importMethodsFrom(ProtGenerics,acquisitionNum)
importMethodsFrom(ProtGenerics,backendInitialize)
importMethodsFrom(ProtGenerics,centroided)
importMethodsFrom(ProtGenerics,collisionEnergy)
importMethodsFrom(ProtGenerics,dataOrigin)
importMethodsFrom(ProtGenerics,extractByIndex)
importMethodsFrom(ProtGenerics,intensity)
importMethodsFrom(ProtGenerics,isolationWindowLowerMz)
importMethodsFrom(ProtGenerics,isolationWindowTargetMz)
importMethodsFrom(ProtGenerics,isolationWindowUpperMz)
importMethodsFrom(ProtGenerics,msLevel)
importMethodsFrom(ProtGenerics,mz)
importMethodsFrom(ProtGenerics,peaksData)
importMethodsFrom(ProtGenerics,polarity)
importMethodsFrom(ProtGenerics,precScanNum)
importMethodsFrom(ProtGenerics,precursorCharge)
importMethodsFrom(ProtGenerics,precursorIntensity)
importMethodsFrom(ProtGenerics,precursorMz)
importMethodsFrom(ProtGenerics,rtime)
importMethodsFrom(ProtGenerics,scanIndex)
importMethodsFrom(ProtGenerics,smoothed)
importMethodsFrom(ProtGenerics,spectraData)
importMethodsFrom(ProtGenerics,spectraNames)
importMethodsFrom(ProtGenerics,spectraVariables)
importMethodsFrom(ProtGenerics,tic)
importMethodsFrom(S4Vectors,isEmpty)
importMethodsFrom(Spectra,"spectraVariableMapping<-")
importMethodsFrom(Spectra,Spectra)
importMethodsFrom(Spectra,intensity)
importMethodsFrom(Spectra,mz)
importMethodsFrom(Spectra,peaksData)
importMethodsFrom(Spectra,peaksVariables)
importMethodsFrom(Spectra,spectraData)
importMethodsFrom(Spectra,spectraVariableMapping)
importMethodsFrom(Spectra,spectrapply)
importMethodsFrom(methods,show)

#' @title Filter Spectra using Python's matchms library
#'
#' @name filterSpectriPy
#'
#' @description
#'
#' The `filterSpectriPy()` function allows to filter/process a `Spectra` object
#' using the `select_by_intensity()`, `select_by_mz()`,
#' `remove_peaks_around_precursor_mz()`, and `normalize_intensities()`
#' functions of the Python
#' [matchms.filtering](https://matchms.readthedocs.io/en/latest/api/matchms.filtering.html)
#' module.
#'
#' Selection and configuration of the algorithm can be performed with one of
#' the parameter objects (equivalent to *matchms*' function names):
#'
#' - `select_by_intensity()`: Keeps only the peaks within defined intensity
#'   range (keep if `intensity_from` >= intensity >= `intensity_to`). See also
#'   the respective [documentation in *matchms*](https://matchms.readthedocs.io/en/latest/api/matchms.filtering.peak_processing.select_by_intensity.html).
#'
#' - `select_by_mz()`: Keeps only the peaks between `mz_from` and `mz_to`
#'   (keep if `mz_from` >= m/z >= `mz_to`). See also the respective
#'   [documentation in *matchms*](https://matchms.readthedocs.io/en/latest/api/matchms.filtering.peak_processing.select_by_mz.html).
#'
#' - `remove_peaks_around_precursor_mz()`: Removes the peaks that are within
#'   `mz_tolerance` (in Da) of the precursor mz, excluding the precursor peak.
#'
#' - `normalize_intensities()`: Normalizes the intensities of peaks
#'   (and losses) to unit height.
#'
#' @note
#'
#' The first call to the `filterSpectriPy()` function can take longer because
#' the Python environment needs to be first set up.
#'
#' `filterSpectriPy()` first translates the `Spectra` to Python, applies the
#' filter functions from the *matchms* Python libraries and then translates
#' the filtered data back to a `Spectra` object. Thus, any spectra variables
#' other than those that are translated between R and Python will be lost
#' during the processing. Use [setSpectraVariableMapping()] to define which
#' spectra variables should be transferred/converted between R and Python.
#' See also examples below for more information.
#'
#' The [Spectra::Spectra()] object returned by `filterSpectriPy()` will
#' **always** use an in-memory backend (i.e. the [Spectra::MsBackendMemory()]),
#' independently of the backend used by the backend used by the input
#' `Spectra`.
#'
#' @param object A [Spectra::Spectra()] object.
#'
#' @param param one of parameter classes listed above (such as
#'     `select_by_intensity()`) defining the filter/processing function in
#'     Python and its parameters.
#'
#' @param mapping named `character()` defining which spectra variables/metadata
#'     should be converted between R and Python and how they should be renamed.
#'     Defaults to `spectraVariableMapping()`. See [setSpectraVariableMapping()]
#'     for more information.
#'
#' @param intensity_from `numeric(1)`: Set lower threshold for peak intensity.
#'     Default is 10.
#'
#' @param intensity_to `numeric(1)`: Set upper threshold for peak intensity.
#'     Default is 200.
#'
#' @param mz_from `numeric(1)`: Set lower threshold for m/z peak positions.
#'     Default is 0.
#'
#' @param mz_to `numeric(1)`: Set upper threshold for m/z peak positions.
#'     Default is 1000.
#'
#' @param mz_tolerance `numeric(1)`: Tolerance of m/z values that are not
#'     allowed to lie within the precursor mz. Default is 17 Da.
#'
#' @param ... ignored.
#'
#' @return `filterSpectriPy()` returns a `Spectra` object on which the
#'     filtering/processing function has been applied
#'
#' @author Thomas Naake
#'
#' @seealso
#'
#' - [Spectra::filterIntensity()], [Spectra::filterMzRange()],
#'   [Spectra::scalePeaks()] in the `Spectra` package for pure R
#'   implementations of filtering/processing calculations.
#'
#' - [rspec_to_pyspec()] or [pyspec_to_rspec()] for the functions used to
#'   translated the MS data between R and Python.
#'
#' @export
#'
#' @importFrom reticulate py_run_string
#'
#' @examples
#'
#' library(Spectra)
#'
#' ## create some example Spectra
#' DF <- DataFrame(
#'     msLevel = c(2L, 2L, 2L),
#'     name = c("Caffeine", "Caffeine", "1-Methylhistidine"),
#'     precursorMz = c(195.0877, 195.0877, 170.0924)
#' )
#' DF$intensity <- list(
#'     c(340.0, 416, 2580, 412),
#'     c(388.0, 3270, 85, 54, 10111),
#'     c(3.407, 47.494, 3.094, 100.0, 13.240))
#' DF$mz <- list(
#'     c(135.0432, 138.0632, 163.0375, 195.0880),
#'     c(110.0710, 138.0655, 138.1057, 138.1742, 195.0864),
#'     c(109.2, 124.2, 124.5, 170.16, 170.52))
#' sps <- Spectra(DF)
#'
#' ## Filter: select_by_intensity
#' res <- filterSpectriPy(
#'     sps, select_by_intensity(intensity_from = 15, intensity_to = 300))
#' ## Only mass peaks with intensities between the specified limits are
#' ## retained
#' intensity(res)
#' ## Compared to the original intensities
#' intensity(sps)
#'
#' ## Note that the spectra variable `"name"` was lost during conversion of
#' ## the MS data between R and Python:
#' sps$name
#' any(spectraVariables(res) == "name")
#'
#' ## Only spectra variables defined by `spectraVariableMapping()` are
#' ## converted and thus retained:
#' spectraVariableMapping()
#'
#' ## We can also pass a custom *spectra variable mapping* with the `mapping`
#' ## parameter to the `filterSpectriPy()` function. Below we create such
#' ## a mapping by adding the translation of a spectra variable `"name"` to
#' ## a metadata name `"compound_name"` to the default spectra variable
#' ## mapping `defaultSpectraVariableMapping()`.
#' map <- c(defaultSpectraVariableMapping(), name = "compound_name")
#' map
#'
#' ## Repeat the filtering operation passing this mapping information:
#' res <- filterSpectriPy(
#'     sps, select_by_intensity(intensity_from = 15, intensity_to = 300),
#'     mapping = map)
#' res$name
#'
NULL

setGeneric("filterSpectriPy", function(object, param, ...)
    standardGeneric("filterSpectriPy"))

setClass("filter_param")
setClass("select_by_intensity",
    contains = "filter_param",
    slots = c(
        intensity_from = "numeric", intensity_to = "numeric"),
    prototype = prototype(
             intensity_from = 20, intensity_to = 200),
    validity = function(object) {
        msg <- NULL
        if (length(object@intensity_from) != 1 || object@intensity_from < 0)
            msg <- c("'intensity_from' has to be a positive number of length 1")
        if (length(object@intensity_to) != 1 || object@intensity_to < 0)
            msg <- c("'intensity_to' has to be a positive number of length 1")
        msg
    }
)
setClass("select_by_mz",
    contains = "filter_param",
    slots = c(
        mz_from = "numeric",
        mz_to = "numeric"),
    prototype = prototype(
        mz_from = 150,
        mz_to = 450),
    validity = function(object) {
        msg <- NULL
        if (length(object@mz_from) != 1 || object@mz_from < 0)
            msg <- c("'mz_from' has to be a positive number of length 1")
        if (length(object@mz_to) != 1 || object@mz_to < 0)
            msg <- c("'mz_to' has to be a positive number of length 1")
        msg
    }
)
setClass("remove_peaks_around_precursor_mz",
    contains = "filter_param",
    slots = c(
        mz_tolerance = "numeric"),
    prototype = prototype(
            mz_tolerance = 20),
    validity = function(object) {
        msg <- NULL
        if (length(object@mz_tolerance) != 1 || object@mz_tolerance < 0)
            msg <- c("'mz_tolerance' has to be a positive number of length 1")
        msg
    }
)
setClass("normalize_intensities",
    contains = "filter_param",
    prototype = prototype(),
    validity = function(object) {
        msg <- NULL
        msg
    }
)

#' @rdname filterSpectriPy
#'
#' @importFrom methods new
#'
#' @export
select_by_intensity <- function(intensity_from = 10, intensity_to = 200) {
    new("select_by_intensity", intensity_from = as.numeric(intensity_from),
        intensity_to = as.numeric(intensity_to))
}

#' @rdname filterSpectriPy
#'
#' @export
select_by_mz <- function(mz_from = 0, mz_to = 1000) {
    new("select_by_mz", mz_from = as.numeric(mz_from),
        mz_to = as.numeric(mz_to))
}

#' @rdname filterSpectriPy
#'
#' @export
remove_peaks_around_precursor_mz <- function(mz_tolerance = 17) {
    new("remove_peaks_around_precursor_mz",
        mz_tolerance = as.numeric(mz_tolerance))
}

#' @rdname filterSpectriPy
#'
#' @export
normalize_intensities <- function() {
    new("normalize_intensities")
}

#' @rdname filterSpectriPy
#'
#' @exportMethod filterSpectriPy
setMethod(
    "filterSpectriPy",
    signature = c(object = "Spectra", param = "filter_param"),
    function(object, param, mapping = spectraVariableMapping(), ...) {
        .filter_spectra_python(object, param, mapping = mapping)
    })

#' Helper method to return a SpectrumProcessor that can be applied to the
#' list of spectra in Python
#'
#' @importFrom reticulate py_dict
#'
#' @noRd
setMethod("py_fun", "select_by_intensity", function(object) {
    matchms_filtering$SpectrumProcessor$create_partial_function(
        matchms_filtering$select_by_intensity,
        py_dict(c("intensity_from", "intensity_to"),
                c(object@intensity_from, object@intensity_to)))
})

setMethod("py_fun", "select_by_mz", function(object) {
    matchms_filtering$SpectrumProcessor$create_partial_function(
        matchms_filtering$select_by_mz,
        py_dict(c("mz_from", "mz_to"),
                c(object@mz_from, object@mz_to)))
})

setMethod("py_fun", "remove_peaks_around_precursor_mz", function(object) {
    matchms_filtering$SpectrumProcessor$create_partial_function(
        matchms_filtering$remove_peaks_around_precursor_mz,
        py_dict(c("mz_tolerance"),
                c(object@mz_tolerance)))
})

setMethod("py_fun", "normalize_intensities", function(object) {
    matchms_filtering$SpectrumProcessor$create_partial_function(
        matchms_filtering$normalize_intensities)
})

#' internal function to filter/processing with python's matchms. `Spectra`
#' will be converted to python `Spectrum` class and matchms' processing
#' functions will be applied on the `Spectrum` objects. After processing, the
#' matchms' `Spectrum` objects will be converted back to `Spectra` objects.
#'
#' @param sps `Spectra` object
#'
#' @param param Parameter object.
#'
#' @return a `Spectra` object
#'
#' @importFrom reticulate py
#'
#' @noRd
#'
#' @author Thomas Naake, Johannes Rainer
.filter_spectra_python <- function(x, param,
                                   mapping = spectraVariableMapping()) {
    ## handle empty input
    if (!length(x))
        return(Spectra())

    proc <- matchms_filtering$SpectrumProcessor$SpectrumProcessor(
        r_to_py(list(py_fun(param))))
    pyspec_to_rspec(
        proc$process_spectra(rspec_to_pyspec(x, mapping = mapping))[0],
        mapping = mapping)
}

caf <- DataFrame(
    msLevel = c(2L, 2L),
    name = "Caffeine",
    precursorMz = c(195.0877, 195.0877)
)
caf$intensity <- list(
    c(340.0, 416, 2580, 412),
    c(388.0, 3270, 85, 54, 10111))
caf$mz <- list(
    c(135.0432, 138.0632, 163.0375, 195.0880),
    c(110.0710, 138.0655, 138.1057, 138.1742, 195.0864))
caf <- Spectra(caf)

mhd <- DataFrame(
    msLevel = c(2L, 2L),
    precursorMz = c(170.0924, 170.0924),
    id = c("HMDB0000001", "HMDB0000001"),
    name = c("1-Methylhistidine", "1-Methylhistidine"))
mhd$mz <- list(
    c(109.2, 124.2, 124.5, 170.16, 170.52),
    c(83.1, 96.12, 97.14, 109.14, 124.08, 125.1, 170.16))
mhd$intensity <- list(
    c(3.407, 47.494, 3.094, 100.0, 13.240),
    c(6.685, 4.381, 3.022, 16.708, 100.0, 4.565, 40.643))
mhd <- Spectra(mhd)
s_all <- c(caf, mhd)

## select_by_intensity
test_that("select_by_intensity constructor works", {
    ## errors
    expect_error(select_by_intensity(intensity_from = c(1.2, 1.5)), "length 1")
    expect_error(select_by_intensity(intensity_from = -1.2), "length 1")
    expect_error(select_by_intensity(intensity_to = c(1, 2)), "length 1")
    expect_error(select_by_intensity(intensity_to = c(1, 2)), "length 1")
    expect_error(select_by_intensity(foo = 1), "unused argument")

    res <- select_by_intensity(intensity_from = 5, intensity_to = 20)
    expect_s4_class(res, "select_by_intensity")
    expect_equal(res@intensity_from, 5)
    expect_equal(res@intensity_to, 20)
})

test_that("py_fun,select_by_intensity works", {
    p <- select_by_intensity(intensity_from = 3, intensity_to = 24)
    res <- py_fun(p)
    expect_true(is(res, "functools.partial"))
    expect_equal(py_to_r(res$keywords["intensity_from"]), 3.0)
    expect_equal(py_to_r(res$keywords["intensity_to"]), 24.0)
})

test_that("filterSpectriPy,Spectra,select_by_intensity works", {
    p <- select_by_intensity(intensity_from = 17, intensity_to = 300)
    res <- filterSpectriPy(s_all, p)
    expect_s4_class(res, "Spectra")
    expect_equal(intensity(res)[[1L]], numeric())
    expect_true(all(intensity(res)[[2L]] > 17 & intensity(res)[[2L]] < 300))
    expect_true(all(intensity(res)[[3L]] > 17 & intensity(res)[[3L]] < 300))
    expect_true(all(intensity(res)[[4L]] > 17 & intensity(res)[[4L]] < 300))
    expect_equal(msLevel(s_all), msLevel(res))
    expect_equal(rtime(s_all), rtime(res))
})

## select_by_mz
test_that("select_by_mz constructor works", {
    ## errors
    expect_error(select_by_mz(mz_from = c(1.2, 1.5)), "length 1")
    expect_error(select_by_mz(mz_from = -1.2), "length 1")
    expect_error(select_by_mz(mz_to = c(1, 2)), "length 1")
    expect_error(select_by_mz(mz_to = c(1, 2)), "length 1")
    expect_error(select_by_mz(foo = 1), "unused argument")

    res <- select_by_mz(mz_from = 0, mz_to = 400)
    expect_s4_class(res, "select_by_mz")
    expect_equal(res@mz_from, 0)
    expect_equal(res@mz_to, 400)
})

test_that("py_fun,select_by_mz works", {
    p <- select_by_mz(mz_from = 3, mz_to = 24)
    res <- py_fun(p)
    expect_true(is(res, "functools.partial"))
    expect_equal(py_to_r(res$keywords["mz_from"]), 3.0)
    expect_equal(py_to_r(res$keywords["mz_to"]), 24.0)
})

test_that("filterSpectriPy,Spectra,select_by_mz works", {
    p <- select_by_mz(mz_from = 100, mz_to = 180)
    res <- filterSpectriPy(s_all, p)
    expect_s4_class(res, "Spectra")
    expect_true(all(mz(res)[[1L]] > 100 & mz(res)[[1L]] < 180))
    expect_true(all(mz(res)[[2L]] > 100 & mz(res)[[2L]] < 180))
    expect_true(all(mz(res)[[3L]] > 100 & mz(res)[[3L]] < 180))
    expect_true(all(mz(res)[[4L]] > 100 & mz(res)[[4L]] < 180))
    expect_equal(msLevel(s_all), msLevel(res))
    expect_equal(rtime(s_all), rtime(res))
})

## remove_peaks_around_precursor_mz
test_that("remove_peaks_around_precursor_mz constructor works", {
    ## errors
    expect_error(remove_peaks_around_precursor_mz(
        mz_tolerance = c(1.2, 1.5)), "length 1")
    expect_error(remove_peaks_around_precursor_mz(
        mz_tolerance = -1.2), "length 1")
    expect_error(remove_peaks_around_precursor_mz(foo = 1), "unused argument")

    res <- remove_peaks_around_precursor_mz(mz_tolerance = 17)
    expect_s4_class(res, "remove_peaks_around_precursor_mz")
    expect_equal(res@mz_tolerance, 17)
})

test_that("py_fun,remove_peaks_around_precursor_mz works", {
    p <- remove_peaks_around_precursor_mz(mz_tolerance = 200)
    res <- py_fun(p)
    expect_true(is(res, "functools.partial"))
    expect_equal(py_to_r(res$keywords["mz_tolerance"]), 200)
})

test_that("filterSpectriPy,Spectra,remove_peaks_around_precursor_mz works", {
    p <- remove_peaks_around_precursor_mz(mz_tolerance = 200)
    res <- filterSpectriPy(s_all, p)
    expect_s4_class(res, "Spectra")
    expect_true(all(lengths(res) == 0))
    expect_equal(msLevel(s_all), msLevel(res))
    expect_equal(rtime(s_all), rtime(res))

    p <- remove_peaks_around_precursor_mz(mz_tolerance = 10)
    res <- filterSpectriPy(s_all, p)
    expect_true(all(lengths(res) < lengths(s_all)))
})

## normalize_intensities
test_that("normalize_intensities constructor works", {
    ## errors
    expect_error(normalize_intensities(foo = 1), "unused argument")

    res <- normalize_intensities()
    expect_s4_class(res, "normalize_intensities")
})

## test_that("filterSpectriPy,normalize_intensitites works", {
##     p <- normalize_intensities()
##     res <- filterSpectriPy(s_all, p)
## })

test_that("py_fun,select_by_intensity works", {
    p <- select_by_mz(mz_from = 3, mz_to = 24)
    res <- py_fun(p)
    expect_true(is(res, "functools.partial"))
    expect_equal(py_to_r(res$keywords["mz_from"]), 3.0)
    expect_equal(py_to_r(res$keywords["mz_to"]), 24.0)
})

test_that("filterSpectriPy,Spectra,select_by_mz works", {
    p <- select_by_mz(mz_from = 100, mz_to = 180)
    res <- filterSpectriPy(s_all, p)
    expect_s4_class(res, "Spectra")
    expect_true(all(mz(res)[[1L]] > 100 & mz(res)[[1L]] < 180))
    expect_true(all(mz(res)[[2L]] > 100 & mz(res)[[2L]] < 180))
    expect_true(all(mz(res)[[3L]] > 100 & mz(res)[[3L]] < 180))
    expect_true(all(mz(res)[[4L]] > 100 & mz(res)[[4L]] < 180))
    expect_equal(msLevel(s_all), msLevel(res))
    expect_equal(rtime(s_all), rtime(res))
})

test_that(".filter_spectra_python works", {
    res <- .filter_spectra_python(Spectra())
    expect_s4_class(res, "Spectra")
    expect_true(length(res) == 0L)
})

## based on https://rstudio.github.io/reticulate/articles/package.html#delay-loading-python-modules
matchms <- NULL
matchms_similarity <- NULL
matchms_filtering <- NULL
spectrum_utils <- NULL

.PY_PKGS <- c(matchms = "matchms>=0.31",
              spectrum_utils = "spectrum_utils>=0.3.2",
              numpy = "numpy>=2.2.0")

#' @importFrom reticulate py_require py_available
.onLoad <- function(libname, pkgname) {
    .check_environment()
    py_require(packages = .PY_PKGS, python_version = ">=3.12")
    .initialize_libraries2(FALSE, FALSE, asNamespace(pkgname))
}

.check_environment <- function() {
    ## Check for old setup
    if (.is_spectripy_use_system())
        warning("Ignoring environment variable 'SPECTRIPY_USE_SYSTEM'. ",
                "Please use 'RETICULATE_PYTHON' or 'RETICULATE_PYTHON_ENV'",
                " instead.")
    if (.is_spectripy_use_conda())
        warning("Environment variable 'SPECTRIPY_USE_CONDA' is no longer ",
                "supported. Please see the package vignette for updated ",
                "Python setup options.")
    se <- .spectripy_env()
    if (length(se) && se != "")
        warning("Environment variable 'SPECTRIPY_ENV' is no longer ",
                "supported. Please use 'RETICULATE_PYTHON_ENV' instead. See ",
                "the package vignette for updated Python setup options.")
    ## Potentially interfering global settings
    if ((res <- Sys.getenv("RETICULATE_PYTHON")) != "")
        packageStartupMessage("Using Python defined by 'RETICULATE_PYTHON': ",
                              res)
    if ((res <- Sys.getenv("RETICULATE_PYTHON_ENV")) != "")
        packageStartupMessage("Using Python environment defined by ",
                              "'RETICULATE_PYTHON_ENV': ", res)
}

.is_spectripy_use_system <- function() {
    res <- getOption("spectripy.use_system", default = NULL)
    if (!length(res))
        res <- Sys.getenv("SPECTRIPY_USE_SYSTEM")
    grepl("1|yes|true", res, ignore.case = TRUE)
}

.is_spectripy_use_conda <- function() {
    res <- getOption("spectripy.use_conda", default = NULL)
    if (!length(res))
        res <- Sys.getenv("SPECTRIPY_USE_CONDA")
    grepl("1|yes|true", res, ignore.case = TRUE)
}

.spectripy_env <- function() {
    if(length(res <- getOption("spectripy.env", default = NULL)))
        return(res)
    Sys.getenv("SPECTRIPY_ENV")
}

#' Load all required Python libraries and assign it to package-internal
#' variables
#'
#' @importFrom reticulate import py_config
#'
#' @noRd
.initialize_libraries2 <- function(delay_load = TRUE, convert = FALSE,
                                   envir = new.env()) {
    if (!reticulate::py_available(initialize = TRUE))
        stop("Unable to initialize the Python environment", call. = FALSE)
    tryCatch({
        assign("matchms", import("matchms", delay_load = delay_load,
                                 convert = convert), envir = envir)
        assign("matchms_similarity",
               import("matchms.similarity", delay_load = delay_load,
                      convert = convert), envir = envir)
        assign("matchms_filtering",
               import("matchms.filtering", delay_load = delay_load,
                      convert = convert), envir = envir)
        assign("spectrum_utils",
               import("spectrum_utils", delay_load = delay_load,
                      convert = convert), envir = envir)
    }, error = function(e) {
        stop("Failed to initialize Python environment and libraries!\n",
             "Original message: ", e, "\nPython configuration:\n",
             print(py_config()), "\nEnvironmental variables:\n",
             print(Sys.getenv()))
    })
}

test_that(".initialize_libraries2 works", {
    a <- new.env()
    import_mock <- function(module, as = NULL, convert = TRUE,
                            delay_load = FALSE) module
    with_mocked_bindings(
        "import" = import_mock,
        .package = "reticulate",
        code = .initialize_libraries2(envir = a)
    )
    expect_equal(sort(ls(a)), sort(c("matchms", "matchms_similarity",
                                     "matchms_filtering", "spectrum_utils")))
    orig <- getOption("spectripy.use.system")
    options(spectripy.use.system = TRUE)
})

test_that(".onLoad works", {
    ## Does not work because the `asNamespace()` call.
})

test_that(".is_spectripy_use_system works", {
    orig_opt <- getOption("spectripy.use_system")
    orig_env <- Sys.getenv("SPECTRIPY_USE_SYSTEM")

    options(spectripy.use.system = NULL)
    Sys.unsetenv("SPECTRIPY_USE_SYSTEM")
    expect_false(.is_spectripy_use_system())
    options(spectripy.use_system = "FALSE")
    expect_false(.is_spectripy_use_system())
    options(spectripy.use_system = "True")
    expect_true(.is_spectripy_use_system())
    Sys.setenv(SPECTRIPY_USE_SYSTEM="FALSE")
    expect_true(.is_spectripy_use_system())
    options(spectripy.use_system = NULL)
    expect_false(.is_spectripy_use_system())
    Sys.setenv(SPECTRIPY_USE_SYSTEM="1")
    expect_true(.is_spectripy_use_system())

    if (orig_env == "")
        Sys.unsetenv("SPECTRIPY_USE_SYSTEM")
    else Sys.setenv(SPECTRIPY_USE_SYSTEM=orig_env)
    options(spectripy.use_system = orig_opt)
})

test_that(".is_spectripy_use_conda works", {
    orig_opt <- getOption("spectripy.use_conda")
    orig_env <- Sys.getenv("SPECTRIPY_USE_CONDA")

    options(spectripy.use.conda = NULL)
    Sys.unsetenv("SPECTRIPY_USE_CONDA")
    expect_false(.is_spectripy_use_conda())
    options(spectripy.use_conda = "FALSE")
    expect_false(.is_spectripy_use_conda())
    options(spectripy.use_conda = "True")
    expect_true(.is_spectripy_use_conda())
    Sys.setenv(SPECTRIPY_USE_CONDA="FALSE")
    expect_true(.is_spectripy_use_conda())
    options(spectripy.use_conda = NULL)
    expect_false(.is_spectripy_use_conda())
    Sys.setenv(SPECTRIPY_USE_CONDA="1")
    expect_true(.is_spectripy_use_conda())

    if (orig_env == "")
        Sys.unsetenv("SPECTRIPY_USE_CONDA")
    else Sys.setenv(SPECTRIPY_USE_CONDA=orig_env)
    options(spectripy.use_conda = orig_opt)
})

test_that(".spectripy_env works", {
    orig_opt <- getOption("spectripy.env")
    orig_env <- Sys.getenv("SPECTRIPY_ENV")

    options(spectripy.env = NULL)
    Sys.unsetenv("SPECTRIPY_ENV")
    expect_equal(.spectripy_env(), "")

    options(spectripy.env = "myenvi")
    expect_equal(.spectripy_env(), "myenvi")
    Sys.setenv(SPECTRIPY_ENV="myotherenvi")
    expect_equal(.spectripy_env(), "myenvi")

    options(spectripy.env = NULL)
    expect_equal(.spectripy_env(), "myotherenvi")

    if (orig_env == "")
        Sys.unsetenv("SPECTRIPY_ENV")
    else Sys.setenv(SPECTRIPY_ENV=orig_env)
    options(spectripy.env = orig_opt)
})

test_that(".check_environment works", {
    sys_opt <- getOption("spectripy.use_system")
    sys_env <- Sys.getenv("SPECTRIPY_USE_SYSTEM")
    con_opt <- getOption("spectripy.use_conda")
    con_env <- Sys.getenv("SPECTRIPY_USE_CONDA")
    env_opt <- getOption("spectripy.env")
    env_env <- Sys.getenv("SPECTRIPY_ENV")
    ret_py <- Sys.getenv("RETICULATE_PYTHON")
    ret_env <- Sys.getenv("RETICULATE_PYTHON_ENV")
    ## cleaning options and environment
    options(spectripy.use_system=NULL)
    options(spectripy.use_conda=NULL)
    options(spectripy.env=NULL)
    Sys.unsetenv("SPECTRIPY_USE_SYSTEM")
    Sys.unsetenv("SPECTRIPY_USE_CONDA")
    Sys.unsetenv("SPECTRIPY_ENV")
    Sys.unsetenv("RETICULATE_PYTHON")
    Sys.unsetenv("RETICULATE_PYTHON_ENV")

    expect_silent(.check_environment())
    options(spectripy.use_system=TRUE)
    expect_warning(.check_environment(), "variable 'SPECTRIPY_USE_SYSTEM'")
    options(spectripy.use_system=NULL)
    options(spectripy.use_conda=1)
    expect_warning(.check_environment(), "variable 'SPECTRIPY_USE_CONDA'")
    options(spectripy.use_conda=NULL)
    options(spectripy.env="something")
    expect_warning(.check_environment(), "variable 'SPECTRIPY_ENV'")
    options(spectripy.env=NULL)
    Sys.setenv(RETICULATE_PYTHON="something")
    expect_message(.check_environment(), "Using Python defined by")
    Sys.unsetenv("RETICULATE_PYTHON")
    Sys.setenv(RETICULATE_PYTHON_ENV="something")
    expect_message(.check_environment(), "Using Python environment defined by")
    Sys.unsetenv("RETICULATE_PYTHON_ENV")

    ## Restoring options and environment
    options(spectripy.use_system = sys_opt)
    options(spectripy.use_conda = con_opt)
    options(spectripy.env = env_opt)
    if (sys_env == "")
        Sys.unsetenv("SPECTRIPY_USE_SYSTEM")
    else Sys.setenv(SPECTRIPY_USE_SYSTEM=sys_env)
    if (con_env == "")
        Sys.unsetenv("SPECTRIPY_USE_CONDA")
    else Sys.setenv(SPECTRIPY_USE_CONDA=con_env)
    if (env_env == "")
        Sys.unsetenv("SPECTRIPY_ENV")
    else Sys.setenv(SPECTRIPY_ENV=env_env)
    if (ret_py == "")
        Sys.unsetenv("RETICULATE_PYTHON")
    else Sys.setenv(RETICULATE_PYTHON=ret_py)
    if (ret_env == "")
        Sys.unsetenv("RETICULATE_PYTHON_ENV")
    else Sys.setenv(RETICULATE_PYTHON_ENV=ret_env)
})

#' @include conversion.R

#' @title A MS data backend for MS data stored in Python
#'
#' @description
#'
#' The `MsBackendPy` allows to access MS data stored as `matchms.Spectrum`
#' or `spectrum_utils.spectrum.MsmsSpectrum` objects from the
#' [*matchms*](https://github.com/matchms/matchms) respectively
#' [*spectrum_utils*](https://github.com/bittremieux-lab/spectrum_utils) Python
#' library directly from R. The MS data (peaks data or spectra variables) are
#' translated on-the-fly when accessed. Thus, the `MsBackendPy` allows a
#' seamless integration of Python MS data structures into [Spectra::Spectra()]
#' based analysis workflows.
#'
#' The `MsBackendPy` object supports replacing values for peaks variables
#' (*m/z* and intensity) and adding/replacing or removing spectra variables.
#' The changes are immediately translated and written back to the Python
#' variable.
#'
#' See the descripion of the `backendInitialize()` method below for creation
#' and initialization of objects from this class. Also, the `setBackend()`
#' method for [Spectra::Spectra()] objects internally uses
#' `backendInitialize()`, thus the same parameters can (and have) to be passed
#' if the backend of a `Spectra` object is changed to `MsBackendPy` using
#' the `setBackend()` method. Special care should also be given to parameter
#' `spectraVariableMapping`, that defines which spectra variables should be
#' considered/translated and how their names should or have to be converted
#' between R and Python. See the description for `backendInitialize()` and the
#' package vignette for details and examples.
#'
#' @details
#'
#' The `MsBackendPy` keeps only a reference to the MS data in Python (i.e. the
#' name of the variable in Python) as well as an index pointing to the
#' individual spectra in Python but no other data. Any data requested from
#' the `MsBackendPy` is accessed and translated on-the-fly from the Python
#' variable. The `MsBackendPy` is thus an interface to the MS data, but not
#' a data container. All changes to the MS data in the Python variable
#' (performed e.g. in Python) immediately affect any `MsBackendPy` instances
#' pointing to this variable.
#'
#' Special care must be taken if the MS data structure in Python is subset or
#' its order is changed (e.g. by another process). In that case it might be
#' needed to re-index the backend using the `reindex()` function:
#' `object <- reindex(object)`. This will update (replace) the index to the
#' individual spectra in Python which is stored within the backend.
#'
#' @note
#'
#' As mentioned in the *details* section the MS data is completely stored in
#' Python and the backend only references to this data through the name of
#' the variable in Python. Thus, each time MS data is requested from the
#' backend, it is retrieved in its **current** state.
#' If for example data was transformed or metadata added or removed in the
#' Python object, it immediately affects the `Spectra`/backend.
#'
#' Any replacement operation uses internally the `spectraData()<-` method,
#' thus replacing/updating values for individual spectra variables or peaks
#' variables will first load the current data from Python to R, update or
#' replace the values and then store the full MS data again to the
#' referenced Python attribute.
#'
#' @section `MsBackendPy` methods:
#'
#' The `MsBackendPy` supports all methods defined by the [Spectra::MsBackend()]
#' interface for access to MS data. Details on the invidual functions can also
#' be found in the main documentation in the *Spectra* package (i.e. for
#' [Spectra::MsBackend()]). Here we provide information for functions with
#' specific properties of the backend.
#'
#' - `backendInitialize()`: this method can be used to either initialize the
#'   backend with data from a referenced and **existing ** MS data structure
#'   in Python, or, through parameter `data`, first convert and store the
#'   provided data to a Python MS data structure and then initialize the
#'   backend pointing to this referenced variable (Python attribute). In both
#'   cases, the name of the Python attribute needs to be provided with the
#'   parameter `pythonVariableName`.
#'   The mapping between the spectra variable names in R and the related
#'   Python metadata variables can be specified with the
#'   `spectraVariableMapping` parameter. It has to be a named `character` with
#'   names being the spectra variables and the values the respective name for
#'   the metadata in the Python MS data structure. It defaults to
#'   [defaultSpectraVariableMapping()] which returns the mapping of some
#'   core spectra variables for the *matchms* Python library. Be aware that
#'   only those spectra variables specified with this parameter are mapped and
#'   translated between R and Python. For `backendInitialize()` with parameter
#'   `data` provided, only the variables defined by `spectraVariableMapping`,
#'   and available in `data`, are converted and stored in Python. Also note
#'   that, for efficiency reasons, core spectra variables (those listed by
#'   [Spectra::coreSpectraVariables()]) defined with `spectraVariableMapping`
#'   but that have only missing values, are ignored.
#'   Parameter `pythonLibrary` must be used to specify the Python library
#'   representing the MS data in Python. It can be either
#'   `pythonLibrary = "matchms"` (the default) or
#'   `pythonLibrary = "spectrum_utils"`. The function returns an initialized
#'   instance of `MsBackendPy`. See examples below for different settings
#'   and conversion of spectra variables.
#'
#' - `intensity()`, `intensity()<-`: get or replace the intensity values.
#'   `intensity()` returns a `NumericList` of length equal to the number of
#'   spectra with each element being the intensity values of the individual
#'   mass peaks per spectrum. `intensity()<-` takes the same list-like
#'   structure as input parameter. Both the number of spectra and the number of
#'   peaks must match the length of the spectra and the number of existing mass
#'   peaks. To change the number of peaks use the `peaksData()<-` method
#'   instead that replaces the *m/z* and intensity values at the same time.
#'   Calling `intensity()<-` will replace the full MS data (spectra variables
#'   as well as peaks variables) of the associated Python variable.
#'
#' - `mz()`, `mz()<-`: get or replace the *m/z* values. `mz()` returns a
#'   `NumericList` of length equal to the number of spectra with each element
#'   being the *m/z* values of the individual mass peaks per spectrum.
#'   `mz()<-` takes the same list-like structure as input parameter. Both the
#'   number of spectra and the number of peaks must match the length of the
#'   spectra and the number of existing mass peaks. To change the number of
#'   peaks use the `peaksData()<-` method instead that replaces the *m/z* and
#'   intensity values at the same time.
#'   Calling `mz()<-` will replace the full MS data (spectra variables
#'   as well as peaks variables) of the associated Python variable.
#'
#' - `peaksData()`: extracts the peaks data matrices from the backend. Python
#'   code is applied to the data structure in Python to
#'   extract the *m/z* and intensity values as a list of (numpy) arrays. These
#'   are then translated into an R `list` of two-column `numeric` matrices.
#'   Because Python does not allow to name columns of an array, an additional
#'   loop in R is required to set the column names to `"mz"` and `"intensity"`.
#'
#' - `peaksData()<-`: replaces the full peaks data (i.e., *m/z* and intensity
#'   values) for all spectra. Parameter `value` has to be a `list`-like
#'   structure with each element being a `numeric` matrix with one column
#'   (named `"mz"`) containing the spectrum's *m/z* and one column (named
#'   `"intensity"`) with the intensity values. This method will replace the
#'   full data of the associated Python variable (i.e., both the spectra as
#'   well as the peaks data).
#'
#' - `spectraData()`: extracts the spectra data from the backend. Which spectra
#'   variables are translated and retrieved from the Python objects depends on
#'   the backend's `spectraVariableMapping()`. All metadata names defined are
#'   retrieved and added to the returned `DataFrame` (with eventually missing
#'   *core* spectra variables filled with `NA`).
#'
#' - `spectraData()<-`: replaces the full spectra (+ peaks) data of the backend
#'   with the values provided with the submitted `DataFrame`. The number of
#'   rows of this `DataFrame` has to match the number of spectra of `object`
#'   (i.e., being equal to `length(object)`) and the `DataFrame` must also
#'   contain the spectras' *m/z* and intensity values.
#'
#' - `spectraVariables()`: retrieves available spectra variables, which include
#'   the names of all metadata attributes in the `matchms.Spectrum` objects
#'   and the *core* spectra variables [Spectra::coreSpectraVariables()].
#'
#' - `spectraVariableMapping()`: get the currently defined mapping for
#'   `spectraVariables()` of the backend.
#'
#' - `spectraVariableMapping<-`: replaces the `spectraVariableMapping` of the
#'   backend (see [setSpectraVariableMapping()] for details and description
#'   of the expected format).
#'
#' - `$`, `$<-`: extract or add/replace values for a spectra variable from/in
#'   the backend. Replacing or adding values for a spectra variable cause the
#'   full data to be replaced. In detail, first the full data is retrieved from
#'   Python, then the values are added/replaced and then the data is again
#'   transferred to Python.
#'
#' @section Additional helper and utility functions:
#'
#' - `reindex()`: update the internal *index* to match `1:length(object)`.
#'   This function is useful if the original data referenced by the backend was
#'   subset or re-ordered by a different process (or a function in Python).
#'
#' @param columns For `spectraData()`: `character` with the names of
#'     columns (spectra variables) to retrieve. Defaults to
#'     `spectraVariables(object)`. For `peaksData()`: `character` with the
#'     names of the peaks variables to retrieve.
#'
#' @param data For `backendInitialize()`: `DataFrame` with the full MS data
#'     (peaks data and spectra data) such as extracted with the
#'     [Spectra::spectraData()] method on another `MsBackend` instance.
#'     Importantly, the `DataFrame` must have columns `"mz"` and
#'     `"intensity"` with the full MS data.
#'
#' @param drop For `spectraData()` and `peaksData()`: `logical(1)` whether,
#'     when a single column is requested, the data should be returned as a
#'     `vector` instead of a `data.frame` or `matrix`.
#'
#' @param name For `$`: `character(1)` with the name of the variable to
#'     retrieve.
#'
#' @param object A `MsBackendPy` object.
#'
#' @param pythonLibrary For `backendInitialize()`: `character(1)` specifying
#'     the Python library used to represent the MS data in Python. Can be
#'     either `pythonLibrary = "matchms"` (the default) or
#'     `pythonLibrary = "spectrum_utils"`.
#'
#' @param pythonVariableName For `backendInitialize()`: `character(1)` with the
#'     name of the variable/Python attribute that contains the list of
#'     `matchms.Spectrum` objects with the MS data.
#'
#' @param spectraVariableMapping For `backendInitialize()`: named `character`
#'     with the mapping between spectra variable names and (`matchms.Spectrum`)
#'     metadata names. See [defaultSpectraVariableMapping()], and the
#'     description of the `backendInitialize()` function for `MsBackendPy`
#'     for more information and details.
#'
#' @param value Replacement value(s).
#'
#' @param x A `MsBackendPy` object
#'
#' @param ... Additional parameters.
#'
#' @author Johannes Rainer and the EuBIC hackathon team
#'
#' @name MsBackendPy
#'
#' @return
#'
#' See description of individual functions for their return values.
#'
#' @examples
#'
#' ## Loading an example MGF file provided by the SpectriPy package.
#' ## As an alternative, the data could also be imported directly in Python
#' ## using:
#' ## import matchms
#' ## from matchms.importing import load_from_mgf
#' ## s_p = list(load_from_mgf(r.fl))
#' library(Spectra)
#' library(SpectriPy)
#' library(MsBackendMgf)
#'
#' fl <- system.file("extdata", "mgf", "test.mgf", package = "SpectriPy")
#' s <- Spectra(fl, source = MsBackendMgf())
#' s
#'
#' ## Translating the MS data to Python and assigning it to a variable
#' ## named "s_p" in the (*reticulate*'s) `py` Python environment. Assigning
#' ## the variable to the Python environment has performance advantages, as
#' ## any Python code applied to the MS data does not require any data
#' ## conversions.
#' py_set_attr(py, "s_p", rspec_to_pyspec(s))
#'
#' ## Create a `MsBackendPy` representing an interface to the data in the
#' ## "s_p" variable in Python:
#' be <- backendInitialize(MsBackendPy(), "s_p")
#' be
#'
#' ## Alternatively, by passing the full MS data with parameter `data`, the
#' ## data is first converted to Python and the backend is initialized with
#' ## that data. The `setBackend()` call from above internally uses this
#' ## code to convert the data.
#' be <- backendInitialize(MsBackendPy(), "s_p3",
#'     data = spectraData(s, c(spectraVariables(s), "mz", "intensity")))
#'
#' ## Create a Spectra object which this backend:
#' s_2 <- Spectra(be)
#' s_2
#'
#' ## An easier way to change the data representation of a `Spectra` object
#' ## from R to Python is to use the `Spectra`'s `setBackend()` method
#' ## selecting a `MsBackendPy` as the target backend representation:
#' s_2 <- setBackend(s, MsBackendPy(), pythonVariableName = "s_p2")
#' s_2
#'
#' ## This moved the data from R to Python, storing it in a Python variable
#' ## with the name `s_p2`. The resulting `s_2` is thus a `Spectra` object
#' ## with all MS data however stored in Python.
#'
#' ## Note that by default only spectra variables that are part of
#' ## `defaultSpectraVariableMapping()` are converted to Python
#' defaultSpectraVariableMapping()
#'
#' ## Thus, for example the precursor m/z is available in `s_2`, but other
#' ## spectra variables from `s`, such as `"SMILES"` are not:
#' precursorMz(s)
#' precursorMz(s_2)
#'
#' s$SMILES |> head()
#' ## s_2$SMILES would throw an error.
#'
#' ## To also translate this spectra variable, it needs to be included and
#' ## specified with the `spectraVariableMapping` parameter. The easiest
#' ## approach is to use the `spectraVariableMapping()` function adding in
#' ## addition to the default mapping for the Python library (`"matchms"`)
#' ## also the mapping of additional spectra variables that should be converted:
#' s_2 <- setBackend(s, MsBackendPy(), pythonVariableName = "s_p2",
#'     spectraVariableMapping = spectraVariableMapping("matchms", c(SMILES = "smiles")))
#' s_2$SMILES |> head()
#'
#' ## Available spectra variables: these include, next to the *core* spectra
#' ## variables, also the names of all metadata stored in the `matchms.Spectrum`
#' ## objects.
#' spectraVariables(s_2)
#'
#' ## Get the full peaks data:
#' peaksData(s_2)
#'
#' ## Get the peaks from the first spectrum
#' peaksData(s_2)[[1L]]
#'
#' ## Get the full spectra data:
#' spectraData(s_2)
#'
#' ## Get the m/z values
#' mz(s_2)
#'
#' ## Plot the first spectrum
#' plotSpectra(s_2[1L])
#'
#'
#' ########
#' ## Using the spectrum_utils Python library
#'
#' ## Below we convert the data to a list of `MsmsSpectrum` object from the
#' ## spectrum_utils library.
#' py_set_attr(py, "su_p", rspec_to_pyspec(s,
#'     spectraVariableMapping("spectrum_utils"), "spectrum_utils"))
#'
#' ## Create a MsBackendPy representing this data. Importantly, we need to
#' ## specify the Python library using the `pythonLibrary` parameter and
#' ## ideally also set the `spectraVariableMapping` to the one specific for
#' ## that library.
#' be <- backendInitialize(MsBackendPy(), "su_p",
#'     spectraVariableMapping = spectraVariableMapping("spectrum_utils"),
#'     pythonLibrary = "spectrum_utils")
#' be
#'
#' ## Get the peaks data for the first 3 spectra
#' peaksData(be[1:3])
#'
#' ## Get the full spectraData
#' spectraData(be)
#'
#' ## Extract the precursor m/z
#' be$precursorMz
NULL

#' @noRd
#'
#' @slot py_var `character(1)` with the **name** of the variable. The variable
#'     could be defined in R or in Python.
#'
#' @slot py_lib `character(1)` defining the Python library representing the
#'     data. Currently not used/defaults to `matchms`.
#'
#' @slot is_in_py `logical(1)` whether the variable is in Python (py$) or in
#'     the R environment.
#'
#' @slot i `integer` with the integer/index of the spectra from the backend
#'     we should represent. This allows to subset using `[` **without**
#'     changing the original data.
setClass("MsBackendPy",
         contains = "MsBackend",
         slots = c(py_var = "character",
                   py_lib = "character",
                   is_in_py = "logical",
                   i = "integer",
                   spectraVariableMapping = "character"),
         prototype = prototype(
             py_var = character(),
             py_lib = "matchms",
             is_in_py = TRUE,
             i = integer(),
             spectraVariableMapping = defaultSpectraVariableMapping(),
             version = "0.1"))

#' @export
MsBackendPy <- function() {
    new("MsBackendPy")
}

#' @importMethodsFrom ProtGenerics backendInitialize
#'
#' @rdname MsBackendPy
setMethod("backendInitialize", "MsBackendPy",
          function(object, pythonVariableName = character(),
                   spectraVariableMapping = defaultSpectraVariableMapping(),
                   pythonLibrary = c("matchms", "spectrum_utils"),
                   ..., data) {
              pythonLibrary <- match.arg(pythonLibrary)
              .check_spectra_variable_mapping(spectraVariableMapping)
              object@spectraVariableMapping <- spectraVariableMapping
              if (!length(pythonVariableName))
                  stop("'pythonVariableName' has to be provided", call. = FALSE)
              if (!is.character(pythonVariableName))
                  stop("'pythonVariableName' is expected to be the name ",
                       "of the variable with the data.", call. = FALSE)
              if (missing(data)) {
                  ## 1) Check if the variable exists - and whether it's in py or
                  ##    in R. Set the `is_in_py` variable depending on that.
                  if (.exists_py_var(pythonVariableName))
                      object@is_in_py <- TRUE
                  else
                      object@is_in_py <- FALSE
                  .check_py_var_exists(pythonVariableName, object@is_in_py)
                  object@py_var <- pythonVariableName
                  ## 2) Check if the variable is of the correct data type.
                  .check_py_var(pythonVariableName, object@is_in_py)
                  object@py_var <- pythonVariableName
                  object <- reindex(object)
                  object@py_lib <- pythonLibrary
                  object
              } else {
                  if (!is(data, "DataFrame"))
                      stop("'data' is expected to be a 'DataFrame'",
                           call. = FALSE)
                  if (!all(c("mz", "intensity") %in% colnames(data)))
                      stop("Columns \"mz\" and \"intensity\" are required",
                           call. = FALSE)
                  ## drop NA core spectra variables!
                  data <- .drop_na_core_spectra_variables(data)
                  spectraVariableMapping <- spectraVariableMapping[
                      names(spectraVariableMapping) %in% colnames(data)]
                  py_set_attr(
                      py, pythonVariableName,
                      switch(pythonLibrary,
                             matchms = .rspec_to_matchms_pyspec(
                                 data, mapping = spectraVariableMapping),
                             spectrum_utils = .rspec_to_spectrum_utils_pyspec(
                                 data, mapping = spectraVariableMapping)))
                  backendInitialize(
                      object, pythonVariableName = pythonVariableName,
                      spectraVariableMapping = spectraVariableMapping,
                      pythonLibrary = pythonLibrary)
              }
          })

#' @importMethodsFrom methods show
setMethod("show", "MsBackendPy", function(object) {
    .check_i(object)
    l <- length(object)
    cat(class(object), "with", l, "spectra\n")
    cat("Data stored in the \"", object@py_var,"\" variable ",
        ifelse(object@is_in_py, "in Python", "in R"), "\n", sep = "")
})

#' @rdname MsBackendPy
setMethod("length", "MsBackendPy", function(x) {
    length(x@i)
})

#' @importMethodsFrom ProtGenerics spectraVariables
#'
#' @importFrom Spectra coreSpectraVariables
#'
#' @rdname MsBackendPy
setMethod("spectraVariables", "MsBackendPy", function(object) {
    if (length(object)) {
        var <- .py_get_metadata_names(object)
        i <- match(var, object@spectraVariableMapping)
        if (length(i))
            var[!is.na(i)] <- names(object@spectraVariableMapping)[i[!is.na(i)]]
    } else var <- character()
    union(names(coreSpectraVariables()), var)
})

#' @importMethodsFrom ProtGenerics spectraData
#'
#' @importFrom Spectra fillCoreSpectraVariables coreSpectraVariables
#'
#' @importFrom IRanges NumericList
#'
#' @importFrom methods as
#'
#' @rdname MsBackendPy
setMethod(
    "spectraData", "MsBackendPy",
    function(object, columns = spectraVariables(object), drop = FALSE) {
        if (!length(object))
            return(as(fillCoreSpectraVariables(
                data.frame()), "DataFrame")[, columns, drop = drop])
        mt <- .py_get_metadata_names(object)
        mt <- mt[names(mt) %in% columns]
        if (length(mt)) {
            SFUN <- switch(
                object@py_lib,
                matchms = .py_matchms_spectrum_spectra_data,
                spectrum_utils = .py_spectrum_utils_spectrum_spectra_data)
            ## Note: creating a pandas.DataFrame in Python is not faster
            d <- .get_py_var(object@py_var, object@is_in_py)
            if (!is.list(d)) d <- py_to_r(d)
            res <- do.call(
                rbindFill,
                lapply(d[object@i], SFUN, mt))
            ## Ensure correct data type for core variables.
            csv <- coreSpectraVariables()
            for (mtc in intersect(colnames(res), names(csv)))
                res[[mtc]] <- as(res[[mtc]], csv[mtc])
        } else
            res <- as.data.frame(matrix(ncol = 0, nrow = length(object)))
        n <- names(coreSpectraVariables())
        res <- fillCoreSpectraVariables(res, n[!n %in% c("mz", "intensity")])
        if (any(columns == "mz")) {
            res <- as(res, "DataFrame")
            res$mz <- NumericList(
                peaksData(object, "mz", TRUE), compress = FALSE)
        }
        if (any(columns == "intensity")) {
            res <- as(res, "DataFrame")
            res$intensity <- NumericList(
                peaksData(object, "intensity", TRUE), compress = FALSE)
        }
        if (any(columns == "dataStorage"))
            res$dataStorage <- object@py_var
        if (!all(columns %in% colnames(res)))
            stop("Column(s) ",
                 paste0("\"", columns[!columns %in% colnames(res)], "\"",
                        collapse = ", "), " not available.", call. = FALSE)
        if (drop && length(columns) == 1)
            res[, columns, drop = TRUE]
        else as(res[, columns, drop = FALSE], "DataFrame")
    })

#' @importMethodsFrom ProtGenerics spectraData<-
#'
#' @rdname MsBackendPy
#'
#' @export
setReplaceMethod("spectraData", "MsBackendPy", function(object, value) {
    if (nrow(value) != length(object))
        stop("The number of rows of the provided 'DataFrame' has to match ",
             "the number of spectra (=", length(object), ").")
    svm <- object@spectraVariableMapping
    ## keep only variable mapping for which we have values in 'value'
    svm <- svm[names(svm) %in% colnames(value)]
    miss <- setdiff(colnames(value), c(names(svm), c("mz", "intensity")))
    svm <- c(svm, .py_variable_map(miss))
    backendInitialize(object, pythonVariableName = object@py_var,
                      pythonLibrary = object@py_lib,
                      spectraVariableMapping = svm,
                      data = value)
})

#' Helper function to choose Python names that match the core ones.
#'
#' @param x `character` with spectra variables.
#'
#' @return named `character` with names corresponding to the spectra variables
#'     and values the respective names in Python.
#'
#' @noRd
.py_variable_map <- function(x = character(), library = "matchms") {
    cm <- spectraVariableMapping(library)
    names(x) <- x
    idx <- match(x, names(cm))
    nna <- !is.na(idx)
    if (any(nna))
        x[nna] <- cm[idx[nna]]
    x
}

#' @importMethodsFrom ProtGenerics peaksData
#'
#' @importFrom reticulate iterate
#'
#' @rdname MsBackendPy
setMethod(
    "peaksData",
    "MsBackendPy",
    function(object, columns = c("mz", "intensity"), drop = FALSE) {
        if (length(object@py_var)) {
            res <- switch(
                object@py_lib,
                matchms = .py_matchms_peaks_data(
                    object@py_var, (object@i - 1L)),
                spectrum_utils = .py_spectrum_utils_peaks_data(
                    object@py_var, (object@i - 1L)))
            if (identical(as.vector(columns), c("mz", "intensity"))) {
                res <- lapply(res, function(z) {
                    colnames(z) <- c("mz", "intensity")
                    z
                })
            } else {
                res <- lapply(res, function(z) {
                    colnames(z) <- c("mz", "intensity")
                    z[, columns, drop = drop]
                })
            }
            res
        } else list()
    })

#' @importMethodsFrom ProtGenerics peaksData<-
#'
#' @rdname MsBackendPy
setReplaceMethod("peaksData", "MsBackendPy", function(object, value) {
    Spectra:::.check_peaks_data_value(value, length(object))
    spd <- spectraData(object, union(names(spectraVariableMapping(object)),
                                     peaksVariables(object)))
    cns <- colnames(value[[1L]])
    for (cn in cns) {
        vals <- lapply(value, "[", , cn)
        if (cn %in% c("mz", "intensity"))
            vals <- NumericList(vals, compress = FALSE)
        spd[[cn]] <- vals
    }
    spectraData(object) <- spd
    object
})

#' @importMethodsFrom ProtGenerics extractByIndex
setMethod("extractByIndex", c("MsBackendPy", "ANY"), function(object, i) {
    object@i <- object@i[i]
    .check_i(object)
    object
})

#' @rdname MsBackendPy
setMethod("$", "MsBackendPy", function(x, name) {
    spectraData(x, name, drop = TRUE)
})

#' @rdname MsBackendPy
#'
#' @importMethodsFrom Spectra peaksVariables
setReplaceMethod("$", "MsBackendPy", function(x, name, value) {
    spd <- spectraData(
        x, union(names(spectraVariableMapping(x)), peaksVariables(x)))
    spd[[name]] <- value
    spectraData(x) <- spd
    x
})

setMethod("lengths", "MsBackendPy", function(x, use.names = FALSE) {
    vapply(peaksData(x), nrow, NA_integer_)
})

#' @importMethodsFrom S4Vectors isEmpty
setMethod("isEmpty", "MsBackendPy", function(x) {
    lengths(x) == 0L
})

#' @importMethodsFrom ProtGenerics acquisitionNum
setMethod("acquisitionNum", "MsBackendPy", function(object) {
    spectraData(object, "acquisitionNum", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics centroided
setMethod("centroided", "MsBackendPy", function(object) {
    spectraData(object, "centroided", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics collisionEnergy
setMethod("collisionEnergy", "MsBackendPy", function(object) {
    spectraData(object, "collisionEnergy", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics dataOrigin
setMethod("dataOrigin", "MsBackendPy", function(object) {
    spectraData(object, "dataOrigin", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics intensity
setMethod("intensity", "MsBackendPy", function(object) {
    NumericList(peaksData(object, "intensity", drop = TRUE), compress = FALSE)
})

#' @importMethodsFrom ProtGenerics intensity<-
#'
#' @rdname MsBackendPy
setReplaceMethod("intensity", "MsBackendPy", function(object, value) {
    .check_mz_intensity(value, length(object), lengths(object))
    spd <- spectraData(object, union(names(spectraVariableMapping(object)),
                                     peaksVariables(object)))
    spd[["intensity"]] <- value
    spectraData(object) <- spd
    object
})

#' @importMethodsFrom ProtGenerics isolationWindowLowerMz
setMethod("isolationWindowLowerMz", "MsBackendPy", function(object) {
    spectraData(object, "isolationWindowLowerMz", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics isolationWindowTargetMz
setMethod("isolationWindowTargetMz", "MsBackendPy", function(object) {
    spectraData(object, "isolationWindowTargetMz", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics isolationWindowUpperMz
setMethod("isolationWindowUpperMz", "MsBackendPy", function(object) {
    spectraData(object, "isolationWindowUpperMz", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics msLevel
setMethod("msLevel", "MsBackendPy", function(object) {
    spectraData(object, "msLevel", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics mz
setMethod("mz", "MsBackendPy", function(object) {
    NumericList(peaksData(object, "mz", drop = TRUE), compress = FALSE)
})

#' @importMethodsFrom ProtGenerics mz<-
#'
#' @rdname MsBackendPy
setReplaceMethod("mz", "MsBackendPy", function(object, value) {
    .check_mz_intensity(value, length(object), lengths(object))
    spd <- spectraData(object, union(names(spectraVariableMapping(object)),
                                     peaksVariables(object)))
    spd[["mz"]] <- value
    spectraData(object) <- spd
    object
})

#' @importMethodsFrom ProtGenerics polarity
setMethod("polarity", "MsBackendPy", function(object) {
    spectraData(object, "polarity", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics precScanNum
setMethod("precScanNum", "MsBackendPy", function(object) {
    spectraData(object, "precScanNum", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics precursorCharge
setMethod("precursorCharge", "MsBackendPy", function(object) {
    spectraData(object, "precursorCharge", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics precursorIntensity
setMethod("precursorIntensity", "MsBackendPy", function(object) {
    spectraData(object, "precursorIntensity", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics precursorMz
setMethod("precursorMz", "MsBackendPy", function(object) {
    spectraData(object, "precursorMz", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics rtime
setMethod("rtime", "MsBackendPy", function(object) {
    res <- spectraData(object, "rtime", drop = TRUE)
    l <- length(object)
    if (length(res) != l)
        rep(NA_real_, l)
    else res
})

#' @importMethodsFrom ProtGenerics scanIndex
setMethod("scanIndex", "MsBackendPy", function(object) {
    spectraData(object, "scanIndex", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics smoothed
setMethod("smoothed", "MsBackendPy", function(object) {
    spectraData(object, "smoothed", drop = TRUE)
})

#' @importMethodsFrom ProtGenerics spectraNames
setMethod("spectraNames", "MsBackendPy", function(object) {
    rownames(spectraData(object))
})

#' @importMethodsFrom ProtGenerics tic
setMethod("tic", "MsBackendPy", function(object, initial = TRUE) {
    if (initial) {
        if (any(spectraVariables(object) == "totIonCurrent"))
            spectraData(object, "totIonCurrent")[, 1L]
        else rep(NA_real_, length(object))
    } else vapply(intensity(object), sum, numeric(1), na.rm = TRUE)
})

#' @importMethodsFrom Spectra spectraVariableMapping<-
#'
#' @rdname MsBackendPy
#'
#' @export
setReplaceMethod(
    "spectraVariableMapping", "MsBackendPy", function(object, value) {
        .check_spectra_variable_mapping(value)
        object@spectraVariableMapping <- value
        object
    })

#' @importMethodsFrom Spectra spectraVariableMapping
#'
#' @exportMethod spectraVariableMapping
#'
#' @rdname MsBackendPy
setMethod("spectraVariableMapping", "MsBackendPy", function(object, value) {
    object@spectraVariableMapping
})

#' @rdname MsBackendPy
#'
#' @export
reindex <- function(object) {
    object@i <- seq_len(.py_var_length(object))
    object
}

################################################################################
##        HELPER FUNCTIONS
################################################################################
.check_spectra_variable_mapping <- function(x) {
    if (length(x) && !length(names(x)))
        stop("'spectraVariableMapping' needs to be a named character vector")
    else TRUE
}

#' Helper to get the `py` variable from *reticulate*.
#'
#' @noRd
.get_py <- function() {
    tryCatch({
        get("py")
    }, error = function(e) {
        stop("Failed to get variable 'py'. Is the 'reticulate' package loaded?")
    })
}

#' @importFrom reticulate py_has_attr
.exists_py_var <- function(x) {
    py_has_attr(.get_py(), x)
}

#' Checks if the variable with the name `x` **exists** in Python (if
#' `is_in_py = TRUE`) or in R (if `is_in_py = FALSE`). It throws and error
#' if there is none.
#'
#' @noRd
.check_py_var_exists <- function(x, is_in_py = TRUE) {
    if ((is_in_py && !.exists_py_var(x)) |
        (!is_in_py && !length(get0(x))))
        stop("No variable of name \"", x, "\" found",
             " in the Python or R environment.")
    TRUE
}

.check_py_var <- function(x, is_in_py = TRUE) {
    var <- .get_py_var(x, is_in_py)
    if (!is(var, "python.builtin.list"))
        stop("Variable \"", x, "\" is supposed to be a Python list, ",
             "but it is a ", class(x)[1L])
        TRUE
}

#' Check that all indices are within the range of the data in Python.
#'
#' @param object `MsBackendPy` object.
#'
#' @return `TRUE` if all is OK - otherwise it throws an error
#'
#' @noRd
.check_i <- function(object) {
    l <- .py_var_length(object)
    if (l > 0 && max(object@i) > l)
        stop("Indices are out of bound. Use `reindex()` to update indices ",
             "of 'object' if \"", object@py_var, "\" was subset by ",
             "another process.", call. = FALSE)
    TRUE
}

#' Helper function to get the variable with the Python data. Can be either
#' in R or in Python.
#'
#' @param x `character(1)` with the name of the variable.
#'
#' @param is_in_py `logical(1)` whether the variable is stored in the `py`
#'     module (`is_in_py = TRUE`, the default) or in the R environment
#'     (`is_in_py = FALSE`).
#'
#' @return the variable for which then name was provided.
#'
#' @noRd
#'
#' @importFrom reticulate py_get_attr
.get_py_var <- function(x, is_in_py = TRUE) {
    if (is_in_py) {
        py_get_attr(.get_py(), x)
    } else get0(x)
}

#' @importFrom reticulate py_run_string py_eval
.py_var_length <- function(x) {
    if (length(x@py_var)) {
        if (x@is_in_py) py_eval(paste0("len(", x@py_var, ")"))
        else length(.get_py_var(x@py_var, FALSE))
    } else 0L
}

#' Get the names of the metadata fields from the referenced Python object.
#' Returns a named vector, elements being the metadata names in Python and
#' names the respective spectra variable names, i.e. the names that should
#' be used in R.
#'
#' @param x `MsBackendPy`
#'
#' @noRd
.py_get_metadata_names <- function(x) {
    var <- .get_py_var(x@py_var, x@is_in_py)
    if (length(var)) {
        switch(
            x@py_lib,
            matchms = {
                mt <- names(var[0L]$metadata)
                names(mt) <- names(
                    x@spectraVariableMapping)[match(
                                      mt, x@spectraVariableMapping)]
                nas <- is.na(names(mt))
                names(mt)[nas] <- mt[nas]
                mt
            },
            spectrum_utils = {
                mt <- c("identifier", "precursor_mz",
                        "precursor_charge", "retention_time")
                mt <- x@spectraVariableMapping[x@spectraVariableMapping %in% mt]
                mt
            })
    }
    else character()
}

#' helper function to drop all columns for core spectra variables that contain
#' only missing values.
#'
#' @importFrom MsCoreUtils vapply1l
#'
#' @noRd
.drop_na_core_spectra_variables <- function(x) {
    svs <- intersect(colnames(x), names(coreSpectraVariables()))
    svs <- svs[!svs %in% c("mz", "intensity")]
    keep <- !vapply1l(svs, function(z) {
        all(is.na(x[[z]]))
    })
    keep <- c(setdiff(colnames(x), svs), svs[keep])
    x[, colnames(x) %in% keep, drop = FALSE]
}

#' helper to check input/validity of intensity or mz: has to be a list-like
#' structure with numeric vectors.
#'
#' @param x `list` or `NumericList`.
#'
#' @param l `integer(1)` with the number of spectra/expected elements in `x`.
#'
#' @param ls `integer` with the number of peaks, i.e., the lengths of the
#'     numeric vectors.
#'
#' @noRd
.check_mz_intensity <- function(x, l = length(x), ls = lengths(x)) {
    if (!(is.list(x) | inherits(x, "SimpleList")))
        stop("'value' has to be a list-like data structure.")
    if (length(x) != l)
        stop("length of 'value' has to match the number of spectra")
    if (!all(lengths(x) == ls))
        stop("lengths(value) has to match the number of peaks per spectrum")
    if (!all(vapply1l(x, is.numeric)))
        stop("elements of 'value' are expected to be numeric vectors.")
}

library(Spectra)
library(MsBackendMgf)
fl <- system.file("extdata", "mgf", "test.mgf", package = "SpectriPy")
s <- Spectra(fl, source = MsBackendMgf())
s$spectrum_index <- seq_along(s)

## Convert the data to Python MS data structures; seems we need to store
## the Python object in Python, otherwise it can not be seen by the unit
## tests below.
py_set_attr(py, "s_p", rspec_to_pyspec(s, c(defaultSpectraVariableMapping(),
                                            spectrum_index = "spctrm_idx")))
expect_true(any(names(py$s_p[[1]]$metadata) == "spctrm_idx"))

## Convert the data also to spectrum_utils
py_set_attr(py, "su_p", rspec_to_pyspec(
                            s, spectraVariableMapping("spectrum_utils"),
                            "spectrum_utils"))
SPECTRUM_UTILS_TOLERANCE <- 1e-7

test_that("MsBackendPy constructor works", {
    res <- MsBackendPy()
    expect_s4_class(res, "MsBackendPy")
    expect_true(res@is_in_py)
    expect_equal(res@spectraVariableMapping, defaultSpectraVariableMapping())
})

test_that(".check_spectra_variable_mapping works", {
    expect_true(.check_spectra_variable_mapping(character()))
    expect_true(.check_spectra_variable_mapping(
        defaultSpectraVariableMapping()))
    expect_error(.check_spectra_variable_mapping("hello"),
                 "named character vector")
})

test_that(".get_py works", {
    res <- .get_py()
    expect_true(is(res, "python.builtin.module"))

    ## expect_error(
    ##     with_mocked_bindings(
    ##         "base::get" = function(x, pos = -1L, envir = as.environment(pos),
    ##                          mode = "any", inherits = TRUE) stop("aaaa"),
    ##         code = .get_py()
    ##     ),
    ##     "Failed to get variable"
    ## )
})

test_that(".exists_py_var works", {
    expect_false(.exists_py_var("what_"))
    expect_true(.exists_py_var("r"))
    expect_true(.exists_py_var("s_p"))
})

test_that(".check_py_var_exists works", {
    expect_true(.check_py_var_exists("r"))
    expect_error(.check_py_var_exists("r", FALSE), "No variable")
    expect_true(.check_py_var_exists("py", FALSE))
    expect_true(.check_py_var_exists("s_p", TRUE))
    ## expect_true(.check_py_var_exists("s", FALSE))
})

test_that(".check_py_var works", {
    expect_true(.check_py_var("s_p", TRUE))
    expect_error(.check_py_var("py", FALSE), "supposed to be")
})

test_that(".get_py_var works", {
    ## For whatever reason I can't get something from the global environment.
    ## res <- .get_py_var("s", FALSE)
    ## expect_s4_class(res, "Spectra")
    py$tmp <- 34
    res <- .get_py_var("tmp", TRUE)
    expect_equal(py_to_r(res), 34)
})

test_that("backendInitialize,MsBackendPy works", {
    expect_error(backendInitialize(MsBackendPy()), "has to be provided")
    expect_error(backendInitialize(MsBackendPy(), 3), "name of the")
    expect_error(backendInitialize(MsBackendPy(), "not exists"),
                 "No variable of name")
    expect_error(backendInitialize(MsBackendPy(), "r"), "Python list")
    res <- backendInitialize(MsBackendPy(), "s_p")
    expect_s4_class(res, "MsBackendPy")
})

test_that("backendInitialize,MsBackendPy works with providing data", {
    expect_error(
        backendInitialize(MsBackendPy(), pythonVariableName = "ttt", data = 4),
        "'DataFrame'")
    d <- spectraData(s)
    expect_error(
        backendInitialize(MsBackendPy(), pythonVariableName = "ttt", data = d),
        "are required")
    d <- spectraData(s@backend)
    res <- backendInitialize(
        MsBackendPy(), pythonVariableName = "ttt", data = d)
    expect_s4_class(res, "MsBackendPy")
    expect_equal(res@py_var, "ttt")
    expect_true(res@is_in_py)
    expect_equal(nrow(d), length(res))
    expect_equal(rtime(s), rtime(res))
    expect_equal(intensity(s), intensity(res))
    expect_equal(mz(s), mz(res))
    expect_true(length(spectraVariables(res)) < length(spectraVariables(s)))
    expect_true(all(spectraVariables(res) %in% names(coreSpectraVariables())))
    expect_true(all(names(coreSpectraVariables()) %in% spectraVariables(res)))

    ## mapping additional variables
    res <- backendInitialize(
        MsBackendPy(), pythonVariableName = "ttt", data = d,
        spectraVariableMapping = spectraVariableMapping(
            "matchms", c(SMILES = "smiles", INCHI = "inchi")))
    expect_true(all(c("SMILES", "INCHI") %in% spectraVariables(res)))
    expect_equal(res$SMILES, s$SMILES)
    expect_equal(res$INCHI, s$INCHI)

    ## Set core variables to NA
    d$rtime <- NA
    d$precursorMz <- NA
    res <- backendInitialize(
        MsBackendPy(), pythonVariableName = "ttt", data = d)
    ## We get a bunch of warnings - because NA was not translated.
    expect_true(all(is.na(res$precursorMz)))
    ## Only non-NA variables have a mapping
    expect_equal(res@spectraVariableMapping,
                 c(precursorCharge = "charge", msLevel = "ms_level"))

    ## Repeat with spectrum_utils
    py_del_attr(py, "ttt")
    res <- backendInitialize(MsBackendPy(), pythonVariableName = "ttt",
                             data = d, pythonLibrary = "spectrum_utils")
    expect_s4_class(res, "MsBackendPy")
    expect_equal(res@py_var, "ttt")
    expect_true(res@is_in_py)
    expect_equal(nrow(d), length(res))
    expect_equal(rtime(s), rtime(res))
    expect_equal(intensity(s), intensity(res),
                 tolerance = SPECTRUM_UTILS_TOLERANCE)
    expect_equal(mz(s), mz(res), tolerance = SPECTRUM_UTILS_TOLERANCE)
})

test_that("setBackend,Spectra backend = MsBackendPy works", {
    expect_error(a <- setBackend(s, MsBackendPy()), "'pythonVariableName'")
    a <- setBackend(s, MsBackendPy(), pythonVariableName = "tmp_s_p")
    expect_s4_class(a, "Spectra")
    expect_s4_class(a@backend, "MsBackendPy")
    expect_equal(rtime(a), rtime(s))
    expect_equal(mz(a), mz(s))
    expect_equal(intensity(a), intensity(s))
})

test_that("show,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    expect_output(show(be), "MsBackendPy")
})

test_that(".py_var_length, length,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    expect_equal(.py_var_length(be), 100L)
    expect_equal(.py_var_length(MsBackendPy()), 0L)
    expect_equal(length(be), 100L)
    expect_equal(length(MsBackendPy()), 0L)

    ## dddd <- 1:10
    ## be2 <- MsBackendPy()
    ## be2@is_in_py <- FALSE
    ## be2@py_var <- "dddd"
    ## expect_equal(SpectriPy:::.py_var_length(be2), 10L)
})

test_that("spectraVariables and .py_get_metadata_names works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    vars <- SpectriPy:::.py_get_metadata_names(be)
    expect_true(is.character(vars))
    expect_true(
        all(c("charge", "collision_energy", "ms_level",
              "precursor_intensity", "precursor_mz", "retention_time") %in%
            vars))
    expect_true(any(vars == "spctrm_idx"))
    expect_true(any(names(vars) == "spctrm_idx"))
    res <- spectraVariables(be)
    expect_true(all(names(coreSpectraVariables()) %in% res))
    expect_true("spctrm_idx" %in% res)

    m <- c(defaultSpectraVariableMapping(), spectrum_index = "spctrm_idx")
    be@spectraVariableMapping <- m
    res <- spectraVariables(be)
    expect_true(all(names(coreSpectraVariables()) %in% res))
    expect_true("spectrum_index" %in% res)

    ## spectrum_utils
    be <- backendInitialize(MsBackendPy(), "su_p",
                            pythonLibrary = "spectrum_utils")
    m <- .py_get_metadata_names(be)
    expect_equal(unname(m), c("precursor_mz", "retention_time"))
    be <- backendInitialize(
        MsBackendPy(), "su_p",
        spectraVariableMapping = spectraVariableMapping("spectrum_utils"),
        pythonLibrary = "spectrum_utils")
    m <- .py_get_metadata_names(be)
    expect_equal(unname(m), c("precursor_mz", "precursor_charge",
                      "retention_time", "identifier"))
    res <- spectraVariables(be)
    expect_true(all(names(m) %in% res))

    be2 <- MsBackendPy()
    be2@py_var <- "hey"
    be2@is_in_py <- FALSE
    hey <- integer()
    expect_equal(.py_get_metadata_names(be2), character())
})

test_that("peaksData,MsBackendPy works", {
    ## Empty data
    be <- MsBackendPy()
    res <- peaksData(be)
    expect_equal(res, list())

    ## Read data
    be <- backendInitialize(MsBackendPy(), "s_p")
    expect_equal(be@i, 1:100)
    res <- peaksData(be)
    expect_equal(res, peaksData(s@backend))

    res <- peaksData(be, c("intensity", "mz"))
    expect_equal(res, peaksData(s@backend, c("intensity", "mz")))
    expect_equal(colnames(res[[1L]]), c("intensity", "mz"))

    res <- peaksData(be, "mz")
    expect_equal(res, peaksData(s@backend, "mz"))
    expect_equal(colnames(res[[1L]]), c("mz"))

    res <- peaksData(be, "mz", drop = TRUE)
    expect_true(is.list(res))
    expect_equal(length(res), 100L)
    expect_true(is.numeric(res[[1L]]))
    expect_false(is.matrix(res[[1L]]))

    ## spectrum_utils
    be <- backendInitialize(MsBackendPy(), "su_p",
                            pythonLibrary = "spectrum_utils")
    res <- peaksData(be)
    expect_true(is.list(res))
    expect_equal(length(res), length(be))
    expect_equal(res, peaksData(s, return.type = "list"),
                 tolerance = SPECTRUM_UTILS_TOLERANCE)
    a <- peaksData(be[4])
    expect_equal(a, res[4])
})

test_that("spectraData,MsBackendPy works", {
    be <- MsBackendPy()
    res <- spectraData(be)
    expect_true(nrow(res) == 0)
    expect_true(is(res, "DataFrame"))

    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- spectraData(be)
    expect_true(nrow(res) == 100)
    expect_s4_class(res, "DataFrame")
    expect_true(all(names(coreSpectraVariables()) %in% colnames(res)))
    expect_true(all(names(be@spectraVariableMapping) %in% colnames(res)))
    expect_true(any(colnames(res) == "spctrm_idx"))

    ## arbitrary order
    cols <- c("rtime", "msLevel", "spctrm_idx", "precursorMz", "polarity")
    res <- spectraData(be, cols)
    expect_equal(colnames(res), cols)
    expect_equal(res$precursorMz, s$precursorMz)
    expect_equal(res$msLevel, s$msLevel)

    ## individual columns
    res <- spectraData(be, "msLevel")
    expect_s4_class(res, "DataFrame")
    expect_equal(res$msLevel, s$msLevel)

    res <- spectraData(be, "spctrm_idx")
    expect_s4_class(res, "DataFrame")
    expect_equal(colnames(res), "spctrm_idx")
    expect_equal(res$spctrm_idx, s$spectrum_index)

    res <- spectraData(be, "smoothed")
    expect_s4_class(res, "DataFrame")
    expect_equal(colnames(res), "smoothed")
    expect_equal(res$smoothed, s$smoothed)

    ## only mz
    res <- spectraData(be, "mz")
    expect_s4_class(res, "DataFrame")
    expect_equal(colnames(res), "mz")
    expect_equal(res$mz, s$mz)

    ## only intensity
    res <- spectraData(be, "intensity")
    expect_s4_class(res, "DataFrame")
    expect_equal(colnames(res), "intensity")
    expect_equal(res$intensity, s$intensity)

    ## only core variables not in backend
    cols <- c("polarity", "smoothed", "centroided")
    res <- spectraData(be, cols)
    expect_s4_class(res, "DataFrame")
    expect_equal(colnames(res), cols)
    expect_true(all(is.na(res)))

    ## only variables that are mapped
    cols <- c("precursorMz", "precursorCharge", "msLevel")
    res <- spectraData(be, cols)
    expect_s4_class(res, "DataFrame")
    expect_equal(colnames(res), cols)
    expect_equal(res$precursorMz, s$precursorMz)
    expect_equal(res$precursorCharge, s$precursorCharge)
    expect_equal(res$msLevel, s$msLevel)

    ## only one with drop TRUE
    res <- spectraData(be, "msLevel", drop = TRUE)
    expect_equal(res, s$msLevel)

    ## non-existant column
    expect_error(spectraData(be, c("msLevel", "other_col")), "not available")

    ## custom spectravariablemapping
    m <- c(defaultSpectraVariableMapping(),
           spectrum_index = "spctrm_idx")
    be@spectraVariableMapping <- m
    res <- spectraData(be)
    expect_true(all(names(m) %in% colnames(res)))
    expect_equal(res$spectrum_index, seq_along(be))

    ## spectrum_utils
    be <- backendInitialize(
        MsBackendPy(), "su_p", pythonLibrary = "spectrum_utils",
        spectraVariableMapping("spectrum_utils"))
    res <- spectraData(be)
    expect_s4_class(res, "DataFrame")
    expect_equal(res$dataStorage, rep("su_p", length(be)))
    expect_equal(res$precursorMz, s$precursorMz)
})

test_that(".check_i works", {
    expect_true(.check_i(MsBackendPy()))
    be <- backendInitialize(MsBackendPy(), "s_p")
    expect_true(.check_i(be))
    be@i <- 1233:3431
    expect_error(.check_i(be), "out of bound")
})

test_that("reindex works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    expect_equal(be@i, 1:100)
    be@i <- c(4L, 3L)
    be <- reindex(be)
    expect_equal(be@i, 1:100)
})

test_that("extractByIndex and [,MsBackendPy work", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- be[c(43, 2)]
    expect_equal(res@i, c(43L, 2L))

    expect_equal(peaksData(res), peaksData(s@backend)[c(43, 2)])
    cols <- c("msLevel", "precursorMz", "rtime")
    expect_equal(spectraData(res, cols),
                 spectraData(s@backend, cols)[c(43, 2), ])

    ## Single element
    res <- be[3L]
    expect_equal(peaksData(res), peaksData(s@backend)[3L])
})

test_that("$,MsBackendPy works", {
    be <- MsBackendPy()
    expect_equal(be$msLevel, integer())
    be <- backendInitialize(MsBackendPy(), "s_p")
    expect_equal(be$msLevel, s$msLevel)
    expect_equal(be$precursorMz, s$precursorMz)
    expect_error(be$not_exists, "not available")

    be <- backendInitialize(
        MsBackendPy(), "su_p",
        spectraVariableMapping = spectraVariableMapping("spectrum_utils"),
        pythonLibrary = "spectrum_utils")
    expect_equal(be$precursorMz, s$precursorMz)
})

test_that("lengths,MsBackendPy works", {
    be <- MsBackendPy()
    expect_equal(lengths(be), integer(0))

    be <- backendInitialize(MsBackendPy(), "s_p")
    expect_equal(lengths(be), lengths(s))
})

test_that("isEmpty,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    expect_false(any(isEmpty(be)))
})

test_that("acquisitionNum,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- acquisitionNum(be)
    expect_true(is.integer(res))
    expect_true(all(is.na(res)))
})

test_that("centroided,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- centroided(be)
    expect_true(is.logical(res))
    expect_true(all(is.na(res)))
})

test_that("collisionEnergy,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- collisionEnergy(be)
    expect_true(is.numeric(res))
    expect_true(all(is.na(res)))
})

test_that("dataOrigin,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- dataOrigin(be)
    expect_true(is.character(res))
    expect_true(all(is.na(res)))
})

test_that("intensity,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- intensity(be)
    expect_equal(res, intensity(s@backend))
    expect_s4_class(res, "NumericList")
})

test_that("isolationWindowLowerMz,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- isolationWindowLowerMz(be)
    expect_true(is.numeric(res))
    expect_equal(res, isolationWindowLowerMz(s@backend))
    expect_true(all(is.na(res)))
})

test_that("isolationWindowUpperMz,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- isolationWindowUpperMz(be)
    expect_true(is.numeric(res))
    expect_equal(res, isolationWindowUpperMz(s@backend))
    expect_true(all(is.na(res)))
})

test_that("isolationWindowTargetMz,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- isolationWindowTargetMz(be)
    expect_true(is.numeric(res))
    expect_equal(res, isolationWindowTargetMz(s@backend))
    expect_true(all(is.na(res)))
})

test_that("msLevel,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- msLevel(be)
    expect_true(is.integer(res))
    expect_equal(res, msLevel(s@backend))
})

test_that("mz,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- mz(be)
    expect_equal(res, mz(s@backend))
    expect_s4_class(res, "NumericList")
})

test_that("polarity,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- polarity(be)
    expect_true(is.integer(res))
    expect_equal(res, polarity(s@backend))
    expect_true(all(is.na(res)))
})

test_that("precScanNum,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- precScanNum(be)
    expect_true(is.integer(res))
    expect_equal(res, precScanNum(s@backend))
    expect_true(all(is.na(res)))
})

test_that("precursorCharge,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- precursorCharge(be)
    expect_true(is.integer(res))
    expect_equal(res, precursorCharge(s@backend))
})

test_that("precursorIntensity,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- precursorIntensity(be)
    expect_true(is.numeric(res))
    expect_equal(res, precursorIntensity(s@backend))
})

test_that("precursorMz,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- precursorMz(be)
    expect_true(is.numeric(res))
    expect_equal(res, precursorMz(s@backend))
})

test_that("rtime,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- rtime(be)
    expect_true(is.numeric(res))
    expect_equal(res, rtime(s@backend))

    s2 <- s[1:4]
    s2$rtime <- c(1.2, 3.4, 5.3, 5.9)
    py_set_attr(py, "s_p2",
                rspec_to_pyspec(s2, defaultSpectraVariableMapping()))
    be <- backendInitialize(MsBackendPy(), "s_p2")
    expect_equal(rtime(be), rtime(s2))
})

test_that("scanIndex,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- scanIndex(be)
    expect_true(is.integer(res))
    expect_equal(res, scanIndex(s@backend))
})

test_that("scanIndex,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- smoothed(be)
    expect_true(is.logical(res))
    expect_equal(res, smoothed(s@backend))
})

test_that("spectraNames,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- spectraNames(be)
    expect_equal(res, spectraNames(s@backend))
})

test_that("tic,MsBackendPy works", {
    be <- backendInitialize(MsBackendPy(), "s_p")
    res <- tic(be, initial = TRUE)
    expect_true(is.numeric(res))
    expect_equal(res, tic(s@backend, initial = TRUE))
    res <- tic(be, initial = FALSE)
    expect_true(is.numeric(res))
    expect_equal(res, tic(s@backend, initial = FALSE))

    s2 <- s[1:4]
    s2$totIonCurrent <- 123.2
    py_set_attr(py, "s_p2",
                rspec_to_pyspec(s2, c(defaultSpectraVariableMapping(),
                                      totIonCurrent = "totIonCurrent")))
    be <- backendInitialize(
        MsBackendPy(), "s_p2",
        spectraVariableMapping = c(defaultSpectraVariableMapping(),
                                   totIonCurrent = "totioncurrent"))
    expect_equal(tic(be, initial = TRUE), c(123.2, 123.2, 123.2, 123.2))
})

test_that("spectraVariableMapping,spectraVariableMapping<- works", {
    s2 <- s[1:4]
    s2$totIonCurrent <- 123.2
    py_set_attr(py, "s_p2",
                rspec_to_pyspec(s2, c(defaultSpectraVariableMapping(),
                                      totIonCurrent = "tot_ion_current")))
    be <- backendInitialize(MsBackendPy(), "s_p2")
    expect_equal(spectraVariableMapping(be), defaultSpectraVariableMapping())

    expect_false("totIonCurrent" %in% spectraVariables(be))
    spectraVariableMapping(be) <- c(spectraVariableMapping(be),
                                    totIonCurrent = "tot_ion_current")
    expect_true("totIonCurrent" %in% spectraVariables(be))

    s3 <- Spectra(be)
    expect_error(spectraVariableMapping(s3) <- "rtime", "needs to be a named")
    spectraVariableMapping(s3) <- c(rtime = "retention_time")
    expect_equal(spectraVariableMapping(s3@backend),
                 c(rtime = "retention_time"))
})

################################################################################
##
##             REPLACEMENT FUNCTIONALITY
##
################################################################################
test_that("spectraData<-,MsBackendPy works", {
    ## Basically, replace the full data from the backend - again.
    ## Expects a `DataFrame` with the same number of rows as there are spectra
    ## in object. This `DataFrame` **has** to contain the m/z and intensity
    ## values! So, basically, we will run a `backendInitialize` with that data.
    a <- setBackend(s, MsBackendPy(),
                    spectraVariableMapping = c(spectrum_index = "spctrm_idx",
                                               defaultSpectraVariableMapping()),
                    pythonVariableName = "set_test")
    ## Subset, this is not synchronized with Python.
    a <- a[20:1]
    expect_true(length(a) == 20)
    expect_true(length(py$set_test) == 100)

    sd <- spectraData(a@backend)
    sd$other_var <- "a"

    spectraData(a@backend) <- sd
    expect_equal(a@backend@i, 1:20)
    expect_true(any(spectraVariables(a) == "other_var"))
    expect_equal(a$other_var, rep("a", 20))
    expect_equal(rtime(a), rtime(s)[20:1])
    expect_equal(mz(a), mz(s)[20:1])

    ## Remove a spectra variable.
    sd <- spectraData(a@backend)
    sd$other_var <- NULL
    sd$centroided <- NULL
    spectraData(a) <- sd
    expect_false(any(spectraVariables(a) == "other_var"))
    expect_equal(centroided(a), rep(NA, 20))

    expect_error(spectraData(a) <- sd[1:4, ], "The number of rows")
})

test_that("$<-,MsBackendPy works", {
    a <- setBackend(
        s, MsBackendPy(), pythonVariableName = "dollar_test",
        spectraVariableMapping = c(spectrum_index = "spctrm_idx",
                                   defaultSpectraVariableMapping()))
    ## set core spectra variable to NULL
    a$msLevel <- NULL
    expect_false(any(names(a@backend@spectraVariableMapping) == "msLevel"))
    expect_true(all(is.na(msLevel(a))))

    ## Re-add the msLevel
    a$msLevel <- 2L
    expect_true(any(names(a@backend@spectraVariableMapping) == "msLevel"))
    expect_true(all(msLevel(a) == 2L))

    ## Add another one.
    a$other_thing <- seq_along(a)
    expect_true(any(names(a@backend@spectraVariableMapping) == "other_thing"))
    expect_equal(a$other_thing, seq_along(a))

    ## Remove it again
    a$other_thing <- NULL
    expect_false(any(names(a@backend@spectraVariableMapping) == "other_thing"))

    expect_error(a$other_thing <- c("a", "b", "c"), "replace")
})

test_that(".py_variable_map works", {
    res <- .py_variable_map(c("msLevel"))
    expect_equal(res, c(msLevel = "ms_level"))
    res <- .py_variable_map(c("other", "msLevel"))
    expect_equal(res, c(other = "other", msLevel = "ms_level"))
    res <- .py_variable_map("other")
    expect_equal(res, c(other = "other"))
    res <- .py_variable_map(c("other", "msLevel", "rtime"))
    expect_equal(res, c(other = "other", msLevel = "ms_level", rtime = "retention_time"))
    res <- .py_variable_map()
    expect_equal(unname(res), character())
})

test_that(".drop_na_core_spectra_variables works", {
    a <- spectraData(s@backend)

    res <- .drop_na_core_spectra_variables(a)
    expect_true(ncol(res) < ncol(a))
    expect_true(any(colnames(res) == "spectrum_index"))

    test_df <- data.frame(a = 1:3, b = "a", c = NA)
    res <- .drop_na_core_spectra_variables(test_df)
    expect_equal(test_df, res)

    test_df$rtime <- NA
    res <- .drop_na_core_spectra_variables(test_df)
    expect_false(any(colnames(res) == "rtime"))

    test_df$rtime <- 1:3
    res <- .drop_na_core_spectra_variables(test_df)
    expect_true(any(colnames(res) == "rtime"))
})

test_that("peaksData<-,MsBackendPy works", {
    a <- setBackend(
        s, MsBackendPy(), pythonVariableName = "pd_test",
        spectraVariableMapping = c(INCHI = "inchi",
                                   defaultSpectraVariableMapping()))@backend
    spd <- spectraData(a)
    svm <- a@spectraVariableMapping
    pd <- peaksData(s)
    expect_error(peaksData(a) <- pd[1:3], "match length")
    pd <- lapply(pd, function(z) {
        z[, 2L] <- z[, 2L] / 2
        z
    })
    peaksData(a) <- pd
    expect_equal(svm, a@spectraVariableMapping)
    pd_2 <- peaksData(a)
    expect_equal(pd, pd_2)
    spd_2 <- spectraData(a)
    expect_equal(colnames(spd), colnames(spd_2))
    expect_equal(lengths(spd$intensity), lengths(spd_2$intensity))
    expect_equal(spd$intensity / 2, spd_2$intensity)
    spd$intensity <- NULL
    spd_2$intensity <- NULL
    expect_equal(spd, spd_2)
})

test_that(".check_mz_intensity works", {
    mzs <- mz(s)
    expect_silent(.check_mz_intensity(mzs, length(mzs), lengths(mzs)))
    expect_silent(.check_mz_intensity(as.list(mzs), length(mzs), lengths(mzs)))
    expect_error(.check_mz_intensity(3, 1, 1), "list-like")
    expect_error(.check_mz_intensity(mzs, 1, 1), "number of spectra")
    expect_error(.check_mz_intensity(mzs[1:3], 3, c(3, 3, 3)),
                 "number of peaks")
    a <- list(c(1:4), c(1.1, 2.2, 3), c("a", "b"))
    expect_error(.check_mz_intensity(a, 4, c(4, 3, 2)), "number of spectra")
    expect_error(.check_mz_intensity(a, 3, c(4, 1, 2)), "number of peaks")
    expect_error(.check_mz_intensity(a, 3, c(4, 3, 2)), "numeric vectors")
})

test_that("mz<-,MsBackendPy works", {
    a <- setBackend(
        s, MsBackendPy(), pythonVariableName = "mz_test",
        spectraVariableMapping = c(INCHI = "inchi",
                                   defaultSpectraVariableMapping()))@backend
    spd <- spectraData(a)
    mzs <- mz(a)
    expect_equal(spd$mz, mzs)
    mzs <- mzs / 3
    mz(a) <- mzs
    expect_equal(mz(a), mzs)
    expect_equal(a$INCHI, spd$INCHI)
    mzs <- as.list(mzs)
    mzs[[1L]] <- mzs[[1L]] * 2
    mz(a) <- mzs
    expect_equal(as.list(mz(a)), mzs)
    ## errors/issues
    expect_error(mz(a) <- mzs[1:3], "match the number")
})

test_that("intensity<-,MsBackendPy works", {
    a <- setBackend(
        s, MsBackendPy(), pythonVariableName = "intensity_test",
        spectraVariableMapping = c(INCHI = "inchi",
                                   defaultSpectraVariableMapping()))@backend
    spd <- spectraData(a)
    ints <- intensity(a)
    expect_equal(spd$intensity, ints)
    ints <- ints / 3
    intensity(a) <- ints
    expect_equal(intensity(a), ints)
    expect_equal(a$INCHI, spd$INCHI)
    ints <- as.list(ints)
    ints[[1L]] <- ints[[1L]] * 2
    intensity(a) <- ints
    expect_equal(as.list(intensity(a)), ints)
    ## errors/issues
    expect_error(intensity(a) <- ints[1:3], "match the number")
})

## Comments, thoughts TODO
## DONE spectraData()<-: replaces the full data and allows adding/removing
##      spectra variables. number of spectra has to match.
## DONE setBackend/backendInitialize: only convert non-NA columns. update the
##      spectraVariableMapping: we can use spectraDataMapping now to add
##      additional columns - missing data will be dropped - and not registered
##      in @spectraVariableMapping
## DONE $<- : replace the full data?
## DONE peaksData()<-: replace the full data?
## DONE mz()<-
## DONE intensity()<-
## TODO applyProcessing(): replace the full data? is that needed? should
##      internally call peaksData()<-
## DONE all other replacement methods.
## TODO support the Spectra unit test suite

library(testthat)
library(Spectra)
library(SpectriPy)
library(MsBackendMgf)

fl <- system.file("extdata", "mgf", "test.mgf", package = "SpectriPy")
s <- Spectra(fl, source = MsBackendMgf())
s <- setBackend(s, MsBackendPy(), pythonVariableName = "py_be")

be <- s@backend

test_check("SpectriPy")

## TODO: add tests from the Spectra unit test suite!

## ## Run the MsBackend spectra variable test suite
## test_suite <- system.file("test_backends", "test_MsBackend",
##                           package = "Spectra")

## ## Run single test file.
## res <- test_dir(test_suite, stop_on_failure = TRUE)

#' @title Converting between R and Python MS data structures
#'
#' @name conversion
#'
#' @description
#'
#' The `rspec_to_pyspec()` and `pyspec_to_rspec()` functions allow to convert
#' (translate) MS data structures between R and Python. At present the
#' R [Spectra::Spectra()] objects can be either translated into a list of
#' [matchms](https://github.com/matchms/matchms) Python `matchms.Spectrum`
#' objects or
#' [spectrum_utils](https://github.com/bittremieux-lab/spectrum_utils) Python
#' `spectrum_utils.spectrum.MsmsSpectrum` objects.
#' For better integration with the *reticulate* R package also a
#' `r_to_py.Spectra()` method is available.
#'
#' The mapping of spectra variables (in R) to (Python) spectra metadata can
#' be configured and defined with the `setSpectraVariableMapping()` and
#' `spectraVariableMapping()`. These get and set the *global* (system wide)
#' setting and are thus also used by the `r_to_py()` method.
#'
#' Properties for translation to the MS data objects of the different Python
#' libraries are:
#'
#' - *matchms*: the `matchms.Spectrum` objects support arbitrary metadata, so
#'   any spectra variable can be translated and stored in these objects.
#'
#' - *spectrum_utils*: the `spectrum_utils.spectrum.MsmsSpectrum` object
#'   supports metadata variables *identifier* (`character`), *precursor_mz*
#'   (`numeric`), *precursor_charge* (`integer`) and optionally also
#'   *retention_time* (`numeric`).
#'
#' See the indivudual function's documentation for more details.
#'
#'
#' @section Translation of MS data objects:
#'
#' MS data structures can be translated between R and Python using the
#' `rspec_to_pyspec()` and `pyspec_to_rspec()` functions, or with the
#' `r_to_py()` method.
#'
#' - `rspec_to_pyspec()` translates an R [Spectra::Spectra()] object into a
#'   list of Python MS data objects, which can be, depending on parameter
#'   `pythonLibrary`, `matchms.Spectrum` objects (for
#'   `pythonLibrary = "matchms"`, the default) or
#'   `spectrum_utils.spectrum.MsmsSpectrum` objects (for
#'   `pythonLibrary = "spectrum_utils"`). Parameter `mapping` allows to specify
#'   which spectra variables from the `Spectra` object `x` should be converted
#'   in addition to the peaks data (m/z and intensity values). It defaults to
#'   `mapping = spectraVariableMapping()` (See the respective help below for
#'   more information on the variable mapping). While being fast, this function
#'   first loads all peaks and spectra data into memory before translating to
#'   Python data structures. A less memory intense operation could be to call
#'   this function in a loop to only load parts of the data at a time into
#'   memory.
#'
#' - `pyspec_to_rspec()` translates a single, or a list of `matchms.Spectrum`
#'   objects (with parameter `pythonLibrary = "matchms"`, the default) or a
#'   list of `spectrum_utils.spectrum.MsmsSpectrum` objects (with parameter
#'   `pythonLibrary = "spectrum_utils"`) to a [Spectra::Spectra()] object.
#'   Parameter `mapping` allows to specify the metadata variables that
#'   should be translated and mapped in addition to the peaks data. The
#'   library used to represent the MS data in Python needs to be specified with
#'   parameter `pythonLibrary`.
#'
#' - `r_to_py.Spectra()` is equivalent to
#'   `rspec_to_pyspec(pythonLibrary = "matchms")`. The spectra
#'   variables that should be converted can be configures with
#'   `setSpectraVariableMapping()` (see documentation below).
#'
#'
#' @section Mapping of spectra variables (metadata):
#'
#' Metadata for MS spectra are represented and stored as *spectra variables*
#' in the R [Spectra::Spectra()] objects. Also Python MS data structures
#' store such metadata along with the mass peak data. While spectra metadata
#' is thus supported by data structures in both programming languages,
#' different names and naming conventions are used. The
#' `spectraVariableMapping()` and `setSpectraVariableMapping()` functions allow
#' to define how the names of spectra metadata (spectra variables) should be
#' translated between R and Python. To support also the different naming
#' conventions used by the Python libraries *matchms* and *spectrum_utils*,
#' `spectraVariableMapping()` defines different mapping schemes for these,
#' using by default the mapping for *matchms*. Note also that *spectrum_utils*
#' supports only few selected metadata/spectra variables, so any additional
#' spectra variables defined by the mapping will be ignored.
#' The `r_to_py()` and `py_to_r()` functions will use the selected naming
#' scheme to name the spectra variables accordingly. Also, only
#' spectra metadata/variables in `spectraVariableMapping()` will be translated.
#' The initial mapping is based on this
#' [definition in matchms](https://github.com/matchms/matchms/blob/master/matchms/data/known_key_conversions.csv).
#'
#' - `defaultSpectraVariableMapping()`: returns the *default* mapping between
#'   spectra variables and Python metadata names for the *matchms* library.
#'
#' - `spectraVariableMapping()`: returns the currently defined spectra variable
#'   mapping as a named character vector, with names
#'   representing the names of the spectra variables in R and elements the
#'   respective names of the spectra metadata in Python.
#'   Use [Spectra::spectraVariables()] on the `Spectra` object that should
#'   be converted with `r_to_py()` to list all available spectra variables.
#'   `r_to_py()` and `py_to_r()` for MS data structures will use this default
#'   mapping.
#'   Calling `spectraVariableMapping()` defining also the Python library (e.g.,
#'   `spectraVariableMapping("matchms")` or
#'   `spectraVariableMapping("spectrum_utils")`) will return the variable
#'   mapping for the specified Python library. Optional parameter `x` allows
#'   to specify a (potentially names) character vector with the names of the
#'   spectra variables that should in addition be included in the mapping.
#'
#' - `setSpectraVariableMapping()`: sets/replaces the currently defined mapping
#'   of spectra variable names to Python metadata names. Setting
#'   `setSpectraVariableMapping(character())` will only convert the mass peaks
#'   data (m/z and intensity values) but no spectra metadata.
#'
#' @param mapping named `character()` vector defining which spectra
#'     variables/metadata should be translated between R and Python and how
#'     they should be renamed. Defaults to `spectraVariableMapping()`.
#'
#' @param x For `rspec_to_pyspec` or `r_to_py.Spectra()`: the
#'     [Spectra::Spectra()]` object that should be translated. For
#'     `pyspec_to_rspec()`: a single `matchms.Spectrum` object or a Python
#'     list of `matchms.Spectrum` objects.
#'     For `spectraVariableMapping()`: optional additional spectra variables
#'     that should be appended to the pre-defined ones for the specified
#'     Python library.
#'
#' @param object For `spectraVariableMapping()`: not used.
#'
#' @param pythonLibrary  For `rspec_to_pyspec()` and `pyspec_to_rspec()`:
#'     `character(1)` defining the Python library to which (or from which)
#'     data structures the data should be converted.
#'     Possible options are `"matchms"` or `"spectrum_utils"` with `"matchms"`
#'     being the default.
#'
#' @param ... For `spectraVariableMapping()`: not used.
#'
#' @return For `r_to_py.Spectra()` and `rspec_to_pyspec()`: Python list of
#'     MS data structures, either `matchms.Spectrum` or
#'     `spectrum_utils.spectrum.MsmsSpectrum` objects. For `pyspec_to_rspec()`:
#'     [Spectra::Spectra()] with the MS data of all `matchms.Spectrum` objects
#'     in the submitted `list`.
#'
#' @author Michael Witting, Johannes Rainer, Wout Bittremieux, Thomas Naake
#'
#' @importFrom reticulate r_to_py py_to_r
#'
#' @importMethodsFrom Spectra spectrapply
#'
#' @examples
#'
#' ## Import a MGF file as a `Spectra` object
#' library(MsBackendMgf)
#' library(SpectriPy)
#' s <- Spectra(
#'     system.file("extdata", "mgf", "spectra2.mgf", package = "SpectriPy"),
#'     source = MsBackendMgf())
#' s
#'
#' #########################
#' ## Conversion R to Python
#'
#' ## A `Spectra` can be translated to a `list` of `matchms.Spectrum` objects
#' ## using either the `r_to_py()` method or the `rspec_to_pyspec()` function:
#' s_py <- r_to_py(s)
#' s_py
#'
#' ## The `s_py` can now be used like any other Python variable within the R
#' ## *reticulate* framework. Below we extract the m/z values of the first
#' ## spectrum
#' s_py[0]$mz
#'
#' ## Extracting that information from the `Spectra` object in R
#' s[1]$mz
#'
#' ## The `spectraVariableMapping()` defines which spectra variables (metadata)
#' ## should be translated between R and Python:
#' spectraVariableMapping()
#'
#' ## The names of that character vector represent the names of the spectra
#' ## variables in R, the elements the name of the metadata variable in Python.
#' ## Below we list the available metadata information from the first
#' ## Spectrum in Python
#' s_py[0]$metadata
#'
#' ## `setSpectraVariableMapping()` allows to replace the default mapping
#' ## of variables. Below we e.g. add a new spectra variable to the `Spectra`
#' ## object.
#' s$new_col <- 1:4
#'
#' ## To translate that variable to Python we need to include it to the
#' ## `spectraVariableMapping()`. Below we define to translate only the
#' ## precursor m/z and the new spectra variable to Python. Be aware that
#' ## `setSpectraVariableMapping()` **globally** sets the default for any
#' ## spectra variable mapping between R and Python. Thus, any subsequent
#' ## calls mapping calls will use the same mapping. It is suggested to
#' ## eventually *restore* the default mapping again after the call or
#' ## use the `rspec_to_pyspec()` function instead, that allows to configure
#' ## the mapping using a parameter `mapping`.
#' setSpectraVariableMapping(
#'     c(precursorMz = "precursor_mz", new_col = "new_col"))
#' s_py <- r_to_py(s)
#'
#' s_py[0]$metadata
#'
#' ## Restoring the global spectra variable mapping configuration to
#' ## the default mapping:
#' setSpectraVariableMapping(defaultSpectraVariableMapping())
#'
#' ## As an alternative to the `r_to_py()` we can use the `rspec_to_pyspec()`
#' ## function and provide a custom mapping using the `mapping` parameter:
#' s_py <- rspec_to_pyspec(
#'     s, mapping = c(precursorMz = "precursor_mz", new_col = "new_col"))
#'
#' ## Convert to MS data objects from the spectrum_utils Python library
#' s_py2 <- rspec_to_pyspec(
#'     s, mapping = spectraVariableMapping("spectrum_utils"),
#'     pythonLibrary = "spectrum_utils")
#'
#' ## Convert the data back to R
#' pyspec_to_rspec(s_py2, pythonLibrary = "spectrum_utils")
#'
#' #########################
#' ## Conversion Python to R
#'
#' ## A `list` of `matchms.Spectrum` objects in Python can be translated into
#' ## the corresponding MS data structure in R (i.e. a `Spectra`) object using
#' ## the `pyspec_to_rspec()` function:
#' res <- pyspec_to_rspec(s_py)
#' res
#'
#' ## All spectra from Python are thus converted into a single `Spectra` object.
#'
#' ## Or providing a custom variable mapping:
#' res <- pyspec_to_rspec(
#'     s_py, mapping = c(precursorMz = "precursor_mz", new_col = "new_col"))
#' res$new_col
NULL

.SPECTRA_2_MATCHMS <- c(
    precursorMz = "precursor_mz",
    precursorIntensity = "precursor_intensity",
    precursorCharge = "charge",
    rtime = "retention_time",
    collisionEnergy = "collision_energy",
    isolationWindowTargetMz = "isolation_window_target_mz",
    ## polarity = "ionmode", # Disabling since matchms does not support int.
    msLevel = "ms_level"
)

.SPECTRA_2_SPECTRUM_UTILS <- c(
    precursorMz = "precursor_mz",
    precursorCharge = "precursor_charge",
    rtime = "retention_time",
    scanIndex = "identifier"
)


#' @importMethodsFrom Spectra spectraVariableMapping
#'
#' @exportMethod spectraVariableMapping
#'
#' @rdname conversion
#'
#' @exportMethod spectraVariableMapping
setMethod("spectraVariableMapping", "character", function(object,
                                                          x = character(),
                                                          ...) {
    object <- object[1L]
    if (is.null(names(x))) names(x) <- x
    if (!object %in% c("matchms", "spectrum_utils"))
        stop("Supported values are \"matchms\" or \"spectrum_utils\"")
    if (object == "matchms") {
        c(.SPECTRA_2_MATCHMS, x[!x %in% names(.SPECTRA_2_MATCHMS)])
    } else {
        if (length(x))
            warning("Provided spectra variables ",
                    paste0(names(x), collapse = ", "),
                    " will not be translated to Python")
        .SPECTRA_2_SPECTRUM_UTILS
    }
})

#' @rdname conversion
setMethod("spectraVariableMapping", "missing", function(object, ...) {
    getOption("spectripy.spectra_variable_mapping", .SPECTRA_2_MATCHMS)
})

#' @rdname conversion
#'
#' @export
setSpectraVariableMapping <- function(x) {
    if (!is.character(x) | length(names(x)) != length(x))
        stop("'x' is expected to be a named character vector")
    options(spectripy.spectra_variable_mapping = x)
}

#' @rdname conversion
#'
#' @export
defaultSpectraVariableMapping <- function() {
    .SPECTRA_2_MATCHMS
}

## -------- R TO PY ----------------------------------------------------------##

#' @rdname conversion
#'
#' @description
#'
#' Function to convert R Spectra objects into a Python list of matchms Spectrum
#' objects using the `reticulate` package.
#'
#' @param x `Spectra` object.
#'
#' @param convert Boolean; should Python objects be automatically converted to
#' their R equivalent? Defaults to `FALSE`.
#'
#' @importFrom reticulate r_to_py
#'
#' @importFrom Spectra spectrapply
#'
#' @export
r_to_py.Spectra <- function(x, convert = FALSE) {
    .rspec_to_matchms_pyspec(x, mapping = spectraVariableMapping())
}

#' @rdname conversion
#'
#' @importFrom methods is
#'
#' @export
rspec_to_pyspec <- function(x, mapping = spectraVariableMapping(),
                            pythonLibrary = c("matchms", "spectrum_utils")) {
    if (!is(x, "Spectra"))
        stop("'x' should be a Spectra object.")
    pythonLibrary <- match.arg(pythonLibrary)
    ## that could be more memory efficient, but slower.
    ## r_to_py(spectrapply(x, .single_rspec_to_pyspec,
    ##                     spectraVariables = mapping,
    ##                     BPPARAM = BPPARAM))
    switch(pythonLibrary,
           matchms = .rspec_to_matchms_pyspec(x, mapping = mapping),
           spectrum_utils = .rspec_to_spectrum_utils_pyspec(
               x, mapping = mapping)
    )
}

#' @description
#'
#' Function to convert a **single** R Spectra object (of length 1) into a
#' Python matchms Spectrum using the `reticulate` package.
#'
#' @param x `Spectra` object **of length 1!**.
#'
#' @param spectraVariables Named `character` vector defining the spectra
#'     variables that should be stored as metadata in `matchms`' metadata. Names
#'     are expected to be the spectra variable names and values the
#'     corresponding metadata fields in `matchms`. Defaults to
#'     `.SPECTRA_2_MATCHMS`. If `spectraVariables = character()` no
#'     metadata will be stored.
#'
#' @return `Spectrum` Single Python Spectrum.
#'
#' @author Michael Witting, Johannes Rainer, Wout Bittremieux
#'
#' @importMethodsFrom Spectra spectraData
#'
#' @importMethodsFrom Spectra peaksData
#'
#' @importMethodsFrom Spectra mz
#'
#' @importMethodsFrom Spectra intensity
#'
#' @importFrom reticulate np_array r_to_py
#'
#' @noRd
.single_rspec_to_pyspec <-
    function(x, mapping = spectraVariableMapping()) {
        pks <- unname(peaksData(x, c("mz", "intensity")))[[1L]]
        if (length(mapping)) {
            slist <- as.list(spectraData(x, columns = names(mapping)))
            names(slist) <- mapping[names(slist)]
            matchms$Spectrum(
                        mz = np_array(pks[, 1L]),
                        intensities = np_array(pks[, 2L]),
                        metadata = r_to_py(slist)
                    )
        } else {
            matchms$Spectrum(
                        mz = np_array(pks[, 1L]),
                        intensities = np_array(pks[, 2L])
                    )
        }
    }

#' Converts a `Spectra::Spectra` (or its `spectraData()`) to a `list` of
#' `matchms.Spectrum` by first extracting the peaks and spectra data and
#' iterating over these. This is a faster, but also more memory heavy
#' implementation as the full peaks and spectra data are read into memory.
#' With the `single_rspec_to_pyspec()` only the data from one spectrum at
#' a time are read.
#'
#' @param x either a `Spectra` object or a `DataFrame` which includes also the
#'    `"mz"` and `"intensity"` values (e.g. by calling `spectraData(x@backend))`
#'
#' @noRd
.rspec_to_matchms_pyspec <- function(x, mapping = spectraVariableMapping()) {
    if (is(x, "Spectra"))
        pks <- peaksData(x, c("mz", "intensity"), return.type = "list")
    else pks <- mapply(mz = x$mz, intensity = x$intensity, FUN = cbind)
    sv <- mapping[!mapping %in% c("mz", "intensity")]
    if (length(sv)) {
        if (is(x, "Spectra"))
            spd <- as.data.frame(spectraData(x, columns = names(sv)))
        else spd <- as.data.frame(x[, names(sv), drop = FALSE])
        colnames(spd) <- sv[colnames(spd)]
        rm(x)
        spd <- split(spd, seq_along(pks))
        r_to_py(mapply(function(y, z) {
            matchms$Spectrum(mz = np_array(z[, 1L]),
                             intensities = np_array(z[, 2L]),
                             metadata = r_to_py(as.list(y)))
        }, spd, pks, SIMPLIFY = FALSE, USE.NAMES = FALSE))
    } else {
        rm(x)
        r_to_py(lapply(pks, function(z)
            matchms$Spectrum(mz = np_array(z[, 1L],),
                             intensities = np_array(z[, 2L]))))
    }
}

#' spectrum_utils have fixed metadata variables. These are:
#'
#' - identifier `character`
#' - mz
#' - intensity
#' - precursor_mz `numeric`
#' - precursor_charge `integer`
#' - retention_time `numeric`
#'
#' We thus need to either extract these from the `spectraData()` or provide
#' them as NA. Also, we drop any additional variables.
#'
#' @noRd
.rspec_to_spectrum_utils_pyspec <-
    function(x, mapping = spectraVariableMapping()) {
        if (is(x, "Spectra"))
            pks <- peaksData(x, c("mz", "intensity"))
        else pks <- mapply(mz = x$mz, intensity = x$intensity, FUN = cbind)
        l <- length(pks)
        sv <- mapping[!mapping %in% c("mz", "intensity")]
        svm <- sv[sv %in% .SPECTRA_2_SPECTRUM_UTILS]
        if (length(w <- setdiff(
                       sv, .SPECTRA_2_SPECTRUM_UTILS)))
            warning("Ignoring variables ",
                    paste0("\"", names(sv)[sv %in% w] ,"\""))
        if (length(svm)) {
            if (is(x, "Spectra"))
                spd <- as.data.frame(spectraData(x, columns = names(svm)))
            else spd <- as.data.frame(x[, names(svm), drop = FALSE])
            colnames(spd) <- svm[colnames(spd)]
        }
        rm(x)
        if (any(svm == "identifier")) {
            id <- as.character(spd$identifier)
        } else id <- rep(NA_character_, l)
        if (any(svm == "precursor_mz")) {
            pmz <- spd$precursor_mz
        } else pmz <- rep(NA_real_, l)
        if (any(svm == "precursor_charge")) {
            pch <- spd$precursor_charge
        } else pch <- rep(NA_integer_, l)
        if (any(svm == "retention_time")) {
            rt <- spd$retention_time
        } else rt <- rep(NA_real_, l)
        if (length(svm)) rm(spd)
        r_to_py(mapply(function(z, id, pmz, pch, rt) {
            spectrum_utils$spectrum$MsmsSpectrum(
                                        mz = np_array(z[, 1L]),
                                        intensity = np_array(z[, 2L]),
                                        identifier = r_to_py(id),
                                        precursor_mz = r_to_py(pmz),
                                        precursor_charge = r_to_py(pch),
                                        retention_time = r_to_py(rt))
        }, pks, id, pmz, pch, rt, SIMPLIFY = FALSE, USE.NAMES = FALSE))
    }

## -------- PY TO R ----------------------------------------------------------##

#' Function to extract the metadata from a (**single**) `matchms.Spectrum`
#' object. This function can be used in a loop over a `list` of objects to
#' create a `Spectra` from it.
#'
#' @param x `matchms.Spectrum`.
#'
#' @param mapping the mapping of spectra variables to metadata names.
#'
#' @return `data.frame()` with the spectra data.
#'
#' @noRd
.py_matchms_spectrum_spectra_data <-
    function(x, mapping = spectraVariableMapping(), ...) {
        pl <- x$metadata
        map <- mapping[mapping %in% names(pl)]
        if (length(map)) {
            ## would be nice to be able to call list(pl.items())
            res <- lapply(map, function(z) py_to_r(pl[z]))
            base::as.data.frame(res[lengths(res) > 0])
        } else data.frame(msLevel = NA_integer_)
    }

#' Function to extract the peaks data as a two-column `matrix` from a
#' (**single!**) `matchms.Spectrum` object. This function can be used in a
#' loop over a `list` of objects to create a `Spectra` from it.
#'
#' @param x `matchms.Spectrum` object.
#'
#' @return `numeric` `matrix` with two columns `"mz"` and `"intensity"`.
#'
#' @noRd
.py_matchms_spectrum_peaks_data <- function(x, ...) {
    m <- py_to_r(x$peaks$to_numpy)
    colnames(m) <- c("mz", "intensity")
    m
}

.py_matchms_spectrum_peaks_data_columns <-
    function(x, columns = c("mz", "intensity"), drop = FALSE, ...) {
        m <- py_to_r(x$peaks$to_numpy)
        colnames(m) <- c("mz", "intensity")
        m[, columns, drop = drop]
    }

.py_spectrum_utils_spectrum_spectra_data <-
    function(x, mapping = spectraVariableMapping(), ...) {
        s <- data.frame(precursor_mz = py_to_r(x$precursor_mz),
                        precursor_charge = py_to_r(x$precursor_charge),
                        retention_time = py_to_r(x$retention_time),
                        identifier = py_to_r(x$identifier))
        s <- s[, colnames(s) %in% mapping, drop = FALSE]
        if (length(s)) {
            colnames(s) <- names(mapping)[match(colnames(s), mapping)]
            s
        } else data.frame(msLevel = NA_integer_)
    }

.py_spectrum_utils_spectrum_peaks_data <- function(x, ...) {
    cbind(mz = py_to_r(x$mz), intensity = py_to_r(x$intensity))
}

#' @description
#'
#' Function to convert a single Python Spectrum object into an R Spectra using
#' the `reticulate` package.
#'
#' @param x `Spectrum` Single Python Spectrum.
#'
#' @param mapping named `character` vector with the names of the
#'     spectra variables that should be extracted. Names are expected to be
#'     the spectra variable names (in `Spectra`) and values the corresponding
#'     names of the variables within the Python Spectrum.
#'
#' @param reference Reference to Python environment matchms
#'
#' @return `Spectra` single R Spectra
#'
#' @author Michael Witting, Johannes Rainer
#'
#' @importFrom IRanges NumericList
#'
#' @importFrom S4Vectors DataFrame
#'
#' @importMethodsFrom Spectra Spectra
#'
#' @importFrom Spectra MsBackendMemory
#'
#' @noRd
.single_pyspec_to_rspec <-
    function(x, mapping = spectraVariableMapping(), ...) {
        be <- MsBackendMemory()
        be@spectraData <- .py_matchms_spectrum_spectra_data(
            x, mapping = mapping)
        be@spectraData$dataStorage <- "<memory>"
        be@peaksData <- list(.py_matchms_spectrum_peaks_data(x))
        Spectra(be)
    }

## Note: disabling this now. Generally, it would be a nice thing to have,
## but it seems R/reticulate starts automatically converting whenever a
## variable from the py is accessed, either throug `py$var` or even with
## `py_get_attr(py, "var")` - the `iterate()` call below is then converting
## the data to a `Spectra` and passing that along to the R function.
## #' @importFrom reticulate py_to_r
## #' @export
## py_to_r.matchms.Spectrum.Spectrum <- function(x) {
##     .single_pyspec_to_rspec(x)
## }

#' @export
#'
#' @importFrom MsCoreUtils rbindFill
#'
#' @importFrom reticulate iterate
#'
#' @importFrom methods slot<-
#'
#' @rdname conversion
pyspec_to_rspec <- function(x, mapping = spectraVariableMapping(),
                            pythonLibrary = c("matchms", "spectrum_utils")) {
    pythonLibrary <- match.arg(pythonLibrary)
    if (is(x, "matchms.Spectrum.Spectrum"))
        return(.single_pyspec_to_rspec(x, mapping = mapping))
    be <- MsBackendMemory()
    orig_mapping <- spectraVariableMapping()
    setSpectraVariableMapping(mapping)
    if (is.list(x))
        ITER <- lapply
    else ITER <- iterate
    switch(pythonLibrary,
           matchms = {
               PFUN <- .py_matchms_spectrum_peaks_data
               SFUN <- .py_matchms_spectrum_spectra_data
           },
           spectrum_utils = {
               PFUN <- .py_spectrum_utils_spectrum_peaks_data
               SFUN <- .py_spectrum_utils_spectrum_spectra_data
           })
    slot(be, "peaksData", check = FALSE) <- ITER(x, PFUN, simplify = FALSE)
    ## Not very efficient and elegant... get the indivudal elements and stuff
    ## into list. pandas.DataFrame can not be easily created unfortunately.
    sdta <- do.call(rbindFill, ITER(x, SFUN, simplify = FALSE))
    sdta$dataStorage <- "<memory>"
    ## Ensure correct data type for core variables.
    csv <- coreSpectraVariables()
    for (mtc in intersect(colnames(sdta), names(csv)))
        sdta[[mtc]] <- as(sdta[[mtc]], csv[mtc])
    slot(be, "spectraData", check = FALSE) <- sdta
    setSpectraVariableMapping(orig_mapping)
    Spectra(be)
}

#' Function to get peaks data matrices from a list of `matchms.Spectrum`
#' objects.
#'
#' @param x `character(1)` with the name of the variable (in Python!) that
#'     contains the MS data.
#'
#' @param i `integer` with the indices of the spectra from which the peaks
#'     data should be extracted.
#'
#' @return `list` of two column `matrix` **without** column names.
#'
#' @importFrom reticulate py_set_attr py_del_attr
#'
#' @noRd
.py_matchms_peaks_data <- function(x, i) {
    if (length(i) > 1)
        py_set_attr(py, "_i_", as.integer(i))
    else py_set_attr(py, "_i_", list(as.integer(i)))
    res <- py_to_r(py_run_string(.py_matchms_peaks_data_cmd(x),
                                 local = TRUE, convert = FALSE))[["_res_"]]
    py_del_attr(py, "_i_")
    res
}

.py_matchms_peaks_data_cmd <- function(x) {
    paste0("_res_ = list()\n",
           "for i in _i_:\n",
           "  _res_.append(", x, "[i].peaks.to_numpy)\n")
}

#' Function to extract peaks data from a list of
#' `spectrum_utils.spectrum.MsmsSpectrum` objects. This function can be called
#' from `pyspec_to_rspec` or from the `MsBackendPy`.
#'
#' @noRd
.py_spectrum_utils_peaks_data <- function(x, i) {
    if (length(i) > 1)
        py_set_attr(py, "_i_", as.integer(i))
    else py_set_attr(py, "_i_", list(as.integer(i)))
    res <- py_to_r(py_run_string(.py_spectrum_utils_peaks_data_cmd(x),
                                 local = TRUE, convert = FALSE))[["_res_"]]
    py_del_attr(py, "_i_")
    res
}

.py_spectrum_utils_peaks_data_cmd <- function(x) {
    paste0("import numpy as np\n",
           "_res_ = list()\n",
           "for i in _i_:\n",
           "  _res_.append(np.column_stack((",x,"[i].mz,",x,"[i].intensity)))")
}

SPECTRUM_UTILS_TOLERANCE <- 1e-7
df <- DataFrame(msLevel = c(1L, 2L, 1L), rtime = c(1.2, 3.2, 5.2))
df$mz <- list(
    c(34.3, 45.6, 199.1),
    c(45.6, 89.3, 100.1, 200.5),
    c(56.1, 56.2)
)
df$intensity <- list(
    1:3,
    1:4,
    1:2
)
sps <- Spectra(df)

test_that("spectraVariableMapping, spectraVariableMapping<- works", {
    res <- spectraVariableMapping()
    expect_true(is.character(res))
    expect_true(length(res) > 0)
    expect_true(length(names(res)) > 0)
    expect_equal(res, .SPECTRA_2_MATCHMS)

    setSpectraVariableMapping(c(a = "b", d = "e"))
    expect_equal(spectraVariableMapping(), c(a = "b", d = "e"))
    setSpectraVariableMapping(.SPECTRA_2_MATCHMS)
    expect_equal(spectraVariableMapping(),
                 .SPECTRA_2_MATCHMS)
    expect_equal(defaultSpectraVariableMapping(), .SPECTRA_2_MATCHMS)

    res <- spectraVariableMapping("spectrum_utils")
    expect_equal(res, .SPECTRA_2_SPECTRUM_UTILS)

    expect_error(spectraVariableMapping("other"), "Supported values")

    res <- spectraVariableMapping("matchms")
    expect_equal(res, .SPECTRA_2_MATCHMS)

    expect_error(setSpectraVariableMapping(3), "named character vector")
    expect_error(setSpectraVariableMapping(c("a", "b")), "named character")

    res <- spectraVariableMapping("matchms", c("some", "other"))
    expect_equal(res, c(spectraVariableMapping("matchms"),
                        c(some = "some", other = "other")))

    expect_warning(
        res <- spectraVariableMapping("spectrum_utils", c("some", "other")),
        "Provided spectra variables")
    expect_equal(res, spectraVariableMapping("spectrum_utils"))
})

#############
##    R to Py

test_that(".single_rspec_to_pyspec works", {
    res <- .single_rspec_to_pyspec(sps[1L])
    expect_true(is(res, "matchms.Spectrum.Spectrum"))
    expect_true(all(.SPECTRA_2_MATCHMS %in% names(res$metadata)))
    expect_equal(sps$mz[[1L]], as.vector(py_to_r(res$mz)))
    expect_equal(sps$intensity[[1L]], as.vector(py_to_r(res$intensities)))
    expect_equal(sps$rtime[1L], py_to_r(res$metadata["retention_time"]))
    expect_equal(sps$msLevel[1L], py_to_r(res$metadata["ms_level"]))
    expect_equal(sps$precursorMz[1L], py_to_r(res$metadata["precursor_mz"]))

    res <- .single_rspec_to_pyspec(sps[2L], character())
    expect_equal(sps$mz[[2L]], as.vector(py_to_r(res$mz)))
    expect_equal(sps$intensity[[2L]], as.vector(py_to_r(res$intensities)))
    expect_equal(names(res$metadata), character())
})

test_that(".rspec_to_matchms_pyspec works", {
    res <- .rspec_to_matchms_pyspec(sps)
    expect_true(is(res, "python.builtin.list"))
    expect_true(length(res) == length(sps))

    expect_equal(sps$mz[[2L]], as.vector(py_to_r(res[[1]]$mz)))
    expect_equal(sps$intensity[[2L]], as.vector(py_to_r(res[[1]]$intensities)))
    expect_true(all(.SPECTRA_2_MATCHMS %in% names(res[[1]]$metadata)))

    res <- .rspec_to_matchms_pyspec(sps, character())
    expect_equal(sps$mz[[2L]], as.vector(py_to_r(res[[1]]$mz)))
    expect_equal(sps$intensity[[2L]], as.vector(py_to_r(res[[1]]$intensities)))
    expect_equal(names(res[[1]]$metadata), character())

    ## passing a `DataFrame` or `data.frame` instead
    x <- spectraData(sps@backend)
    res <- .rspec_to_matchms_pyspec(x)
    expect_true(is(res, "python.builtin.list"))
    expect_true(length(res) == length(sps))
    expect_equal(sps$mz[[2L]], as.vector(py_to_r(res[[1]]$mz)))
    expect_equal(sps$intensity[[2L]], as.vector(py_to_r(res[[1]]$intensities)))
    expect_true(all(.SPECTRA_2_MATCHMS %in% names(res[[1]]$metadata)))

    res <- .rspec_to_matchms_pyspec(x, character())
    expect_equal(sps$mz[[2L]], as.vector(py_to_r(res[[1]]$mz)))
    expect_equal(sps$intensity[[2L]], as.vector(py_to_r(res[[1]]$intensities)))
    expect_equal(names(res[[1]]$metadata), character())
})

test_that(".rspec_to_spectrum_utils_pyspec works", {
    res <- expect_warning(.rspec_to_spectrum_utils_pyspec(sps),
                          "Ignoring variables")
    expect_equal(length(res), length(sps))
    expect_true(is(res, "python.builtin.list"))
    expect_true(is(res[[1L]], "spectrum_utils.spectrum.MsmsSpectrum"))
    expect_true(is.na(py_to_r(res[[1L]]$precursor_mz)))
    expect_equal(py_to_r(res[[1L]]$retention_time), sps$rtime[2L])

    res <- .rspec_to_spectrum_utils_pyspec(sps, c())
    expect_equal(length(res), length(sps))
    expect_true(is(res, "python.builtin.list"))
    expect_true(is(res[[1L]], "spectrum_utils.spectrum.MsmsSpectrum"))
    expect_true(is.na(py_to_r(res[[1L]]$precursor_mz)))
    expect_true(is.na(py_to_r(res[[1L]]$retention_time)))

    sps$my_id <- c("a", "b", "c")
    res <- .rspec_to_spectrum_utils_pyspec(sps, c(my_id = "identifier"))
    expect_equal(length(res), length(sps))
    expect_true(is(res, "python.builtin.list"))
    expect_true(is(res[[1L]], "spectrum_utils.spectrum.MsmsSpectrum"))
    expect_true(is.na(py_to_r(res[[1L]]$precursor_mz)))
    expect_true(is.na(py_to_r(res[[1L]]$retention_time)))
    expect_equal(py_to_r(res[[1L]]$identifier), "b")

    res <- .rspec_to_spectrum_utils_pyspec(sps, .SPECTRA_2_SPECTRUM_UTILS)
    expect_equal(length(res), length(sps))
    expect_true(is(res, "python.builtin.list"))
    expect_true(is(res[[1L]], "spectrum_utils.spectrum.MsmsSpectrum"))
    expect_true(is.na(py_to_r(res[[1L]]$precursor_mz)))

    ## passing a `DataFrame` or `data.frame` instead
    x <- spectraData(sps@backend)
    res <- .rspec_to_spectrum_utils_pyspec(x)
    expect_true(is(res, "python.builtin.list"))
    expect_true(length(res) == length(sps))
    expect_equal(sps$mz[[2L]], as.vector(py_to_r(res[[1]]$mz)),
                 tolerance = SPECTRUM_UTILS_TOLERANCE)
    expect_equal(sps$intensity[[2L]], as.vector(py_to_r(res[[1]]$intensity)))
    expect_equal(sps$rtime[2], py_to_r(res[[1]]$retention_time))
    expect_equal(sps$precursorMz[2], py_to_r(res[[1]]$precursor_mz))

    res <- .rspec_to_spectrum_utils_pyspec(x, character())
    expect_equal(sps$mz[[2L]], as.vector(py_to_r(res[[1]]$mz)),
                 tolerance = SPECTRUM_UTILS_TOLERANCE)
    expect_equal(sps$intensity[[2L]], as.vector(py_to_r(res[[1]]$intensity)))
    expect_true(is.na(py_to_r(res[[1]]$retention_time)))
    expect_true(is.na(py_to_r(res[[1]]$precursor_mz)))
})

test_that("rspec_to_pyspec works", {
    expect_error(rspec_to_pyspec(4), "Spectra object")

    res <- r_to_py(sps)
    expect_equal(res, rspec_to_pyspec(sps))
    expect_true(is(res, "python.builtin.list"))
    expect_equal(length(res), length(sps))
    expect_equal(as.numeric(py_to_r(res[2]$peaks$mz)), mz(sps)[[3]])
    expect_equal(rtime(sps)[1], py_to_r(res[0]$metadata$retention_time))
    expect_equal(msLevel(sps)[1], py_to_r(res[0]$metadata$ms_level))
    expect_equal(msLevel(sps)[2], py_to_r(res[1]$metadata$ms_level))
    expect_equal(msLevel(sps)[3], py_to_r(res[2]$metadata$ms_level))
    expect_equal(precursorMz(sps)[1], py_to_r(res[0]$metadata$precursor_mz))

    ## mapping only specific spectra variables.
    sps$new_col <- c(9, 3, 1)
    setSpectraVariableMapping(c(collisionEnergy = "collision_energy",
                                new_col = "new_col"))
    res <- r_to_py(sps)
    expect_true(is(res, "python.builtin.list"))
    expect_equal(length(res), length(sps))
    expect_equal(sort(names(res[0]$metadata)),
                 sort(c("collision_energy", "new_col")))
    expect_equal(py_to_r(res[0]$metadata$new_col), 9)

    s <- sps
    s$rtime <- NULL
    setSpectraVariableMapping(c(precursorMz = "precursor_mz",
                                rtime = "retention_time"))
    res <- r_to_py(s)
    expect_equal(sort(names(res[0]$metadata)),
                 c("precursor_mz", "retention_time"))
    expect_equal(py_to_r(res[0]$metadata$precursor_mz), NA_real_)
    ## It's a bit odd - a retention time of NA is converted to a NULL
    expect_equal(py_to_r(res[0]$metadata$retention_time), NULL)

    s$new_col <- NULL
    setSpectraVariableMapping(c(precursorMz = "precursor_mz",
                                new_col = "new_col"))
    expect_error(res <- r_to_py(s), "requested spectra variables")

    ## No mapping at all.
    setSpectraVariableMapping(character())
    res <- rspec_to_pyspec(s)
    expect_equal(names(res[0]$metadata), character())

    setSpectraVariableMapping(.SPECTRA_2_MATCHMS)

    ## spectrum_utils
    expect_error(rspec_to_pyspec(sps, pythonLibrary = "other"))
    res <- rspec_to_pyspec(sps, spectraVariableMapping("spectrum_utils"),
                           "spectrum_utils")
    expect_equal(length(res), length(sps))
    expect_true(is(res, "python.builtin.list"))
    expect_true(is(res[[1L]], "spectrum_utils.spectrum.MsmsSpectrum"))
})

#############
##    Py to R

test_that(".py_matchms_spectrum_spectra_data works", {
    p <- r_to_py(sps)
    res <- .py_matchms_spectrum_spectra_data(p[[1]])
    expect_true(is.data.frame(res))
    expect_true(all(names(.SPECTRA_2_MATCHMS) %in% colnames(res)))
    expect_equal(res$msLevel, sps$msLevel[2])
    expect_equal(res$rtime, sps$rtime[2])
    expect_equal(res$precursorCharge, sps$precursorCharge[2])
    expect_equal(res$precursorMz, sps$precursorMz[2],
                 tolerance = SPECTRUM_UTILS_TOLERANCE)
    expect_equal(res$precursorIntensity, sps$precursorIntensity[2])
    expect_equal(res$collisionEnergy, sps$collisionEnergy[2])

    res <- .py_matchms_spectrum_spectra_data(
        p[[1]], mapping = c(msLevel = "ms_level", other_col = "other_col"))
    expect_equal(colnames(res), "msLevel")
    expect_equal(res$msLevel, sps$msLevel[2])

    res <- .py_matchms_spectrum_spectra_data(p[[1]], mapping = character())
    expect_equal(colnames(res), "msLevel")
    expect_equal(res$msLevel, NA_integer_)
})

test_that(".py_matchms_spectrum_peaks_data works", {
    p <- r_to_py(sps)
    res <- .py_matchms_spectrum_peaks_data(p[[1]])
    expect_true(is.matrix(res))
    expect_equal(colnames(res), c("mz", "intensity"))
    expect_equal(res, peaksData(sps)[[2L]])
})

test_that(".py_matchms_spectrum_peaks_data_columns works", {
    p <- r_to_py(sps)
    res <- .py_matchms_spectrum_peaks_data_columns(p[[1]])
    expect_true(is.matrix(res))
    expect_equal(colnames(res), c("mz", "intensity"))
    expect_equal(res, peaksData(sps)[[2L]])
    res <- .py_matchms_spectrum_peaks_data_columns(
                           p[[1]], columns = c("intensity", "mz"))
    expect_true(is.matrix(res))
    expect_equal(colnames(res), c("intensity", "mz"))
    res <- .py_matchms_spectrum_peaks_data_columns(
                           p[[1]], columns = c("intensity"))
    expect_true(is.matrix(res))
    expect_equal(colnames(res), c("intensity"))

    res <- .py_matchms_spectrum_peaks_data_columns(
                           p[[1]], columns = c("intensity"), drop = TRUE)
    expect_true(is.numeric(res))
    expect_false(is.matrix(res))
})

test_that("pyspec_to_rspec and .single_pyspec_to_rspec work", {
    p <- r_to_py(sps)
    res <- pyspec_to_rspec(p[0])
    expect_s4_class(res, "Spectra")
    expect_equal(peaksData(res), peaksData(sps[1L]))
    expect_equal(res$msLevel, sps[1L]$msLevel)
    expect_equal(res$rtime, sps[1L]$rtime)

    res <- pyspec_to_rspec(p)
    expect_s4_class(res, "Spectra")
    expect_equal(peaksData(res), peaksData(sps))
    expect_equal(res$msLevel, sps$msLevel)
    expect_equal(res$rtime, sps$rtime)

    res2 <- pyspec_to_rspec(py_to_r(p))
    expect_s4_class(res2, "Spectra")
    expect_equal(peaksData(res2), peaksData(sps))
    expect_equal(res2$msLevel, sps$msLevel)
    expect_equal(res2$rtime, sps$rtime)

    res <- .single_pyspec_to_rspec(
        p[1L], mapping = c(rtime = "retention_time"))
    expect_equal(res$rtime, sps[2L]$rtime)
    expect_true(is.na(res$msLevel))

    res <- .single_pyspec_to_rspec(p[1L], mapping = character())
    expect_true(is.na(res$rtime))
    expect_true(is.na(res$msLevel))

    ## spectrum_utils
    map <- spectraVariableMapping("spectrum_utils")
    p <- rspec_to_pyspec(sps, map, "spectrum_utils")
    res <- expect_warning(pyspec_to_rspec(p, map, "spectrum_utils"), "NAs")
    expect_s4_class(res, "Spectra")
    expect_s4_class(res@backend, "MsBackendMemory")
    expect_equal(rtime(res), rtime(sps), tolerance = SPECTRUM_UTILS_TOLERANCE)
    expect_true(validObject(res))
    expect_equal(peaksData(res, return.type = "list"),
                 peaksData(sps, return.type = "list"),
                 tolerance = SPECTRUM_UTILS_TOLERANCE)
})

test_that("pyspec_to_rspec works", {
    p <- r_to_py(sps)
    res <- pyspec_to_rspec(p)
    expect_s4_class(res, "Spectra")
    expect_equal(peaksData(res), peaksData(sps))
    expect_equal(spectraData(res, names(spectraVariableMapping())),
                 spectraData(sps, names(spectraVariableMapping())))

    res_1 <- pyspec_to_rspec(p[0])
    expect_equal(res_1, res[1])

    ## custom mapping.
    sps$new_col <- seq_along(sps)
    maps <- c(rtime = "retention_time", new_col = "r_new_col")
    p <- rspec_to_pyspec(sps, mapping = maps)
    res <- pyspec_to_rspec(p, mapping = maps)
    expect_equal(peaksData(res), peaksData(sps))
    expect_equal(res$rtime, sps$rtime)
    expect_equal(res$new_col, sps$new_col)
    expect_equal(spectraVariableMapping(), .SPECTRA_2_MATCHMS)

    p <- rspec_to_pyspec(sps, character())
    res <- pyspec_to_rspec(p)
    expect_true(all(is.na(res$msLevel)))
    expect_true(all(is.na(res$rtime)))
    expect_equal(spectraVariableMapping(), .SPECTRA_2_MATCHMS)
})

test_that(".py_matchms_peaks_data works", {
    ## Python variable in R
    p <- r_to_py(sps)
    res <- .py_matchms_peaks_data("r.p", (1:3 - 1L))
    expect_true(is.list(res))
    expect_equal(length(res), length(sps))
    expect_true(is.matrix(res[[1L]]))
    res <- lapply(res, function(z) {
        colnames(z) <- c("mz", "intensity")
        z
    })
    expect_equal(res, peaksData(sps@backend))

    res <- .py_matchms_peaks_data("r.p", (c(3, 1, 3) - 1L))
    res <- lapply(res, function(z) {
        colnames(z) <- c("mz", "intensity")
        z
    })
    expect_equal(res, peaksData(sps@backend)[c(3, 1, 3)])

    ## Python variable in Python
    rm("p")
    py_set_attr(py, "p", r_to_py(sps))
    res <- .py_matchms_peaks_data("p", (1:3 - 1L))
    expect_true(is.list(res))
    expect_equal(length(res), length(sps))
    expect_true(is.matrix(res[[1L]]))
    res <- lapply(res, function(z) {
        colnames(z) <- c("mz", "intensity")
        z
    })
    expect_equal(res, peaksData(sps@backend))
})

test_that(".py_matchms_peaks_data_cmd works", {
    res <- .py_matchms_peaks_data_cmd("AAA")
    expect_match(res, "AAA")
})

## spectrum_utils

test_that(".py_spectrum_utils_peaks_data_cmd works", {
    res <- .py_spectrum_utils_peaks_data_cmd("BB")
    expect_match(res, "BB")
})

test_that(".py_spectrum_utils_peaks_data works", {
    py_set_attr(py, "p", rspec_to_pyspec(sps, pythonLibrary = "spectrum_utils"))

    res <- .py_spectrum_utils_peaks_data("p", 0:2)
    expect_true(is.list(res))
    expect_true(is.matrix(res[[1L]]))
    expect_true(ncol(res[[1L]]) == 2)
    expect_equal(res[[1L]][, 2L], 1:3)
    expect_equal(res[[1L]], unname(peaksData(sps)[[1L]]),
                 tolerance = SPECTRUM_UTILS_TOLERANCE)

    a <- .py_spectrum_utils_peaks_data("p", 0)
    expect_equal(a[[1L]], res[[1L]])
    a <- .py_spectrum_utils_peaks_data("p", 1)
    expect_equal(a[[1L]], res[[2L]])
    a <- .py_spectrum_utils_peaks_data("p", 2)
    expect_equal(a[[1L]], res[[3L]])
})

test_that(".py_spectrum_utils_spectrum_peaks_data works", {
    tmp <- rspec_to_pyspec(sps, spectraVariableMapping("spectrum_utils"),
                           "spectrum_utils")
    res <- .py_spectrum_utils_spectrum_peaks_data(tmp[0])
    expect_true(is.matrix(res))
    expect_true(ncol(res) == 2)
    expect_equal(colnames(res), c("mz", "intensity"))
    expect_equal(res, peaksData(sps)[[1L]],
                 tolerance = SPECTRUM_UTILS_TOLERANCE)
    res <- .py_spectrum_utils_spectrum_peaks_data(tmp[1])
    expect_true(is.matrix(res))
    expect_true(ncol(res) == 2)
    expect_equal(colnames(res), c("mz", "intensity"))
    expect_equal(res, peaksData(sps)[[2L]],
                 tolerance = SPECTRUM_UTILS_TOLERANCE)
    rm(tmp)
})

test_that(".py_spectrum_utils_spectrum_spectra_data works", {
    map <- spectraVariableMapping("spectrum_utils")
    tmp <- rspec_to_pyspec(sps, map, "spectrum_utils")
    res <- .py_spectrum_utils_spectrum_spectra_data(tmp[0])
    expect_true(is.data.frame(res))
    expect_true(nrow(res) == 1)
    expect_true(ncol(res) == 2)
    expect_equal(colnames(res), c("precursorMz", "rtime"))
    expect_equal(res$rtime, sps$rtime[1L], tolerance = SPECTRUM_UTILS_TOLERANCE)
    expect_equal(res$precursorMz, sps$precursorMz[1L])

    res <- .py_spectrum_utils_spectrum_spectra_data(tmp[1], map)
    expect_true(is.data.frame(res))
    expect_true(nrow(res) == 1)
    expect_true(ncol(res) == 4)
    expect_equal(colnames(res), c("precursorMz", "precursorCharge", "rtime",
                                  "scanIndex"))
    ## expect_identical(res$precursorCharge, NA_integer_)
    expect_equal(res$rtime, sps$rtime[2L], tolerance = SPECTRUM_UTILS_TOLERANCE)

    res <- .py_spectrum_utils_spectrum_spectra_data(tmp[0], mapping = c())
    expect_true(is.data.frame(res))
    expect_equal(colnames(res), "msLevel")
    expect_equal(res$msLevel, NA_integer_)
})

#' @title Spectra similarity calculations using Python's matchms library
#'
#' @name compareSpectriPy
#'
#' @description
#'
#' The `compareSpectriPy()` function allows to calculate spectral similarity
#' scores using the `calculate_scores()` function of the Python
#' [matchms.similarity](https://matchms.readthedocs.io/en/latest/api/matchms.similarity.html).
#' module.
#'
#' Selection and configuration of the algorithm can be performed with one of the
#' *parameter* objects/functions:
#'
#' - `CosineGreedy()`: calculate the *cosine similarity score* between
#'   spectra. The score is calculated by finding the best possible matches
#'   between peaks of two spectra. Two peaks are considered a potential match if
#'   their m/z ratios lie within the given `tolerance`. The underlying peak
#'   assignment problem is here solved in a *greedy* way. This can perform
#'   notably faster, but does occasionally deviate slightly from a fully correct
#'   solution (as with the `CosineHungarian` algorithm). In practice this
#'   will rarely affect similarity scores notably, in particular for smaller
#'   tolerances. The algorithm can be configured with parameters `tolerance`,
#'   `mz_power` and `intensity_power` (see parameter description for more
#'   details). See also
#'   [matchms CosineGreedy](https://matchms.readthedocs.io/en/latest/api/matchms.similarity.CosineGreedy.html) for more information.
#'
#' - `CosineHungarian()`: calculate the *cosine similarity score* as with
#'   `CosineGreedy`, but using the Hungarian algorithm to find the best
#'   matching peaks between the compared spectra. The algorithm can be
#'   configured with parameters `tolerance`, `mz_power` and `intensity_power`
#'   (see parameter description for more details). See also
#'   [matchms CosingHungarian](https://matchms.readthedocs.io/en/latest/api/matchms.similarity.CosineHungarian.html) for more information.
#'
#' - `ModifiedCosine()`: The modified cosine score aims at quantifying the
#'   similarity between two mass spectra. The score is calculated by finding
#'   the best possible matches between peaks of two spectra. Two peaks are
#'   considered a potential match if their m/z ratios lie within the given
#'   `tolerance`, or if their m/z ratios lie within the tolerance once a
#'   mass shift is applied. The mass shift is simply the difference in
#'   precursor-m/z between the two spectra.
#'   See also [matchms ModifiedCosine](https://matchms.readthedocs.io/en/latest/api/matchms.similarity.ModifiedCosine.html) for more information.
#'
#' - `NeutralLossesCosine()`: The neutral losses cosine score aims at
#'   quantifying the similarity between two mass spectra. The score is
#'   calculated by finding the best possible matches between peaks of two
#'   spectra. Two peaks are considered a potential match if their m/z ratios lie
#'   within the given `tolerance` once a mass shift is applied. The mass shift
#'   is the difference in precursor-m/z between the two spectra. See also
#'   [matchms NeutralLossesCosine](https://matchms.readthedocs.io/en/latest/api/matchms.similarity.NeutralLossesCosine.html) for more information.
#'
#' - `FingerprintSimilarity()`: Calculate similarity between molecules based
#'   on their fingerprints. For this similarity measure to work, fingerprints
#'   are expected to be derived by running *add_fingerprint()*. See also
#'   [matchms FingerprintSimilarity](https://matchms.readthedocs.io/en/latest/api/matchms.similarity.FingerprintSimilarity.html) for more information.
#'
#'
#' @note
#'
#' Parameters and algorithms are named as originally defined in the *matchms*
#' library (i.e. all parameters in *snake_case* while *CamelCase* is used
#' for the algorithms.
#'
#' @param x A [Spectra::Spectra()] object.
#'
#' @param y A [Spectra::Spectra()] object to compare against. If missing,
#'   spectra similarities are calculated between all spectra in `x`.
#'
#' @param param One of the parameter classes listed above (such as
#'   `CosineGreedy`) defining the similarity scoring function in Python
#'   and its parameters.
#'
#' @param tolerance `numeric(1)`: tolerated differences in the peaks' m/z. Peaks
#'   with m/z differences `<= tolerance` are considered matching.
#'
#' @param mz_power `numeric(1)`: the power to raise m/z to in the cosine
#'   function. The default is 0, in which case the peak intensity products will
#'   not depend on the m/z ratios.
#'
#' @param ignore_peaks_above_precursor For `NeutralLossesCosine()`:
#'   `logical(1)`: if `TRUE` (the default), peaks with m/z values larger than
#'   the precursor m/z are ignored.
#'
#' @param intensity_power `numeric(1)`: the power to raise intensity to in the
#'   cosine function. The default is 1.
#'
#' @param ... ignored.
#'
#' @return `compareSpectriPy()` Returns a `numeric` matrix with the scores,
#'   with the number of rows equal to `length(x)` and the number of columns
#'   equal to `length(y)`.
#'
#' @author Carolin Huber, Michael Witting, Johannes Rainer, Helge Hecht,
#'   Marilyn De Graeve
#'
#' @seealso [Spectra::compareSpectra()] in the *Spectra* package for pure R
#'   implementations of spectra similarity calculations.
#'
#' @export
#'
#' @examples
#'
#' library(Spectra)
#' ## Create some example Spectra.
#' DF <- DataFrame(
#'     msLevel = c(2L, 2L, 2L),
#'     name = c("Caffeine", "Caffeine", "1-Methylhistidine"),
#'     precursorMz = c(195.0877, 195.0877, 170.0924)
#' )
#' DF$intensity <- list(
#'     c(340.0, 416, 2580, 412),
#'     c(388.0, 3270, 85, 54, 10111),
#'     c(3.407, 47.494, 3.094, 100.0, 13.240)
#' )
#' DF$mz <- list(
#'     c(135.0432, 138.0632, 163.0375, 195.0880),
#'     c(110.0710, 138.0655, 138.1057, 138.1742, 195.0864),
#'     c(109.2, 124.2, 124.5, 170.16, 170.52)
#' )
#' sps <- Spectra(DF)
#'
#' ## Calculate pairwise similarity beween all spectra within sps with
#' ## matchms' CosineGreedy algorithm
#' ## Note: the first compareSpectriPy will take longer because the Python
#' ## environment needs to be set up.
#' res <- compareSpectriPy(sps, param = CosineGreedy())
#' res
#'
#' ## Next we calculate similarities for all spectra against the first one
#' res <- compareSpectriPy(sps, sps[1], param = CosineGreedy())
#'
#' ## Calculate pairwise similarity of all spectra in sps with matchms'
#' ## ModifiedCosine algorithm
#' res <- compareSpectriPy(sps, param = ModifiedCosine())
#' res
#'
#' ## Note that the ModifiedCosine method requires the precursor m/z to be
#' ## known for all input spectra. Thus, it is advisable to remove spectra
#' ## without precursor m/z before using this algorithm.
#' sps <- sps[!is.na(precursorMz(sps))]
#' compareSpectriPy(sps, param = ModifiedCosine())
NULL

setGeneric("compareSpectriPy", function(x, y, param, ...) {
    standardGeneric("compareSpectriPy")
})

setClass(
    "CosineGreedy",
    slots = c(
        tolerance = "numeric", mzPower = "numeric", intensityPower = "numeric"
    ),
    prototype = prototype(tolerance = 0.1, mzPower = 0.0, intensityPower = 1.0),
    validity = function(object) {
        msg <- NULL
        if (length(object@tolerance) != 1 || object@tolerance < 0) {
            msg <- c("'tolerance' has to be a positive number of length 1")
        }
        if (length(object@mzPower) != 1) {
            msg <- c(msg, "'mz_power' has to be a number of length 1")
        }
        if (length(object@intensityPower) != 1) {
            msg <- c(msg, "'intensity_power' has to be a number of length 1")
        }
        msg
    }
)
setClass(
    "CosineHungarian",
    contains = "CosineGreedy")
setClass(
    "ModifiedCosine",
    contains = "CosineGreedy")
setClass(
    "NeutralLossesCosine",
    contains = "CosineGreedy",
    slots = c(ignorePeaksAbovePrecursor = "logical"),
    prototype = prototype(ignorePeaksAbovePrecursor = TRUE),
    validity = function(object) {
        msg <- NULL
        if (length(object@ignorePeaksAbovePrecursor) != 1) {
            msg <- paste0("'ignorePeaksAbovePrecursor' has to be a ",
                          "positive number of length 1")
        }
        msg
    }
)
setClass(
    "FingerprintSimilarity",
    contains = "CosineGreedy",
    slots = c(similarityMeasure = "character"),
    prototype = prototype(similarityMeasure = "jaccard"))

#' @rdname compareSpectriPy
#'
#' @importFrom methods new
#'
#' @export
CosineGreedy <- function(tolerance = 0.1, mz_power = 0.0,
                         intensity_power = 1.0) {
    new("CosineGreedy",
        tolerance = as.numeric(tolerance),
        mzPower = as.numeric(mz_power),
        intensityPower = as.numeric(intensity_power)
        )
}

#' @rdname compareSpectriPy
#'
#' @export
CosineHungarian <- function(tolerance = 0.1, mz_power = 0.0,
                            intensity_power = 1.0) {
    new("CosineHungarian",
        tolerance = as.numeric(tolerance),
        mzPower = as.numeric(mz_power),
        intensityPower = as.numeric(intensity_power)
    )
}

#' @rdname compareSpectriPy
#'
#' @export
ModifiedCosine <- function(tolerance = 0.1, mz_power = 0.0,
                                intensity_power = 1.0) {
    new("ModifiedCosine",
        tolerance = as.numeric(tolerance),
        mzPower = as.numeric(mz_power),
        intensityPower = as.numeric(intensity_power)
    )
}

#' @rdname compareSpectriPy
#'
#' @export
NeutralLossesCosine <- function(tolerance = 0.1, mz_power = 0.0,
                                intensity_power = 1.0,
                                ignore_peaks_above_precursor = TRUE) {
    new("NeutralLossesCosine",
        tolerance = as.numeric(tolerance),
        mzPower = as.numeric(mz_power),
        intensityPower = as.numeric(intensity_power),
        ignorePeaksAbovePrecursor = as.logical(ignore_peaks_above_precursor)
    )
}

## #' @rdname compareSpectriPy
## #'
## #' @export
## FingerprintSimilarity <- function(similarity_measure = "jaccard") {
##     new("FingerprintSimilarity",
##         similarityMeasure = similarity_measure
##         )
## }

#' @rdname compareSpectriPy
#'
#' @exportMethod compareSpectriPy
setMethod(
    "compareSpectriPy",
    signature = c(x = "Spectra", y = "Spectra", param = "CosineGreedy"),
    function(x, y, param, ...) {
        .compare_spectra_python(x, y, param)
    }
)

#' @rdname compareSpectriPy
setMethod(
    "compareSpectriPy",
    signature = c(x = "Spectra", y = "missing", param = "CosineGreedy"),
    function(x, y, param, ...) {
        .compare_spectra_python(x, y = NULL, param)
    }
)

#' Method to construct a python function name for a param object.
#'
#' @noRd
setGeneric("py_fun", function(object, ...) {
    standardGeneric("py_fun")
})
setMethod("py_fun", "CosineGreedy", function(object) {
    matchms_similarity$CosineGreedy(object@tolerance, object@mzPower,
                                    object@intensityPower)
})
setMethod("py_fun", "CosineHungarian", function(object) {
    matchms_similarity$CosineHungarian(object@tolerance, object@mzPower,
                                       object@intensityPower)
})
setMethod("py_fun", "ModifiedCosine", function(object) {
    matchms_similarity$ModifiedCosine(object@tolerance, object@mzPower,
                                      object@intensityPower)
})
setMethod("py_fun", "NeutralLossesCosine", function(object) {
    matchms_similarity$NeutralLossesCosine(object@tolerance, object@mzPower,
                                           object@intensityPower,
                                           object@ignorePeaksAbovePrecursor)
})

.fun_name <- function(x) {
    class(x)[1L]
}

#' Internal function to calculate similarities with Python's matchms. `Spectra`
#' will be converted to Python.
#'
#' @param x `Spectra`
#'
#' @param y `Spectra`
#'
#' @param param Parameter object.
#'
#' @param value `character(1)` defining which value should be returned from the
#'     Python call.
#'
#' @return a `numeric` `matrix` nrow being length of `x`, nrow length `y`.
#'
#' @noRd
#'
#' @author Carolin Huber, Johannes Rainer, Wout Bittremieux
#'
#' @importFrom reticulate r_to_py py_to_r
.compare_spectra_python <- function(x, y = NULL, param,
                                    mapping = spectraVariableMapping()) {
    ## Handle empty input.
    if (!length(x) || (!length(y) & !is.null(y))) {
        return(matrix(NA_real_, ncol = length(y), nrow = length(x)))
    }

    ## Convert R spectra to Python.
    py_x <- rspec_to_pyspec(x, mapping = mapping)
    py_y <- if (!is.null(y)) rspec_to_pyspec(y) else py_x
    is_symmetric <- is.null(y)

    ## Compute the similarity scores with matchms.
    scores <- matchms$calculate_scores(
        py_x, py_y, py_fun(param),
        is_symmetric = is_symmetric
        )
    ## Collect results
    py_to_r(scores$to_array()[paste0(.fun_name(param), "_score")])
}

caf <- DataFrame(
    msLevel = c(2L, 2L),
    name = "Caffeine",
    precursorMz = c(195.0877, 195.0877)
)
caf$intensity <- list(
    c(340.0, 416, 2580, 412),
    c(388.0, 3270, 85, 54, 10111))
caf$mz <- list(
    c(135.0432, 138.0632, 163.0375, 195.0880),
    c(110.0710, 138.0655, 138.1057, 138.1742, 195.0864))
caf <- Spectra(caf)

mhd <- DataFrame(
    msLevel = c(2L, 2L),
    precursorMz = c(170.0924, 170.0924),
    id = c("HMDB0000001", "HMDB0000001"),
    name = c("1-Methylhistidine", "1-Methylhistidine"))
mhd$mz <- list(
    c(109.2, 124.2, 124.5, 170.16, 170.52),
    c(83.1, 96.12, 97.14, 109.14, 124.08, 125.1, 170.16))
mhd$intensity <- list(
    c(3.407, 47.494, 3.094, 100.0, 13.240),
    c(6.685, 4.381, 3.022, 16.708, 100.0, 4.565, 40.643))
mhd <- Spectra(mhd)

test_that("param constructors work", {
    ## Errors
    expect_error(CosineGreedy(tolerance = c(1.2, 1.5)), "length 1")
    expect_error(CosineGreedy(tolerance = -1.2), "length 1")
    expect_error(CosineGreedy(mz_power = c(1, 2)), "length 1")
    expect_error(CosineGreedy(intensity_power = c(1, 2)), "length 1")
    expect_error(NeutralLossesCosine(
        ignore_peaks_above_precursor = c(TRUE, FALSE)), "length 1")
})

test_that("CosineGreedy constructor works", {
    res <- CosineGreedy(tolerance = 5)
    expect_s4_class(res, "CosineGreedy")
    expect_equal(res@tolerance, 5)
})

test_that("CosineHungarian constructor works", {
    res <- CosineHungarian(intensity_power = 1.3)
    expect_s4_class(res, "CosineGreedy")
    expect_equal(res@intensityPower, 1.3)
})

test_that("ModifiedCosine constructor works", {
    res <- ModifiedCosine(mz_power = 4.3)
    expect_s4_class(res, "ModifiedCosine")
    expect_equal(res@mzPower, 4.3)
})

test_that("NeutralLossesCosine constructor works", {
    res <- NeutralLossesCosine(ignore_peaks_above_precursor = FALSE)
    expect_s4_class(res, "NeutralLossesCosine")
    expect_false(res@ignorePeaksAbovePrecursor)
})

test_that(".fun_name works", {
    a <- CosineGreedy()
    expect_equal(.fun_name(a), "CosineGreedy")
    a <- CosineHungarian()
    expect_equal(.fun_name(a), "CosineHungarian")
    a <- ModifiedCosine()
    expect_equal(.fun_name(a), "ModifiedCosine")
    a <- NeutralLossesCosine()
    expect_equal(.fun_name(a), "NeutralLossesCosine")
})

test_that("py_fun works", {
    res <- py_fun(CosineGreedy(tolerance = 0.5, mz_power = 0.3,
                                           intensity_power = 0.2))
    expect_equal(class(res)[1L],
                 "matchms.similarity.CosineGreedy.CosineGreedy")
    res <- py_fun(CosineHungarian(tolerance = 0.4, mz_power = 0.3,
                                           intensity_power = 0.2))
    expect_equal(
        class(res)[1L], "matchms.similarity.CosineHungarian.CosineHungarian")
    res <- py_fun(ModifiedCosine(tolerance = 0.1, mz_power = 0.3,
                                           intensity_power = 0.2))
    expect_equal(
        class(res)[1L], "matchms.similarity.ModifiedCosine.ModifiedCosine")
    res <- py_fun(NeutralLossesCosine(tolerance = 0.1, mz_power = 0.3,
                                                  intensity_power = 0.2,
                                                  ignore_peaks_above_precursor = FALSE))
    expect_equal(
        class(res)[1L],
        "matchms.similarity.NeutralLossesCosine.NeutralLossesCosine")
})

test_that(".compare_spectra_python works", {
    all <- c(caf, mhd)
    res <- .compare_spectra_python(all, caf, CosineGreedy())
    expect_true(nrow(res) == 4)
    expect_true(ncol(res) == 2)
    expect_equal(res[1, 1], 1)
    expect_equal(res[2, 2], 1)
    expect_true(all(res[3:4, ]== 0))

    res_sps <- compareSpectra(all, caf, tolerance = 0.1, ppm = 0)
    diffs <- abs(as.numeric(res - res_sps))
    expect_true(all(diffs < 0.01))

    ## only one spectra
    res <- .compare_spectra_python(all, param = CosineGreedy())
    res_2 <- .compare_spectra_python(all, all, param = CosineGreedy())
    expect_equal(res, res_2)

    ## try with empty Spectra
    res <- .compare_spectra_python(all, all[integer()], CosineGreedy())
    expect_true(is.numeric(res))
    expect_true(nrow(res) == length(all))
    expect_true(ncol(res) == 0)

    res <- .compare_spectra_python(all[integer()], param = CosineGreedy())
    expect_true(is.numeric(res))
    expect_true(nrow(res) == 0)
    expect_true(ncol(res) == 0)

    res <- .compare_spectra_python(all[integer()], all, CosineGreedy())
    expect_true(is.numeric(res))
    expect_true(nrow(res) == 0)
    expect_true(ncol(res) == length(all))
})

test_that("compareSpectriPy works", {
    all <- c(caf, mhd)
    res_all <- compareSpectriPy(all, param = CosineGreedy())
    expect_true(is.numeric(res_all))
    expect_true(nrow(res_all) == length(all))
    expect_true(ncol(res_all) == length(all))
    expect_equal(diag(res_all), c(1, 1, 1, 1))

    ## CosineGreedy
    res <- compareSpectriPy(caf, mhd, param = CosineGreedy())
    expect_true(nrow(res) == 2)
    expect_true(ncol(res) == 2)
    expect_true(all(res == 0))
    res <- compareSpectriPy(caf, mhd, param = CosineGreedy(tolerance = 10))
    expect_true(any(res > 0))

    res <- compareSpectriPy(caf[1L], mhd[1L], param = CosineGreedy())
    expect_true(is.numeric(res))
    expect_true(nrow(res) == 1)
    expect_true(ncol(res) == 1)
    expect_true(all(res == 0))
    expect_equal(res[1, 1], res_all[1, 3])

    ## CosineHungarian
    res <- compareSpectriPy(caf, mhd, param = CosineHungarian())
    expect_true(nrow(res) == 2)
    expect_true(ncol(res) == 2)
    expect_true(all(res == 0))
    res <- compareSpectriPy(caf, mhd, param = CosineHungarian(tolerance = 10))
    expect_true(any(res > 0))

    ## ModifiedCosine
    res <- compareSpectriPy(caf, mhd, param = ModifiedCosine())
    expect_true(nrow(res) == 2)
    expect_true(ncol(res) == 2)
    expect_true(all(res > 0))
    res <- compareSpectriPy(caf, mhd, param = ModifiedCosine(tolerance = 10))
    expect_true(all(res > 0.3))

    all_mod <- all
    all_mod$precursorMz[3] <- NA_real_
    expect_error(compareSpectriPy(all_mod, all, param = ModifiedCosine()),
                 "Expect precursor to be positive")

    ## NeutralLossesCosine
    res <- compareSpectriPy(caf, mhd, param = NeutralLossesCosine())
    expect_true(nrow(res) == 2)
    expect_true(ncol(res) == 2)
    expect_true(all(res == 0))
    res <- compareSpectriPy(caf, mhd, param = NeutralLossesCosine(tolerance = 10))
    expect_true(all(res > 0))
})

## -----------------------------------------------------------------------------
#| eval: false
# install.packages("BiocManager")
# BiocManager::install("SpectriPy")


## -----------------------------------------------------------------------------
#| label: libraries
#| message: false
#' R session:

library(Spectra)
library(SpectriPy)


## -----------------------------------------------------------------------------
stopifnot(require("MsDataHub"))


## -----------------------------------------------------------------------------
#| label: load-data
#| message: false
#' R session:

#' Loading the data from a mzML file as a `Spectra` object
library(MsDataHub)
fl <- MsDataHub::PestMix1_DDA.mzML()
mzml_r <- Spectra(fl)


## -----------------------------------------------------------------------------
#| label: filter MS level
#' R session:

#' Restrict to MS level 2 spectra
mzml_r <- filterMsLevel(mzml_r, 2)
mzml_r <- mzml_r[lengths(mzml_r) >= 3]
mzml_r


## -----------------------------------------------------------------------------
#| label: convert spectra to matchms
#' R session:

#' Convert the R Spectra to a list of Python matchms.Spectrum objects
tmp <- rspec_to_pyspec(mzml_r)


## -----------------------------------------------------------------------------
#' R session:

#' Class of the converted variable
class(tmp)

#' First element
class(tmp[1])


## #' Python session:
## 
## #' Access the Python data structure stored in the R session
## r.tmp

## -----------------------------------------------------------------------------
#| label: assign variable to Python
#' R session:

#' Assign the data to a variable in the Python environment
py_set_attr(py, "mzml_py", rspec_to_pyspec(mzml_r))
names(py)


## #' Python session:
## 
## #' Data type of the variable:
## type(mzml_py)
## 
## #' The length of the list:
## len(mzml_py)
## 
## #' Data type of the first element:
## type(mzml_py[0])
## 
## #' Intensities of the first spectrum:
## mzml_py[0].peaks.intensities

## #' Python session:
## 
## import matchms.filtering as mms_filt
## 
## #' Iterate over the Spectrum list and scale the intensities
## for i in range(len(mzml_py)):
##     mzml_py[i] = mms_filt.normalize_intensities(mzml_py[i])
## 
## #' Intensities for the first spectrum
## mzml_py[0].peaks.intensities

## -----------------------------------------------------------------------------
#| label: inspect changed intensities in R
#' R session:

#' Access the intensities of the first spectrum
py_get_attr(py, "mzml_py")[0]$peaks$intensities


## #' Python session:
## 
## #' Access the `fl` variable from the R session:
## r.fl
## 
## type(r.fl)

## #' Python session:
## 
## #' Access the `Spectra` object with the original data from R; the data
## #' gets directly translated on-the-fly
## r.mzml_r
## 
## #' Access the intensities of the first spectrum
## r.mzml_r[0].peaks.intensities

## -----------------------------------------------------------------------------
#| label: define MGF file
#' R session:

f_mgf <- system.file("extdata", "mgf", "test.mgf", package = "SpectriPy")


## #' Python session
## 
## import matchms
## from matchms.importing import load_from_mgf
## 
## mgf_py = list(load_from_mgf(r.f_mgf))
## mgf_py

## -----------------------------------------------------------------------------
#' R session:

#' Access the first spectrum
py$mgf_py[[1]]


## -----------------------------------------------------------------------------
#' R session:

#' Extract the peaks matrix from the first spectrum
py_to_r(py$mgf_py[[1]]$peaks$to_numpy)


## -----------------------------------------------------------------------------
#' R session:

#' Convert and copy the data to R
mgf_r <- pyspec_to_rspec(py_get_attr(py, "mgf_py"))
mgf_r

#' Extract intensities
mgf_r$intensity


## -----------------------------------------------------------------------------
#| label: initialize MsBackendPy for Python variable
#' R session:

#' Create a Spectra object with a MsBackendPy backend for the
#' attribute "mgf_py"
mgf <- Spectra("mgf_py", source = MsBackendPy())
mgf


## -----------------------------------------------------------------------------
#| label: access MS level and intensity through MsBackendPy
#' R session:

#' Extract MS level
msLevel(mgf)

#' Extract intensity values
intensity(mgf)


## #' Python session:
## 
## #' Scale intensities
## for i in range(len(mgf_py)):
##     mgf_py[i] = mms_filt.normalize_intensities(mgf_py[i])
## 

## -----------------------------------------------------------------------------
#| label: get intensities after modifying in Python
#' R session:

#' Get intensities after scaling the intensities in Python:
intensity(mgf)


## -----------------------------------------------------------------------------
#' R session:

mzml_r


## -----------------------------------------------------------------------------
#' R session:

mzml_r <- setBackend(mzml_r, backend = MsBackendPy(),
                     pythonVariableName = "mzml_p")
mzml_r


## -----------------------------------------------------------------------------
#' R session:

print(object.size(mzml_r), units = "MB")


## -----------------------------------------------------------------------------
#' R session:

mz(mzml_r)


## #' Python session:
## 
## len(mzml_p)
## 
## mzml_p[0].peaks.mz

## -----------------------------------------------------------------------------
#' R session:

spectraVariableMapping("matchms")


## -----------------------------------------------------------------------------
#' R session:

spectraVariableMapping("spectrum_utils")


## -----------------------------------------------------------------------------
#' R session:

spectraVariableMapping(mzml_r)


## -----------------------------------------------------------------------------
#' R session:

library(MsBackendMgf)
sps <- Spectra(system.file("extdata", "mgf", "test.mgf",
                           package = "SpectriPy"), source = MsBackendMgf())
spectraVariables(sps)


## -----------------------------------------------------------------------------
#' R session:

mp <- spectraVariableMapping("matchms", c(SMILES = "smiles", INCHI = "inchi"))
mp


## -----------------------------------------------------------------------------
#' R session:

sps <- setBackend(sps, backend = MsBackendPy(),
                  pythonVariableName = "sps_p",
                  spectraVariableMapping = mp)
spectraVariables(sps)


## -----------------------------------------------------------------------------
#' R session:

sps$SMILES |> head()


## -----------------------------------------------------------------------------
#' R session:

#' List the *default* spectra variable mapping in R and python, respectively
defaultSpectraVariableMapping()


## #' Python session:
## 
## #' Available metadata for the first spectrum
## mgf_py[0].metadata.keys()

## -----------------------------------------------------------------------------
#' R session:

#' Add mapping for additional spectra variables to the default mapping in R and
#' python, respectively
map <- c(defaultSpectraVariableMapping(), smiles = "smiles",
         inchi = "inchi", name = "compound_name")
map


## -----------------------------------------------------------------------------
#| label: convert R to Python with spectra variables
#' R session:

#' Convert the Python MS data structures to an R `Spectra`
mgf_r <- pyspec_to_rspec(py_get_attr(py, "mgf_py"), mapping = map)
spectraVariables(mgf_r)


## -----------------------------------------------------------------------------
#| label: access spectra variables from Python in R
#' R session:

#' Show the first values for the spectra variable "name"
mgf_r$name |>
    head()

#' Show the first values for the spectra variable "inchi"
mgf_r$inchi |>
    head()


## -----------------------------------------------------------------------------
#' R session:

#' List available spectra variables
spectraVariables(mgf)


## -----------------------------------------------------------------------------
#' R session:

#' Get the first entries from the inchi variable
mgf$inchi |>
    head()


## -----------------------------------------------------------------------------
#| label: add additional mapping
#' R session:

#' Add mapping for additional spectra variables to the `MsBackendPy`
m <- defaultSpectraVariableMapping()
m["precursorIntensity"] <- "pepmassint"
spectraVariableMapping(mgf) <- m

#' List available spectra variables
spectraVariables(mgf)

#' Get spectra data for two variables
spectraData(mgf, c("precursorMz", "precursorIntensity"))


## -----------------------------------------------------------------------------
#| label: compareSpectra with mzML and MGF
#' R session:

#' Create a `Spectra` for the scaled MS data in Python
mzml <- Spectra("mzml_py", source = MsBackendPy())

#' Calculate the pairwise similarity between all spectra
sim <- compareSpectra(mzml, mgf, tolerance = 0.1)
dim(sim)


## #' Python session:
## 
## import matchms.similarity as mms_similarity
## 
## #' Calculate similarity scores
## scores = matchms.calculate_scores(
##     mzml_py, mgf_py, mms_similarity.CosineHungarian(tolerance = 0.1))
## 
## #' Extract similarity scores
## sim = scores.to_array()["CosineHungarian_score"]

## -----------------------------------------------------------------------------
#| label: compare dot product vs Cosine Hungarian
#' R session:

#' Plot the similarity scores against each other
plot(sim, py$sim, pch = 21, col = "#000000ce", bg = "#00000060",
     xlab = "Dot product", ylab = "Cosine Hungarian")
grid()


## -----------------------------------------------------------------------------
#' R session:

sessionInfo()

test_that("spectraVariableMapping,Spectra works", {
    expect_true(all(names(spectraVariableMapping(s)) %in%
                    names(coreSpectraVariables())))
    expect_true(length(spectraVariableMapping(s)) <
                length(coreSpectraVariables()))
    ## <-
    a <- s
    spectraVariableMapping(a) <- c(precursorMz = "precursor_mz",
                                   msLevel = "ms_level")
    expect_equal(a@backend@spectraVariableMapping,
                 c(precursorMz = "precursor_mz", msLevel = "ms_level"))
    expect_true(all(is.na(precursorCharge(a))))
})

---
title: "Detailed information on installation and configuration"
package: SpectriPy
format:
  html:
    minimal: true
    theme: flatly
vignette: >
  %\VignetteIndexEntry{Detailed information on installation and configuration}
  %\VignetteKeywords{Mass Spectrometry, MS, MSMS, Metabolomics, Infrastructure, Quantitative}
  %\VignettePackage{SpectriPy}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
  %\VignetteDepends{BiocStyle,SpectriPy,reticulate}
---

# Introduction

This document provides detailed installation and configuration instructions for
the [*SpectriPy*](https://github.com/RforMassSpectrometry/SpectriPy)
package. For first time R users see also section @sec-first in the appendix. For
advanced Python/system configuration and troubleshooting see sections
@sec-python and @sec-fix in the appendix.


# Installation

*SpectriPy* relies on, and extends, the
[*reticulate*](https://rstudio.github.io/reticulate/index.html) package for
interoperability between Python and R. For installation of *SpectriPy*, Python
needs to be installed on the system. See also the [Install Python
Packages](https://rstudio.github.io/reticulate/articles/python_packages.html)
documentation from the *reticulate* package for information on installing
Python. Since version 0.99.5, *SpectriPy* uses the new
[`py_require()`](https://rstudio.github.io/reticulate/reference/py_require.html)
function from the *reticulate* package to install and manage all required Python
libraries (i.e., *matchms*, *spectrum_utils* and *numpy*) including all their
dependencies.

Installation of the package using:

```{r}
#| eval: false
#' R session:
install.packages("BiocManager")
BiocManager::install("SpectriPy")
```

should work on most systems. See also section [*Fixing package installation or
loading
problems*](https://rformassspectrometry.github.io/SpectriPy/articles/SpectriPy.html#sec-fix)
in the appendix of the *SpectriPy* vignette if installing or loading of the
package fails.


## Configure reticulate with host system Python environment (optional)

*SpectriPy* (respectively *reticulate*) can be configured to use an available
Python environment of the host system, instead of using the automatic setup. In
this case, however, the user **must** first install the required Python
libraries manually in that local environment. The required libraries are:

- [*matchms*](https://github.com/matchms) 0.31
- [*spectrum_utils*](https://github.com/bittremieux-lab/spectrum_utils) 0.3.2
- *numpy* 2.2.0

To use *SpectriPy* with a local, system, Python environment either the
`RETICULATE_PYTHON` or the `RETICULATE_PYTHON_ENV` environment variable needs to
be set. The former should point to the *python* binary, while the latter should
point to the (virtualenv or conda) environment with all necessary libraries
installed. These environment variables can either be set globally in the
operation system settings, or specifically for R. See also the
[Environment](https://docs.posit.co/ide/user/ide/guide/environments/r/managing-r.html)
documentation from Posit for more information on how to define environment
variables for R. In the example below, we set the variable using `Sys.setenv()`
**before** the *SpectriPy* or *reticulate* libraries are loaded:

```{r}
#| eval: false
#' R session:

Sys.setenv(RETICULATE_PYTHON_ENV="<path-to-the-virtualenv-with-libraries>")
library(SpectriPy)
```

The `Sys.setenv()` call can also be added to a *.Rprofile* file to automatically
set the environmental variable when R starts.


## SpectriPy pre-requisites and installation instructions

### Installing Bioconductor

Bioconductor is required to install *SpectriPy*, as described below and from
source [https://bioconductor.org/install/](https://bioconductor.org/install/).

The current release of Bioconductor is version 3.21; it works with R version
4.5.0. Users of older R and Bioconductor must update their installation to take
advantage of new features and to access packages that have been added to
Bioconductor since the last release.

The development version of Bioconductor is version 3.22; it works with R version
4.5.0. More recent *devel* versions of R (if available) will be supported during
the next Bioconductor release cycle.

Once R has been installed, get the latest version of Bioconductor by starting R
and entering the following commands.

It may be possible to change the Bioconductor version of an existing
installation; see the *Changing version* section of the BiocManager vignette.

Details, including instructions to install additional packages and to update,
find, and troubleshoot are provided below. A devel version of Bioconductor is
available. There are good reasons for using `BiocManager::install()` for
managing Bioconductor resources.

```{r}
#| eval: false
#' R session:

if (!require("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install(version = "3.21")
```

### Installing *SpectriPy*

*SpectriPy* can be installed through Bioconductor:

```{r}
#| eval: false
#' R session:

BiocManager::install("SpectriPy")
```

### Check installation completed

The status of installation can be easily checked by starting R and entering the
following commands.

This command loads the *SpectriPy* package, if correct
installed. `py_availble()` should return `TRUE`.

```{r}
#' R session:

library(SpectriPy)
py_available()
```


# Appendix

## Installation instructions for first-time R users{#sec-first}

Instructions to install R and RStudio for the first time are described below,
from source
[https://rstudio-education.github.io/hopr/packages2.html](https://rstudio-education.github.io/hopr/packages2.html).

### Installing R and RStudio

To get started with R, you need to acquire your own copy. This appendix will
show you how to download R as well as RStudio, a software application that makes
R easier to use. You’ll go from downloading R to opening your first R session.

Both R and RStudio are free and easy to download.

### How to Download and Install R

R is maintained by an international team of developers who make the language
available through the web page of The Comprehensive R Archive Network. The top
of the web page provides three links for downloading R. Follow the link that
describes your operating system: Windows, Mac, or Linux.

#### Windows

To install R on Windows, click the “Download R for Windows” link. Then click the
*base* link. Next, click the first link at the top of the new page. This link
should say something like *Download R 3.0.3 for Windows*, except the `3.0.3`
will be replaced by the most current version of R. The link downloads an
installer program, which installs the most up-to-date version of R for
Windows. Run this program and step through the installation wizard that
appears. The wizard will install R into your program files folders and place a
shortcut in your Start menu. Note that you’ll need to have all of the
appropriate administration privileges to install new software on your machine.

#### Mac

To install R on a Mac, click the *Download R for Mac* link. Next, click on the
`R-3.0.3` package link (or the package link for the most current release of
R). An installer will download to guide you through the installation process,
which is very easy. The installer lets you customize your installation, but the
defaults will be suitable for most users. I’ve never found a reason to change
them. If your computer requires a password before installing new progams, you’ll
need it here.

> **Binaries Versus Source**
>
> R can be installed from precompiled binaries or built from source on any
> operating system. For Windows and Mac machines, installing R from binaries is
> extremely easy. The binary comes preloaded in its own installer. Although you
> can build R from source on these platforms, the process is much more
> complicated and won’t provide much benefit for most users. For Linux systems,
> the opposite is true. Precompiled binaries can be found for some systems, but
> it is much more common to build R from source files when installing on
> Linux. The download pages on [CRAN’s website](https://cran.r-project.org/)
> provide information about building R from source for the Windows, Mac, and
> Linux platforms.

#### Linux

R comes preinstalled on many Linux systems, but you’ll want the newest version
of R if yours is out of date. The [CRAN website](https://cran.r-project.org/)
provides files to build R from source on Debian, Redhat, SUSE, and Ubuntu
systems under the link *Download R for Linux.* Click the link and then follow
the directory trail to the version of Linux you wish to install on. The exact
installation procedure will vary depending on the Linux system you use. CRAN
guides the process by grouping each set of source files with documentation or
README files that explain how to install on your system.

> **32-bit Versus 64-bit**
>
> R comes in both 32-bit and 64-bit versions. Which should you use? In most
> cases, it won’t matter. Both versions use 32-bit integers, which means they
> compute numbers to the same numerical precision. The difference occurs in the
> way each version manages memory. 64-bit R uses 64-bit memory pointers, and
> 32-bit R uses 32-bit memory pointers. This means 64-bit R has a larger memory
> space to use (and search through).  As a rule of thumb, 32-bit builds of R are
> faster than 64-bit builds, though not always. On the other hand, 64-bit builds
> can handle larger files and data sets with fewer memory management
> problems. In either version, the maximum allowable vector size tops out at
> around 2 billion elements. If your operating system doesn’t support 64-bit
> programs, or your RAM is less than 4 GB, 32-bit R is for you. The Windows and
> Mac installers will automatically install both versions if your system
> supports 64-bit R.

### Using R

R isn’t a program that you can open and start using, like Microsoft Word or
Internet Explorer. Instead, R is a computer language, like C, C++, or UNIX. You
use R by writing commands in the R language and asking your computer to
interpret them. In the old days, people ran R code in a UNIX terminal window—as
if they were hackers in a movie from the 1980s. Now almost everyone uses R with
an application called RStudio, and I recommend that you do, too.

> **R and UNIX**
>
> You can still run R in a UNIX or BASH window by typing the command:
>
> `R`
>
> which opens an R interpreter. You can then do your work and close the
> interpreter by running q() when you are finished.

### RStudio

RStudio is an application like Microsoft Word—except that instead of helping you
write in English, RStudio helps you write in R. I use RStudio throughout the
book because it makes using R much easier. Also, the RStudio interface looks the
same for Windows, Mac OS, and Linux. That will help me match the book to your
personal experience.

You can download [RStudio](https://posit.co/products/open-source/rstudio/) for
free. Just click the *Download RStudio* button and follow the simple
instructions that follow. Once you’ve installed RStudio, you can open it like
any other program on your computer—usually by clicking an icon on your desktop.

> **The R GUIs**
>
> Windows and Mac users usually do not program from a terminal window, so the
> Windows and Mac downloads for R come with a simple program that opens a
> terminal-like window for you to run R code in. This is what opens when you
> click the R icon on your Windows or Mac computer. These programs do a little
> more than the basic terminal window, but not much. You may hear people refer
> to them as the Windows or Mac R GUIs.

When you open RStudio, a window appears with three panes in it, as in Figure 1.
The largest pane is a console window. This is where you’ll run your R code and
see results. The console window is exactly what you’d see if you ran R from a
UNIX console or the Windows or Mac GUIs. Everything else you see is unique to
RStudio. Hidden in the other panes are a text editor, a graphics window, a
debugger, a file manager, and much more. You’ll learn about these panes as they
become useful throughout the course of this book.  The RStudio IDE for R.

![The RStudio IDE for R](https://rstudio-education.github.io/hopr/images/hopr_aa01.png)

Figure 1: The RStudio IDE for R.

> **Do I still need to download R?**
>
> Even if you use RStudio, you’ll still need to download R to your
> computer. RStudio helps you use the version of R that lives on your computer,
> but it doesn’t come with a version of R on its own.

### Opening R

Now that you have both R and RStudio on your computer, you can begin using R by
opening the RStudio program. Open RStudio just as you would any program, by
clicking on its icon or by typing *RStudio* at the Windows Run prompt.

## Startup and Python configuration {#sec-python}

The way Python dependencies are defined and managed has changed in *SpectriPy*
beginning with version 0.99.6. *SpectriPy* now declares Python dependencies
using the `py_require()` function from the *reticulate* package. These Python
package dependencies requested via `py_require()` will automatically be
provisioned and made available for the user when the *SpectriPy* package is
loaded, *via* an ephemeral Python virtual environment. Eventually missing
libraries are downloaded and installed automatically.

The previous *virtualenv* or *conda*-based setup is now replaced by the
preferred `py_require()` approach, hence the previous package's options and
environment variables `"spectripy.use_system"`, `SPECTRIPY_USE_SYSTEM`,
`"spectripy.use_conda"`, `SPECTRIPY_USE_CONDA`, `"spectripy.env"`,
`SPECTRIPY_ENV` are now ignored.

A pre-defined Python environment can be used by pointing the `RETICULATE_PYTHON`
environment variable to the respective Python binary or the
`RETICULATE_PYTHON_ENV` to the path of the environment. All Python libraries
required by *SpectriPy* need however to be installed and available in that
Python environment, as *SpectriPy* respectively *reticulate*'s `py_require()`
functionality will be bypassed. These Python libraries are:

- *matchms* >= 0.31
- *spectrum_utils* >= 0.3.2
- *numpy* >= 2.2.0

More information on this manual setup can be found in the *Detailed information
on installation and configuration* vignette.

See also the help of the *reticulate* package for more information on
configuring Python with R.


## Fixing package installation or loading problems {#sec-fix}

*SpectriPy* loads and imports the required Python libraries *matchms*,
*spectrum_utils* and *numpy* during package loading/attaching (e.g. using
`library(SpectriPy)`). Installation or loading of the package can thus fail if
these libraries or the required versions can not be found. *SpectriPy* uses the
newer `py_require()` approach from the *reticulate* package to manage Python
requirements.

Trouble shooting:

1) Check the output of `Sys.getenv()`: is there a system variable
   `RETICULATE_PYTHON` or `RETICULATE_PYTHON_ENV` defined? Problem: *reticulate*
   will use the specified Python or Python environment and skip automatic
   installation of the Python libraries. Solution: a) unset this environment
   variables b) manually install the required Python libraries in that Python
   environment (see section @sec-python).

2) Is there a *default* *r-reticulate* virtual environment (e.g. created by
   another package or a previous version of *reticulate*) on the system, i.e.,
   does `virtualenv_exists("r-reticulate")` return `TRUE`? Problem: *reticulate*
   will use this environment instead of the *ephemeral virtual environment* that
   would be managed through `py_require()`. Solution: define an environment
   variable `RETICULATE_USE_MANAGED_VENV="yes"`. This variable can either be
   defined system wide, or by adding a line `RETICULATE_USE_MANAGED_VENV="yes"`
   to a file named *.Renviron* in the user's home directory. See also the
   [Environment](https://docs.posit.co/ide/user/ide/guide/environments/r/managing-r.html)
   documentation from Posit for more information on how to define environment
   variables for R.

If this does not solve the issue, have also a look at the [order of
discovery](https://rstudio.github.io/reticulate/articles/versions.html#order-of-discovery)
documentation of *reticulate* which clearly explains how *reticulate* tries to
define and use the Python setup. Most *SpectriPy* installation/loading problems
come from the fact that another Python setup than the *ephemeral virtual
environment* (which is the current suggested mode) is used. The solution
suggested in point 2 above should solve most of these problems.

More information can also be found in the *reticulate*'s [Python version
configuration](https://rstudio.github.io/reticulate/articles/versions.html)
documentation.


# Session information

```{r}
#' R session:

sessionInfo()
```

---
title: "Enabling integration of Python libraries and R packages for combined mass spectrometry data analysis"
package: SpectriPy
format:
  html:
    minimal: true
    theme: flatly
vignette: >
  %\VignetteIndexEntry{Enabling integration of Python libraries and R packages for combined mass spectrometry data analysis}
  %\VignetteKeywords{Mass Spectrometry, MS, MSMS, Metabolomics, Infrastructure, Quantitative}
  %\VignettePackage{SpectriPy}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
  %\VignetteDepends{Spectra,BiocStyle,SpectriPy,reticulate,MsBackendMgf,MsDataHub,mzR}
---

**Compiled**: `r date()`

**Note**: since version 0.99.6 *SpectriPy* uses the newer, recommended approach
to install and configure required Python libraries (i.e., through *reticulate*'s
`py_require()` functionality). This affects also the way a pre-defined Python
environment can be used. See section [Startup and Python configuration](https://rformassspectrometry.github.io/SpectriPy/articles/detailed-installation-configuration.html#sec-python) in the
*Detailed information on installation and configuration* vignette for updated
information.

# Introduction

Powerful software libraries for mass spectrometry (MS) data are available in
both Python and R, covering specific and sometimes distinct aspects in the
analysis of proteomics and metabolomics data. R's *reticulate* package converts
basic data types between Python and R and enables a seamless interoperability of
both programming languages. The *SpectriPy* package extends *reticulate*
allowing to translate MS data data structures between R and Python,
specifically, between R's `Spectra` objects and Python MS data structures from
the [*matchms*](https://github.com/matchms/matchms) and
[*spectrum_utils*](https://github.com/bittremieux-lab/spectrum_utils) Python
libraries. In addition, functionality from Python's *matchms* library is
directly wrapped into R functions allowing a seamless integration into R based
workflows. *SpectriPy* thus enables powerful proteomics or metabolomics analysis
workflows combining the strengths of Python and R MS libraries.

This vignette provides information on how to share and translate MS data
structures between R and Python to enable combined Python and R-based analysis
workflows. For an example use case analysis, see the *SpectriPy tutorial:
Annotation of LC-MS/MS spectra* vignette.


# System requirements

The [*reticulate*](https://rstudio.github.io/reticulate/) package enables a
seamless integration of Python with R by translating the core data structures
between the two programming languages and sharing a Python (respectively R)
runtime environment, including shared variables, that can be accessed from the
other programming language. If the *reticulate* package is not already
available, it will be installed during the installation of *SpectriPy*.
Alternatively, it can be installed using `install.packages("reticulate")`.

The *SpectriPy* package builds on *reticulate* providing in addition the
functionality to translate data structures for MS data between the two
languages. Specifically, the package translates between R's `Spectra` objects
(from the `r BiocStyle::Biocpkg("Spectra")` package) and Python's
`matchms.Spectrum` and `spectrum_utils.spectrum.MsmsSpectrum` objects from the
[*matchms*](https://github.com/matchms/matchms) and
[*spectrum_utils*](https://github.com/bittremieux-lab/spectrum_utils) Python
libraries, respectively.


# Installation

If the *BiocManager* package is not already available, please install it with
`install.packages("BiocManager")`. As a system dependency, the package requires
Python (version >= 3.12) to be available. During package installation,
*SpectriPy* will by default install all required Python libraries
automatically. See section [Startup and Python
configuration](https://rformassspectrometry.github.io/SpectriPy/articles/detailed-installation-configuration.html#sec-python)
in the *Detailed information on installation and configuration* vignette for
information on manual library installation or usage of a pre-defined or system
Python environment. See section [Fixing package installation or loading
problems](https://rformassspectrometry.github.io/SpectriPy/articles/detailed-installation-configuration.html#sec-fix)
in the *Detailed information on installation and configuration* vignette for
some hints how to solve package installation or loading problems.

To install the package use the code below:

```{r}
#| eval: false
install.packages("BiocManager")
BiocManager::install("SpectriPy")
```


# Translating data structures between R and Python

This section describes how MS data is being converted between R and Python to
enable analyses combining both programming/analysis languages and accessing the
same shared MS data. The first two sections describe how MS data can be
converted between R and Python. See also section @sec-backend for a more
elegant approach and improved integration with [*Spectra*](https://)-based
workflows in R.

## Library loading and system setup

Below we load all required packages. By loading *SpectriPy*, the package will
evaluate if the required Python libraries (i.e., *matchms* version >= 0.1,
*spectrum_utils* version >= 0.3.2 and *numpy* version >= 2.2) are available. If
they are not available, *SpectriPy* will install them using functionality from
the *reticulate* R package. See section [Startup and Python
configuration](https://rformassspectrometry.github.io/SpectriPy/articles/detailed-installation-configuration.html#sec-python)
in the *Detailed information on installation and configuration* vignette for
more configuration options of *SpectriPy*.

The *reticulate* package will be loaded by *SpectriPy*, ensuring the R/Python
integration provided by that package to be available as well. To better
discriminate between R and Python code chunks, we add the comment `#' R
session:` or `#' Python session:` to label the R and Python code chunks,
respectively.

```{r}
#| label: libraries
#| message: false
#' R session:

library(Spectra)
library(SpectriPy)
```


## Converting MS data from R to Python

In this section we show how MS data can be converted from R to Python. Below we
first define the name and path of a data file in mzML format (provided by the `r
BiocStyle::Biocpkg("MsDataHub")` package) and load that data as an R `Spectra`
object. We thus first ensure that the package *MsDataHub* is installed. If the
line below results in an error you need to first install *MsDataHub* using
`BiocManager::install("MsDataHub")`.

```{r}
stopifnot(require("MsDataHub"))
```

Next, we define the file path to the test file and load the data as a `Spectra`
object in R.

```{r}
#| label: load-data
#| message: false
#' R session:

#' Loading the data from a mzML file as a `Spectra` object
library(MsDataHub)
fl <- MsDataHub::PestMix1_DDA.mzML()
mzml_r <- Spectra(fl)
```

We next restrict the data to MS level 2 and remove spectra with less than 3
(fragment) peaks.

```{r}
#| label: filter MS level
#' R session:

#' Restrict to MS level 2 spectra
mzml_r <- filterMsLevel(mzml_r, 2)
mzml_r <- mzml_r[lengths(mzml_r) >= 3]
mzml_r
```

This `Spectra` object can now be converted to equivalent data structures in
Python using the `rspec_to_pyspec()` function:

```{r}
#| label: convert spectra to matchms
#' R session:

#' Convert the R Spectra to a list of Python matchms.Spectrum objects
tmp <- rspec_to_pyspec(mzml_r)
```

The `tmp` variable is now a Python list of `matchms.Spectrum` objects:

```{r}
#' R session:

#' Class of the converted variable
class(tmp)

#' First element
class(tmp[1])
```

Note that this Python data structure is now stored within the R session.
Therefore, we have two full copies of the data in memory, i.e., an R `Spectra`
object and a Python `matchms.Spectrum` object (see Section @sec-backend for an
alternative to avoid multiple data copies). We can access the variable from the
associated Python environment through *reticulate*'s special `r` attribute using
`r.<variable name in R>`.

```{python}
#' Python session:

#' Access the Python data structure stored in the R session
r.tmp
```

While it is thus possible to access the variable, it is suggested, and has also
performance advantages, if the Python data is stored directly in the
Python environment. Similar to the special `r` attribute, *reticulate* provides
the `py` variable in R that allows access to the main Python environment
associated with the R session. Attributes can be assigned to the Python
environment with the `py_set_attr()` function. We repeat our data conversion
operation but assign the result to a Python attribute `"mzml_py"`:

```{r}
#| label: assign variable to Python
#' R session:

#' Assign the data to a variable in the Python environment
py_set_attr(py, "mzml_py", rspec_to_pyspec(mzml_r))
names(py)
```

We can now access this data directly from within Python:

```{python}
#| label: data types of variables
#' Python session:

#' Data type of the variable:
type(mzml_py)

#' The length of the list:
len(mzml_py)

#' Data type of the first element:
type(mzml_py[0])

#' Intensities of the first spectrum:
mzml_py[0].peaks.intensities
```

This allows us to analyze and process the MS data directly in Python. As an
example we load the *matchms.filtering* library and scale the intensity
values of each spectrum with the `normalize_intensities()` function such that
their total sum is 1.

```{python}
#| label: normalize intensities with matchms
#' Python session:

import matchms.filtering as mms_filt

#' Iterate over the Spectrum list and scale the intensities
for i in range(len(mzml_py)):
    mzml_py[i] = mms_filt.normalize_intensities(mzml_py[i])

#' Intensities for the first spectrum
mzml_py[0].peaks.intensities
```

We can also access the changed data from R. The `py_get_attr()` function can be
used to retrieve the variable with the changed Python data. Through the
*reticulate* package it is then also possible to call attributes and Python
functions directly from R. To extract the intensities of the first spectrum we
can use the same code as above, just replacing `.` with `$`. Note also that,
since the variable returned by `py_get_attr()` is a Python object, we need to
use index `0` to access the first element.

```{r}
#| label: inspect changed intensities in R
#' R session:

#' Access the intensities of the first spectrum
py_get_attr(py, "mzml_py")[0]$peaks$intensities
```

Note that with the `rspec_to_pyspec()` function, we created a copy of the
original data in Python. We have thus now two variables, the `mzml_r` variable
in R with the original, unchanged, intensity values, and the `mzml_py` attribute
in Python with the scaled intensity values. To get the processing results back
to R we need to convert the data from Python to R. This can be done with the
`pyspec_to_rspec()` function which translates Python MS data structures into an
R `Spectra` object (see the following @sec-py-to-r section for more
information). A more elegant way is to use the `MsBackendPy` backend for
R' `Spectra` objects (see Section @sec-backend).

With *reticulate* we can use the `py` special variable in R to access attributes
defined in the associated Python environment. Similarly, *reticulate* defines an
attribute `r` in the Python session that allows to access variables in R. When
variables are accessed this way, they are automatically converted to the
corresponding data types in the other programming language if an `r_to_py()`
method (or `py_to_r()` method) is implemented for them. Such methods are
defined for the basic data types, so, when accessing for example the `fl`
R variable from Python, it is converted from the R `character` data type to the
equivalent Python `str` data type:

```{python}
#' Python session:

#' Access the `fl` variable from the R session:
r.fl

type(r.fl)
```

*SpectriPy* implements an `r_to_py()` method for `Spectra` objects, so, when
`Spectra` objects are accessed from Python, they are also automatically
translated to a Python list of `matchms.Spectrum` objects:

```{python}
#' Python session:

#' Access the `Spectra` object with the original data from R; the data
#' gets directly translated on-the-fly
r.mzml_r

#' Access the intensities of the first spectrum
r.mzml_r[0].peaks.intensities
```

While this automatic conversion is convenient in some cases, the manual
translation with `rspec_to_pyspec()` and `pyspec_to_rspec()` is preferred, as it
allows to configure the handling of the spectra variables (i.e. metadata) and
avoids eventual repeated translation of the data.


## Converting MS data from Python to R {#sec-py-to-r}

To show conversion of MS data from Python to R, we import a test data file in
MGF format using the Python *matchms* library. This MGF file is provided within
the *SpectriPy* package so we first define its file name and path in R.

```{r}
#| label: define MGF file
#' R session:

f_mgf <- system.file("extdata", "mgf", "test.mgf", package = "SpectriPy")
```

We next load the required Python library and import the data in Python. We can
access the variable (defined in the R session) with the file name to import the
data from through the `r.<variable name>`.

```{python}
#| label: load MGF in Python
#| warning: false
#' Python session

import matchms
from matchms.importing import load_from_mgf

mgf_py = list(load_from_mgf(r.f_mgf))
mgf_py
```

The MS data from the MGF is now loaded in Python as a list of
`matchms.Spectrum` objects. We can also directly access this variable from the R
session through `py$<variable name in Python>`. Below we access the first
spectrum in that list:

```{r}
#' R session:

#' Access the first spectrum
py$mgf_py[[1]]
```

The data is thus provided in a *matchms.Spectrum* object. We extract the *m/z*
and intensity peak matrix from that spectrum using the built-in functionality
from the *reticulate* package that allows to call Python functions from R or
translate between basic data types. As an example we below get the `peaks`
attribute from the first spectrum and convert that to a Python `numpy`
array. This array can then be translated to an R `matrix` with the `py_to_r()`
function (see also the [*Calling Python from
R*](https://rstudio.github.io/reticulate/articles/calling_python.html) vignette
from the *reticulate* package for examples on accessing data from Python):

```{r}
#' R session:

#' Extract the peaks matrix from the first spectrum
py_to_r(py$mgf_py[[1]]$peaks$to_numpy)
```

While such functionality can thus be used to extract the MS data from Python,
it is for MS data analysis in R more convenient to transform the full MS data to
an R `Spectra` object using *SpectriPy*'s `pyspec_to_rspec()` function. Note
that below we use the `py_get_attr()` function to get the Python variable with
the MS data instead of `py$mgf_py`. Accessing the Python attribute through
`py$mgf_py` immediately translates the Python list into an R `list` through the
default `py_to_r()` function and `pyspec_to_rspec()` thus iterates over the R
`list` (in R). With `py_get_attr()` the attribute is accessed in its
native Python data type and `pyspec_to_rspec()` hence iterates over the data in
Python, which has a minimal performance advantage.

```{r}
#' R session:

#' Convert and copy the data to R
mgf_r <- pyspec_to_rspec(py_get_attr(py, "mgf_py"))
mgf_r

#' Extract intensities
mgf_r$intensity
```

The full MS data is thus now available as a `Spectra` object. Note however that
with `pyspec_to_rspec()` the complete MS data gets **copied**. We have now
two (detached) variables containing the same MS data, one in R and one in
Python.


## Using a dedicated MS data *backend* for MS data in Python {#sec-backend}

An alternative, and more elegant, approach to avoid keeping multiple copies of
spectral objects (i.e., `Spectra`, `matchms.Spectrum`) in memory, is to use a
dedicated data *backend* for the `Spectra` object. This allows direct access of
MS data in Python from R without having an additional copy in R. The R's
`Spectra` object separates by design the functionality to analyze MS data from
the code to *represent* or retrieve the MS data, which is provided by dedicated
data *backends*. These allow for example to import MS data from different file
formats, or to store and access the data with different storage modes,
respectively (see also [this
tutorial](https://jorainer.github.io/SpectraTutorials/articles/Spectra-backends.html)
for more information). The *SpectriPy* packages defines such a backend, the
`MsBackendPy`, that allows to directly access MS data stored in Python (in their
respective data formats).

As an example we create below a `Spectra` object for the MS data (in Python)
previously imported from the MGF file (i.e. `"mgf_py"`) and using the
`MsBackendPy` as the data `source` (i.e., backend).

```{r}
#| label: initialize MsBackendPy for Python variable
#' R session:

#' Create a Spectra object with a MsBackendPy backend for the
#' attribute "mgf_py"
mgf <- Spectra("mgf_py", source = MsBackendPy())
mgf
```

In contrast to using the `pyspec_to_rspec()` function, no data was copied or
converted by this call. The `Spectra` object (through its backend) does only
keep a *reference* to the original data attribute in Python, but no data. Data
is retrieved and translated on-the-fly each time it is requested from the
`Spectra` object. Thus, calling e.g., `msLevel()` or `intensity()` causes the
backend to iterate over the MS data in Python, extract the respective
information, translate it and return it to the user.

```{r}
#| label: access MS level and intensity through MsBackendPy
#' R session:

#' Extract MS level
msLevel(mgf)

#' Extract intensity values
intensity(mgf)
```

While the performance is a little lower, compared to translating all data to R
using the `pyspec_to_rspec()` function, this approach is much more memory
efficient, because no additional copies of MS data are generated and only the MS
data currently required for a certain analysis task is loaded and translated at
a time.

Also, because data is always retrieved on-the-fly from Python, any changes to
the MS data attribute in Python are also immediately reflected in the respective
`Spectra` object. To illustrate this we below scale the intensities of the mass
peaks in Python:

```{python}
#| label: normalize intensities of MGF in Python
#' Python session:

#' Scale intensities
for i in range(len(mgf_py)):
    mgf_py[i] = mms_filt.normalize_intensities(mgf_py[i])

```

The intensities retrieved from the `Spectra` object are now also scaled.

```{r}
#| label: get intensities after modifying in Python
#' R session:

#' Get intensities after scaling the intensities in Python:
intensity(mgf)
```

While we created a `Spectra` object from an existing data object in Python, it
is similarly possible change the backend of a `Spectra` object to a
`MsBackendPy`. This follows the general concept of `Spectra` objects to support
changing between data representations, respectively backends, at any time. The
first data set in this vignette was imported from mzML files:

```{r}
#' R session:

mzml_r
```

This `Spectra` object uses the `MsBackedMzR` backend as data representation. We
can change the backend of `Spectra` objects using the `setBackend()`
function. Thus, below we change the backend to `MsBackendPy`, specifying also
the name of the variable in Python that should be used to store the data.

```{r}
#' R session:

mzml_r <- setBackend(mzml_r, backend = MsBackendPy(),
                     pythonVariableName = "mzml_p")
mzml_r
```

This converted all the MS data of the `mzml_r` `Spectra` object to Python
objects and assigned these to the Python variable with the name `"mzml_p"`. The
resulting `Spectra` object uses now a `MsBackendPy` backend which keeps only the
reference to the Python variable but no data. The size of the `Spectra` object
is thus very small.

```{r}
#' R session:

print(object.size(mzml_r), units = "MB")
```

And accessing the data will retrieve and convert the data from the referenced
Python objects:

```{r}
#' R session:

mz(mzml_r)
```

And all data can also be accessed in Python:

```{python}
#' Python session:

len(mzml_p)

mzml_p[0].peaks.mz
```

In addition to the MS data (i.e., the *m/z* and intensity values),
`setBackend()` will also convert core general spectra variables and store them
in the Python MS data structure as spectrum related metadata. By default, only a
limited set of spectra variables, the ones with the direct counterpart in
Python, are translated. These are for the *matchms* library:

```{r}
#' R session:

spectraVariableMapping("matchms")
```

with the names being the spectra variable names in R and the values the names of
the respective metadata fields in Python. The *spectrum_utils* library supports
only few, fixed, variables:

```{r}
#' R session:

spectraVariableMapping("spectrum_utils")
```

The spectra variables that have been transferred to Python with the
`setBackend()` call above are:

```{r}
#' R session:

spectraVariableMapping(mzml_r)
```

Thus, only some of the variables from the original `Spectra` object are now
available after changing the backend to `MsBackendPy`. However, since *matchms*
supports additional, arbitrary, metadata fields, it is also possible to convert
and transfer other spectra variables with `MsBackendPy`. As an example, we below
import data from an MGF.

```{r}
#' R session:

library(MsBackendMgf)
sps <- Spectra(system.file("extdata", "mgf", "test.mgf",
                           package = "SpectriPy"), source = MsBackendMgf())
spectraVariables(sps)
```

There are thus also spectra variables, such as `"SMILES"` and `"INCHI"` that are
not part of the default `spectraVariableMapping("matchms")` and would thus not
be converted/transferred by default when changing to a `MsBackendPy`. We can
however use the `spectraVariableMapping` parameter with `setBackend()` to define
which spectra variables should be transferred. The parameter takes a named
character with names being the names of the spectra variables to transfer and
values the names that should be used for the metadata in Python. We use the
`spectraVariableMapping()` function to append the mapping for the spectra
variables `"SMILES"` and `"INCHI"` to the default one.

```{r}
#' R session:

mp <- spectraVariableMapping("matchms", c(SMILES = "smiles", INCHI = "inchi"))
mp
```

We pass this character vector to the `setBackend()` call.

```{r}
#' R session:

sps <- setBackend(sps, backend = MsBackendPy(),
                  pythonVariableName = "sps_p",
                  spectraVariableMapping = mp)
spectraVariables(sps)
```

The spectra variables `"INCHI"` and `"SMILES"` have thus also been stored in
Python:

```{r}
#' R session:

sps$SMILES |> head()
```

See also the next section for more information on the mapping between
`Spectra`'s spectra variables and *matchms* metadata.

The `MsBackendPy` has full *read/write* support, i.e., it allows to add new
spectra variables or change existing spectra and/or peaks variables through the
available replacement methods `spectraData()<-`, `peaksData()<-`,
`intensity()<-`, `mz()<-` and `$<-`. Currently, this operations change however
the full MS data of the associated Python variable. Thus, even if for example
only the retention times are replaced using `$rtime<-`, the (full!) data is
first loaded from Python to R, the retention time values are then replaced and
finally the full data (spectra and peaks variables) are again stored to the
associated Python variable. Replacement operations are thus quite memory
demanding.

## Conversion of spectra variables

Conversion of the MS peaks data (i.e. the *m/z* and intensity values) is always
performed by the `rspec_to_pyspec()` and `pyspec_to_rspec()` functions. But next
to the peaks data, also additional information are available for individual
spectra. In R/*Spectra* these variables are called *spectra variables* while in
*matchms* they are stored as a *metadata* attribute to a `matchms.Spectrum`
object. The *SpectriPy* package defines a core set of spectra variables
that are by default converted by the `rspec_to_pyspec()` and `pyspec_to_rspec()`
function. These default variables can be accessed using the
`defaultSpectraVariableMapping()` function:

```{r}
#' R session:

#' List the *default* spectra variable mapping in R and python, respectively
defaultSpectraVariableMapping()
```

These variables, if present in the respective data object, are transferred (and
renamed) to the MS data structure of the other programming language. The names
of this character vector represent the name of the spectra variable in the
`Spectra` object, the elements (values) the names of the respective metadata
keys in Python's `matchms.Spectrum` class. This default mapping thus transfers
for example the `precursorMz()` spectra variable of R to the `"precursor_mz"`
spectrum metadata in Python (and *vice versa*). Note that any spectra variable
or metadata **not** being part of such a `mapping` will be ignored and hence not
converted.

Below we inspect the available metadata in the `matchms.Spectrum` objects that
were imported from the MGF file.

```{python}
#' Python session:

#' Available metadata for the first spectrum
mgf_py[0].metadata.keys()
```

Several additional metadata variables, such as `"smiles"`, `"inchi"` or
`"compound_name"`, not part of the default variables, are available. To
transfer also these, we create below a custom mapping adding in addition also a
mapping for these 3 variables:

```{r}
#' R session:

#' Add mapping for additional spectra variables to the default mapping in R and
#' python, respectively
map <- c(defaultSpectraVariableMapping(), smiles = "smiles",
         inchi = "inchi", name = "compound_name")
map
```

Such custom mapping can be passed with the parameter `mapping` to the
`pyspec_to_rspec()` (and also the `rspec_to_pyspec()`) function which will then
convert the full data to R.

```{r}
#| label: convert R to Python with spectra variables
#' R session:

#' Convert the Python MS data structures to an R `Spectra`
mgf_r <- pyspec_to_rspec(py_get_attr(py, "mgf_py"), mapping = map)
spectraVariables(mgf_r)
```

The respective metadata values have thus been added as new spectra variables to
our `Spectra` object and can also be extracted:

```{r}
#| label: access spectra variables from Python in R
#' R session:

#' Show the first values for the spectra variable "name"
mgf_r$name |>
    head()

#' Show the first values for the spectra variable "inchi"
mgf_r$inchi |>
    head()
```

When using a `MsBackendPy`, all metadata attributes from the `matchms.Spectrum`
objects can be accessed and extracted with the `spectraData()` function. The
`spectraVariables()` function lists all available columns:

```{r}
#' R session:

#' List available spectra variables
spectraVariables(mgf)
```

We thus have already access to e.g. the `"inchi"` variable:

```{r}
#' R session:

#' Get the first entries from the inchi variable
mgf$inchi |>
    head()
```

A `spectraVariableMapping()` can however be used to rename variables. Below we
add for example a mapping of the `matchms.Spectrum` metadata attribute
`"pepmassint"` to the (core) variable `"precursorIntensity"` using the
`MsBackendPy`:

```{r}
#| label: add additional mapping
#' R session:

#' Add mapping for additional spectra variables to the `MsBackendPy`
m <- defaultSpectraVariableMapping()
m["precursorIntensity"] <- "pepmassint"
spectraVariableMapping(mgf) <- m

#' List available spectra variables
spectraVariables(mgf)

#' Get spectra data for two variables
spectraData(mgf, c("precursorMz", "precursorIntensity"))
```

The `r_to_py()` methods do not support additional parameters, thus, in order to
use a similar mapping also with the `r_to_py()` method for `Spectra` the
*global* spectra variable mapping need to be changed. See the help of the
`setSpectraVariableMapping()` function for more details.


## Combined MS data analysis

With the functionality to translate between R and Python MS data structures, the
*SpectriPy* package enables thus a MS data analysis combining functionality
provided by both R and Python libraries. Some of the functionality of the
*matchms* Python library are directly wrapped into R functions simplifying
their use and inclusion in R-based workflows (see the help of
`compareSpectraPy()` and `filterSpectriPy()` functions). Combined analyses with
code chunks in both programming languages has however the advantage to use the
respective functionality provided by the original package/library.

As a simple use case we calculate below spectra similarities using two different
similarity scores between spectra from the mzML and the MGF files.

In R we can use the `compareSpectra()` function that by default calculates the
normalized dot product similarity between the compared spectra. The MS data from
the mzML file was processed in Python. To use this data we first create a
`Spectra` object with a `MsBackendPy` backend. For the MS data from the MGF file
we re-use the `mgf` variable, which is also a `Spectra` with a `MsBackendPy`
backend referencing the MS data imported and processed in Python.

```{r}
#| label: compareSpectra with mzML and MGF
#' R session:

#' Create a `Spectra` for the scaled MS data in Python
mzml <- Spectra("mzml_py", source = MsBackendPy())

#' Calculate the pairwise similarity between all spectra
sim <- compareSpectra(mzml, mgf, tolerance = 0.1)
dim(sim)
```

We can calculate the *Cosine Hungarian* similarity score in Python using the
functionality from the *matchms* library. Here we use the *original* attributes
available in Python, i.e. `mzml_py` and `mgf_py`.

```{python}
#| label: Cosine Hungarian similarity in Python
#' Python session:

import matchms.similarity as mms_similarity

#' Calculate similarity scores
scores = matchms.calculate_scores(
    mzml_py, mgf_py, mms_similarity.CosineHungarian(tolerance = 0.1))

#' Extract similarity scores
sim = scores.to_array()["CosineHungarian_score"]
```

Alternatively, we can also directly use any `Spectra` object from R as the
data is converted automatically when accessed from Python. We can for
example use `r.mzml` instead of `mzml_py` in the call above, which converts
the `Spectra` object `mzml` in R to Python before calculating the similarities.

We can also directly compare the scores calculated using the two different
algorithms.

```{r}
#| label: compare dot product vs Cosine Hungarian
#' R session:

#' Plot the similarity scores against each other
plot(sim, py$sim, pch = 21, col = "#000000ce", bg = "#00000060",
     xlab = "Dot product", ylab = "Cosine Hungarian")
grid()
```


## Summary

By translating between R and Python MS data structures, the *SpectriPy* package
enables data analyses that combine functionalities from both programming
language. In terms of (memory) efficiency, usage of *SpectriPy*'s `MsBackendPy`
backend for `Spectra` objects has clear advantages over the repeated translation
and copying of the MS data. See also section @sec-comments for general comments.


# Appendix

## General comments {#sec-comments}

- Be careful accessing Python attributes using `py$<attribute name>`: base
  Python data types will be automatically converted to the equivalent R data
  type. For MS data, it might be better to get the attributes using the
  `py_get_attr()` functions.

- Since MS data can be large, it is suggested the user converts MS data mostly
  manually, and only if/when needed, using `rspec_to_pyspec()` and
  `pyspec_to_rspec()`.

- The `rspec_to_pyspec()` and `pyspec_to_rspec()` functions **copy** the data
  while transferring. Thus, there will be eventually two (detached) copies of
  the same data in Python and R.

- The `MsBackendPy` backend allows to directly interface MS data from
  Python. Data will be converted on-the-fly, so no additional copies of the data
  exist.

- Be aware that, since the `MsBackendPy` backend does not contain any MS data
  but simply interfaces the MS data in Python, any changes to this data in
  Python affect also the `Spectra` object using that backend.


# Session information

```{r}
#' R session:

sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterSpectriPy.R
\name{filterSpectriPy}
\alias{filterSpectriPy}
\alias{select_by_intensity}
\alias{select_by_mz}
\alias{remove_peaks_around_precursor_mz}
\alias{normalize_intensities}
\alias{filterSpectriPy,Spectra,filter_param-method}
\title{Filter Spectra using Python's matchms library}
\usage{
select_by_intensity(intensity_from = 10, intensity_to = 200)

select_by_mz(mz_from = 0, mz_to = 1000)

remove_peaks_around_precursor_mz(mz_tolerance = 17)

normalize_intensities()

\S4method{filterSpectriPy}{Spectra,filter_param}(object, param, mapping = spectraVariableMapping(), ...)
}
\arguments{
\item{intensity_from}{\code{numeric(1)}: Set lower threshold for peak intensity.
Default is 10.}

\item{intensity_to}{\code{numeric(1)}: Set upper threshold for peak intensity.
Default is 200.}

\item{mz_from}{\code{numeric(1)}: Set lower threshold for m/z peak positions.
Default is 0.}

\item{mz_to}{\code{numeric(1)}: Set upper threshold for m/z peak positions.
Default is 1000.}

\item{mz_tolerance}{\code{numeric(1)}: Tolerance of m/z values that are not
allowed to lie within the precursor mz. Default is 17 Da.}

\item{object}{A \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object.}

\item{param}{one of parameter classes listed above (such as
\code{select_by_intensity()}) defining the filter/processing function in
Python and its parameters.}

\item{mapping}{named \code{character()} defining which spectra variables/metadata
should be converted between R and Python and how they should be renamed.
Defaults to \code{spectraVariableMapping()}. See \code{\link[=setSpectraVariableMapping]{setSpectraVariableMapping()}}
for more information.}

\item{...}{ignored.}
}
\value{
\code{filterSpectriPy()} returns a \code{Spectra} object on which the
filtering/processing function has been applied
}
\description{
The \code{filterSpectriPy()} function allows to filter/process a \code{Spectra} object
using the \code{select_by_intensity()}, \code{select_by_mz()},
\code{remove_peaks_around_precursor_mz()}, and \code{normalize_intensities()}
functions of the Python
\href{https://matchms.readthedocs.io/en/latest/api/matchms.filtering.html}{matchms.filtering}
module.

Selection and configuration of the algorithm can be performed with one of
the parameter objects (equivalent to \emph{matchms}' function names):
\itemize{
\item \code{select_by_intensity()}: Keeps only the peaks within defined intensity
range (keep if \code{intensity_from} >= intensity >= \code{intensity_to}). See also
the respective \href{https://matchms.readthedocs.io/en/latest/api/matchms.filtering.peak_processing.select_by_intensity.html}{documentation in \emph{matchms}}.
\item \code{select_by_mz()}: Keeps only the peaks between \code{mz_from} and \code{mz_to}
(keep if \code{mz_from} >= m/z >= \code{mz_to}). See also the respective
\href{https://matchms.readthedocs.io/en/latest/api/matchms.filtering.peak_processing.select_by_mz.html}{documentation in \emph{matchms}}.
\item \code{remove_peaks_around_precursor_mz()}: Removes the peaks that are within
\code{mz_tolerance} (in Da) of the precursor mz, excluding the precursor peak.
\item \code{normalize_intensities()}: Normalizes the intensities of peaks
(and losses) to unit height.
}
}
\note{
The first call to the \code{filterSpectriPy()} function can take longer because
the Python environment needs to be first set up.

\code{filterSpectriPy()} first translates the \code{Spectra} to Python, applies the
filter functions from the \emph{matchms} Python libraries and then translates
the filtered data back to a \code{Spectra} object. Thus, any spectra variables
other than those that are translated between R and Python will be lost
during the processing. Use \code{\link[=setSpectraVariableMapping]{setSpectraVariableMapping()}} to define which
spectra variables should be transferred/converted between R and Python.
See also examples below for more information.

The \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object returned by \code{filterSpectriPy()} will
\strong{always} use an in-memory backend (i.e. the \code{\link[Spectra:MsBackend]{Spectra::MsBackendMemory()}}),
independently of the backend used by the backend used by the input
\code{Spectra}.
}
\examples{

library(Spectra)

## create some example Spectra
DF <- DataFrame(
    msLevel = c(2L, 2L, 2L),
    name = c("Caffeine", "Caffeine", "1-Methylhistidine"),
    precursorMz = c(195.0877, 195.0877, 170.0924)
)
DF$intensity <- list(
    c(340.0, 416, 2580, 412),
    c(388.0, 3270, 85, 54, 10111),
    c(3.407, 47.494, 3.094, 100.0, 13.240))
DF$mz <- list(
    c(135.0432, 138.0632, 163.0375, 195.0880),
    c(110.0710, 138.0655, 138.1057, 138.1742, 195.0864),
    c(109.2, 124.2, 124.5, 170.16, 170.52))
sps <- Spectra(DF)

## Filter: select_by_intensity
res <- filterSpectriPy(
    sps, select_by_intensity(intensity_from = 15, intensity_to = 300))
## Only mass peaks with intensities between the specified limits are
## retained
intensity(res)
## Compared to the original intensities
intensity(sps)

## Note that the spectra variable `"name"` was lost during conversion of
## the MS data between R and Python:
sps$name
any(spectraVariables(res) == "name")

## Only spectra variables defined by `spectraVariableMapping()` are
## converted and thus retained:
spectraVariableMapping()

## We can also pass a custom *spectra variable mapping* with the `mapping`
## parameter to the `filterSpectriPy()` function. Below we create such
## a mapping by adding the translation of a spectra variable `"name"` to
## a metadata name `"compound_name"` to the default spectra variable
## mapping `defaultSpectraVariableMapping()`.
map <- c(defaultSpectraVariableMapping(), name = "compound_name")
map

## Repeat the filtering operation passing this mapping information:
res <- filterSpectriPy(
    sps, select_by_intensity(intensity_from = 15, intensity_to = 300),
    mapping = map)
res$name

}
\seealso{
\itemize{
\item \code{\link[Spectra:filterMsLevel]{Spectra::filterIntensity()}}, \code{\link[Spectra:filterMsLevel]{Spectra::filterMzRange()}},
\code{\link[Spectra:addProcessing]{Spectra::scalePeaks()}} in the \code{Spectra} package for pure R
implementations of filtering/processing calculations.
\item \code{\link[=rspec_to_pyspec]{rspec_to_pyspec()}} or \code{\link[=pyspec_to_rspec]{pyspec_to_rspec()}} for the functions used to
translated the MS data between R and Python.
}
}
\author{
Thomas Naake
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/conversion.R
\name{conversion}
\alias{conversion}
\alias{spectraVariableMapping,character-method}
\alias{spectraVariableMapping,missing-method}
\alias{setSpectraVariableMapping}
\alias{defaultSpectraVariableMapping}
\alias{r_to_py.Spectra}
\alias{rspec_to_pyspec}
\alias{pyspec_to_rspec}
\title{Converting between R and Python MS data structures}
\usage{
\S4method{spectraVariableMapping}{character}(object, x = character(), ...)

\S4method{spectraVariableMapping}{missing}(object, ...)

setSpectraVariableMapping(x)

defaultSpectraVariableMapping()

\method{r_to_py}{Spectra}(x, convert = FALSE)

rspec_to_pyspec(
  x,
  mapping = spectraVariableMapping(),
  pythonLibrary = c("matchms", "spectrum_utils")
)

pyspec_to_rspec(
  x,
  mapping = spectraVariableMapping(),
  pythonLibrary = c("matchms", "spectrum_utils")
)
}
\arguments{
\item{object}{For \code{spectraVariableMapping()}: not used.}

\item{x}{\code{Spectra} object.}

\item{...}{For \code{spectraVariableMapping()}: not used.}

\item{convert}{Boolean; should Python objects be automatically converted to
their R equivalent? Defaults to \code{FALSE}.}

\item{mapping}{named \code{character()} vector defining which spectra
variables/metadata should be translated between R and Python and how
they should be renamed. Defaults to \code{spectraVariableMapping()}.}

\item{pythonLibrary}{For \code{rspec_to_pyspec()} and \code{pyspec_to_rspec()}:
\code{character(1)} defining the Python library to which (or from which)
data structures the data should be converted.
Possible options are \code{"matchms"} or \code{"spectrum_utils"} with \code{"matchms"}
being the default.}
}
\value{
For \code{r_to_py.Spectra()} and \code{rspec_to_pyspec()}: Python list of
MS data structures, either \code{matchms.Spectrum} or
\code{spectrum_utils.spectrum.MsmsSpectrum} objects. For \code{pyspec_to_rspec()}:
\code{\link[Spectra:Spectra]{Spectra::Spectra()}} with the MS data of all \code{matchms.Spectrum} objects
in the submitted \code{list}.
}
\description{
The \code{rspec_to_pyspec()} and \code{pyspec_to_rspec()} functions allow to convert
(translate) MS data structures between R and Python. At present the
R \code{\link[Spectra:Spectra]{Spectra::Spectra()}} objects can be either translated into a list of
\href{https://github.com/matchms/matchms}{matchms} Python \code{matchms.Spectrum}
objects or
\href{https://github.com/bittremieux-lab/spectrum_utils}{spectrum_utils} Python
\code{spectrum_utils.spectrum.MsmsSpectrum} objects.
For better integration with the \emph{reticulate} R package also a
\code{r_to_py.Spectra()} method is available.

The mapping of spectra variables (in R) to (Python) spectra metadata can
be configured and defined with the \code{setSpectraVariableMapping()} and
\code{spectraVariableMapping()}. These get and set the \emph{global} (system wide)
setting and are thus also used by the \code{r_to_py()} method.

Properties for translation to the MS data objects of the different Python
libraries are:
\itemize{
\item \emph{matchms}: the \code{matchms.Spectrum} objects support arbitrary metadata, so
any spectra variable can be translated and stored in these objects.
\item \emph{spectrum_utils}: the \code{spectrum_utils.spectrum.MsmsSpectrum} object
supports metadata variables \emph{identifier} (\code{character}), \emph{precursor_mz}
(\code{numeric}), \emph{precursor_charge} (\code{integer}) and optionally also
\emph{retention_time} (\code{numeric}).
}

See the indivudual function's documentation for more details.

Function to convert R Spectra objects into a Python list of matchms Spectrum
objects using the \code{reticulate} package.
}
\section{Translation of MS data objects}{


MS data structures can be translated between R and Python using the
\code{rspec_to_pyspec()} and \code{pyspec_to_rspec()} functions, or with the
\code{r_to_py()} method.
\itemize{
\item \code{rspec_to_pyspec()} translates an R \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object into a
list of Python MS data objects, which can be, depending on parameter
\code{pythonLibrary}, \code{matchms.Spectrum} objects (for
\code{pythonLibrary = "matchms"}, the default) or
\code{spectrum_utils.spectrum.MsmsSpectrum} objects (for
\code{pythonLibrary = "spectrum_utils"}). Parameter \code{mapping} allows to specify
which spectra variables from the \code{Spectra} object \code{x} should be converted
in addition to the peaks data (m/z and intensity values). It defaults to
\code{mapping = spectraVariableMapping()} (See the respective help below for
more information on the variable mapping). While being fast, this function
first loads all peaks and spectra data into memory before translating to
Python data structures. A less memory intense operation could be to call
this function in a loop to only load parts of the data at a time into
memory.
\item \code{pyspec_to_rspec()} translates a single, or a list of \code{matchms.Spectrum}
objects (with parameter \code{pythonLibrary = "matchms"}, the default) or a
list of \code{spectrum_utils.spectrum.MsmsSpectrum} objects (with parameter
\code{pythonLibrary = "spectrum_utils"}) to a \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object.
Parameter \code{mapping} allows to specify the metadata variables that
should be translated and mapped in addition to the peaks data. The
library used to represent the MS data in Python needs to be specified with
parameter \code{pythonLibrary}.
\item \code{r_to_py.Spectra()} is equivalent to
\code{rspec_to_pyspec(pythonLibrary = "matchms")}. The spectra
variables that should be converted can be configures with
\code{setSpectraVariableMapping()} (see documentation below).
}
}

\section{Mapping of spectra variables (metadata)}{


Metadata for MS spectra are represented and stored as \emph{spectra variables}
in the R \code{\link[Spectra:Spectra]{Spectra::Spectra()}} objects. Also Python MS data structures
store such metadata along with the mass peak data. While spectra metadata
is thus supported by data structures in both programming languages,
different names and naming conventions are used. The
\code{spectraVariableMapping()} and \code{setSpectraVariableMapping()} functions allow
to define how the names of spectra metadata (spectra variables) should be
translated between R and Python. To support also the different naming
conventions used by the Python libraries \emph{matchms} and \emph{spectrum_utils},
\code{spectraVariableMapping()} defines different mapping schemes for these,
using by default the mapping for \emph{matchms}. Note also that \emph{spectrum_utils}
supports only few selected metadata/spectra variables, so any additional
spectra variables defined by the mapping will be ignored.
The \code{r_to_py()} and \code{py_to_r()} functions will use the selected naming
scheme to name the spectra variables accordingly. Also, only
spectra metadata/variables in \code{spectraVariableMapping()} will be translated.
The initial mapping is based on this
\href{https://github.com/matchms/matchms/blob/master/matchms/data/known_key_conversions.csv}{definition in matchms}.
\itemize{
\item \code{defaultSpectraVariableMapping()}: returns the \emph{default} mapping between
spectra variables and Python metadata names for the \emph{matchms} library.
\item \code{spectraVariableMapping()}: returns the currently defined spectra variable
mapping as a named character vector, with names
representing the names of the spectra variables in R and elements the
respective names of the spectra metadata in Python.
Use \code{\link[Spectra:spectraData]{Spectra::spectraVariables()}} on the \code{Spectra} object that should
be converted with \code{r_to_py()} to list all available spectra variables.
\code{r_to_py()} and \code{py_to_r()} for MS data structures will use this default
mapping.
Calling \code{spectraVariableMapping()} defining also the Python library (e.g.,
\code{spectraVariableMapping("matchms")} or
\code{spectraVariableMapping("spectrum_utils")}) will return the variable
mapping for the specified Python library. Optional parameter \code{x} allows
to specify a (potentially names) character vector with the names of the
spectra variables that should in addition be included in the mapping.
\item \code{setSpectraVariableMapping()}: sets/replaces the currently defined mapping
of spectra variable names to Python metadata names. Setting
\code{setSpectraVariableMapping(character())} will only convert the mass peaks
data (m/z and intensity values) but no spectra metadata.
}
}

\examples{

## Import a MGF file as a `Spectra` object
library(MsBackendMgf)
library(SpectriPy)
s <- Spectra(
    system.file("extdata", "mgf", "spectra2.mgf", package = "SpectriPy"),
    source = MsBackendMgf())
s

#########################
## Conversion R to Python

## A `Spectra` can be translated to a `list` of `matchms.Spectrum` objects
## using either the `r_to_py()` method or the `rspec_to_pyspec()` function:
s_py <- r_to_py(s)
s_py

## The `s_py` can now be used like any other Python variable within the R
## *reticulate* framework. Below we extract the m/z values of the first
## spectrum
s_py[0]$mz

## Extracting that information from the `Spectra` object in R
s[1]$mz

## The `spectraVariableMapping()` defines which spectra variables (metadata)
## should be translated between R and Python:
spectraVariableMapping()

## The names of that character vector represent the names of the spectra
## variables in R, the elements the name of the metadata variable in Python.
## Below we list the available metadata information from the first
## Spectrum in Python
s_py[0]$metadata

## `setSpectraVariableMapping()` allows to replace the default mapping
## of variables. Below we e.g. add a new spectra variable to the `Spectra`
## object.
s$new_col <- 1:4

## To translate that variable to Python we need to include it to the
## `spectraVariableMapping()`. Below we define to translate only the
## precursor m/z and the new spectra variable to Python. Be aware that
## `setSpectraVariableMapping()` **globally** sets the default for any
## spectra variable mapping between R and Python. Thus, any subsequent
## calls mapping calls will use the same mapping. It is suggested to
## eventually *restore* the default mapping again after the call or
## use the `rspec_to_pyspec()` function instead, that allows to configure
## the mapping using a parameter `mapping`.
setSpectraVariableMapping(
    c(precursorMz = "precursor_mz", new_col = "new_col"))
s_py <- r_to_py(s)

s_py[0]$metadata

## Restoring the global spectra variable mapping configuration to
## the default mapping:
setSpectraVariableMapping(defaultSpectraVariableMapping())

## As an alternative to the `r_to_py()` we can use the `rspec_to_pyspec()`
## function and provide a custom mapping using the `mapping` parameter:
s_py <- rspec_to_pyspec(
    s, mapping = c(precursorMz = "precursor_mz", new_col = "new_col"))

## Convert to MS data objects from the spectrum_utils Python library
s_py2 <- rspec_to_pyspec(
    s, mapping = spectraVariableMapping("spectrum_utils"),
    pythonLibrary = "spectrum_utils")

## Convert the data back to R
pyspec_to_rspec(s_py2, pythonLibrary = "spectrum_utils")

#########################
## Conversion Python to R

## A `list` of `matchms.Spectrum` objects in Python can be translated into
## the corresponding MS data structure in R (i.e. a `Spectra`) object using
## the `pyspec_to_rspec()` function:
res <- pyspec_to_rspec(s_py)
res

## All spectra from Python are thus converted into a single `Spectra` object.

## Or providing a custom variable mapping:
res <- pyspec_to_rspec(
    s_py, mapping = c(precursorMz = "precursor_mz", new_col = "new_col"))
res$new_col
}
\author{
Michael Witting, Johannes Rainer, Wout Bittremieux, Thomas Naake
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/MsBackendPython.R
\name{MsBackendPy}
\alias{MsBackendPy}
\alias{backendInitialize,MsBackendPy-method}
\alias{length,MsBackendPy-method}
\alias{spectraVariables,MsBackendPy-method}
\alias{spectraData,MsBackendPy-method}
\alias{spectraData<-,MsBackendPy-method}
\alias{peaksData,MsBackendPy-method}
\alias{peaksData<-,MsBackendPy-method}
\alias{$,MsBackendPy-method}
\alias{$<-,MsBackendPy-method}
\alias{intensity<-,MsBackendPy-method}
\alias{mz<-,MsBackendPy-method}
\alias{spectraVariableMapping<-,MsBackendPy-method}
\alias{spectraVariableMapping,MsBackendPy-method}
\alias{reindex}
\title{A MS data backend for MS data stored in Python}
\usage{
\S4method{backendInitialize}{MsBackendPy}(
  object,
  pythonVariableName = character(),
  spectraVariableMapping = defaultSpectraVariableMapping(),
  pythonLibrary = c("matchms", "spectrum_utils"),
  ...,
  data
)

\S4method{length}{MsBackendPy}(x)

\S4method{spectraVariables}{MsBackendPy}(object)

\S4method{spectraData}{MsBackendPy}(object, columns = spectraVariables(object), drop = FALSE)

\S4method{spectraData}{MsBackendPy}(object) <- value

\S4method{peaksData}{MsBackendPy}(object, columns = c("mz", "intensity"), drop = FALSE)

\S4method{peaksData}{MsBackendPy}(object) <- value

\S4method{$}{MsBackendPy}(x, name)

\S4method{$}{MsBackendPy}(x, name) <- value

\S4method{intensity}{MsBackendPy}(object) <- value

\S4method{mz}{MsBackendPy}(object) <- value

\S4method{spectraVariableMapping}{MsBackendPy}(object) <- value

\S4method{spectraVariableMapping}{MsBackendPy}(object, value)

reindex(object)
}
\arguments{
\item{object}{A \code{MsBackendPy} object.}

\item{pythonVariableName}{For \code{backendInitialize()}: \code{character(1)} with the
name of the variable/Python attribute that contains the list of
\code{matchms.Spectrum} objects with the MS data.}

\item{spectraVariableMapping}{For \code{backendInitialize()}: named \code{character}
with the mapping between spectra variable names and (\code{matchms.Spectrum})
metadata names. See \code{\link[=defaultSpectraVariableMapping]{defaultSpectraVariableMapping()}}, and the
description of the \code{backendInitialize()} function for \code{MsBackendPy}
for more information and details.}

\item{pythonLibrary}{For \code{backendInitialize()}: \code{character(1)} specifying
the Python library used to represent the MS data in Python. Can be
either \code{pythonLibrary = "matchms"} (the default) or
\code{pythonLibrary = "spectrum_utils"}.}

\item{...}{Additional parameters.}

\item{data}{For \code{backendInitialize()}: \code{DataFrame} with the full MS data
(peaks data and spectra data) such as extracted with the
\code{\link[Spectra:spectraData]{Spectra::spectraData()}} method on another \code{MsBackend} instance.
Importantly, the \code{DataFrame} must have columns \code{"mz"} and
\code{"intensity"} with the full MS data.}

\item{x}{A \code{MsBackendPy} object}

\item{columns}{For \code{spectraData()}: \code{character} with the names of
columns (spectra variables) to retrieve. Defaults to
\code{spectraVariables(object)}. For \code{peaksData()}: \code{character} with the
names of the peaks variables to retrieve.}

\item{drop}{For \code{spectraData()} and \code{peaksData()}: \code{logical(1)} whether,
when a single column is requested, the data should be returned as a
\code{vector} instead of a \code{data.frame} or \code{matrix}.}

\item{value}{Replacement value(s).}

\item{name}{For \code{$}: \code{character(1)} with the name of the variable to
retrieve.}
}
\value{
See description of individual functions for their return values.
}
\description{
The \code{MsBackendPy} allows to access MS data stored as \code{matchms.Spectrum}
or \code{spectrum_utils.spectrum.MsmsSpectrum} objects from the
\href{https://github.com/matchms/matchms}{\emph{matchms}} respectively
\href{https://github.com/bittremieux-lab/spectrum_utils}{\emph{spectrum_utils}} Python
library directly from R. The MS data (peaks data or spectra variables) are
translated on-the-fly when accessed. Thus, the \code{MsBackendPy} allows a
seamless integration of Python MS data structures into \code{\link[Spectra:Spectra]{Spectra::Spectra()}}
based analysis workflows.

The \code{MsBackendPy} object supports replacing values for peaks variables
(\emph{m/z} and intensity) and adding/replacing or removing spectra variables.
The changes are immediately translated and written back to the Python
variable.

See the descripion of the \code{backendInitialize()} method below for creation
and initialization of objects from this class. Also, the \code{setBackend()}
method for \code{\link[Spectra:Spectra]{Spectra::Spectra()}} objects internally uses
\code{backendInitialize()}, thus the same parameters can (and have) to be passed
if the backend of a \code{Spectra} object is changed to \code{MsBackendPy} using
the \code{setBackend()} method. Special care should also be given to parameter
\code{spectraVariableMapping}, that defines which spectra variables should be
considered/translated and how their names should or have to be converted
between R and Python. See the description for \code{backendInitialize()} and the
package vignette for details and examples.
}
\details{
The \code{MsBackendPy} keeps only a reference to the MS data in Python (i.e. the
name of the variable in Python) as well as an index pointing to the
individual spectra in Python but no other data. Any data requested from
the \code{MsBackendPy} is accessed and translated on-the-fly from the Python
variable. The \code{MsBackendPy} is thus an interface to the MS data, but not
a data container. All changes to the MS data in the Python variable
(performed e.g. in Python) immediately affect any \code{MsBackendPy} instances
pointing to this variable.

Special care must be taken if the MS data structure in Python is subset or
its order is changed (e.g. by another process). In that case it might be
needed to re-index the backend using the \code{reindex()} function:
\code{object <- reindex(object)}. This will update (replace) the index to the
individual spectra in Python which is stored within the backend.
}
\note{
As mentioned in the \emph{details} section the MS data is completely stored in
Python and the backend only references to this data through the name of
the variable in Python. Thus, each time MS data is requested from the
backend, it is retrieved in its \strong{current} state.
If for example data was transformed or metadata added or removed in the
Python object, it immediately affects the \code{Spectra}/backend.

Any replacement operation uses internally the \verb{spectraData()<-} method,
thus replacing/updating values for individual spectra variables or peaks
variables will first load the current data from Python to R, update or
replace the values and then store the full MS data again to the
referenced Python attribute.
}
\section{\code{MsBackendPy} methods}{


The \code{MsBackendPy} supports all methods defined by the \code{\link[Spectra:MsBackend]{Spectra::MsBackend()}}
interface for access to MS data. Details on the invidual functions can also
be found in the main documentation in the \emph{Spectra} package (i.e. for
\code{\link[Spectra:MsBackend]{Spectra::MsBackend()}}). Here we provide information for functions with
specific properties of the backend.
\itemize{
\item \code{backendInitialize()}: this method can be used to either initialize the
backend with data from a referenced and **existing ** MS data structure
in Python, or, through parameter \code{data}, first convert and store the
provided data to a Python MS data structure and then initialize the
backend pointing to this referenced variable (Python attribute). In both
cases, the name of the Python attribute needs to be provided with the
parameter \code{pythonVariableName}.
The mapping between the spectra variable names in R and the related
Python metadata variables can be specified with the
\code{spectraVariableMapping} parameter. It has to be a named \code{character} with
names being the spectra variables and the values the respective name for
the metadata in the Python MS data structure. It defaults to
\code{\link[=defaultSpectraVariableMapping]{defaultSpectraVariableMapping()}} which returns the mapping of some
core spectra variables for the \emph{matchms} Python library. Be aware that
only those spectra variables specified with this parameter are mapped and
translated between R and Python. For \code{backendInitialize()} with parameter
\code{data} provided, only the variables defined by \code{spectraVariableMapping},
and available in \code{data}, are converted and stored in Python. Also note
that, for efficiency reasons, core spectra variables (those listed by
\code{\link[Spectra:spectraData]{Spectra::coreSpectraVariables()}}) defined with \code{spectraVariableMapping}
but that have only missing values, are ignored.
Parameter \code{pythonLibrary} must be used to specify the Python library
representing the MS data in Python. It can be either
\code{pythonLibrary = "matchms"} (the default) or
\code{pythonLibrary = "spectrum_utils"}. The function returns an initialized
instance of \code{MsBackendPy}. See examples below for different settings
and conversion of spectra variables.
\item \code{intensity()}, \verb{intensity()<-}: get or replace the intensity values.
\code{intensity()} returns a \code{NumericList} of length equal to the number of
spectra with each element being the intensity values of the individual
mass peaks per spectrum. \verb{intensity()<-} takes the same list-like
structure as input parameter. Both the number of spectra and the number of
peaks must match the length of the spectra and the number of existing mass
peaks. To change the number of peaks use the \verb{peaksData()<-} method
instead that replaces the \emph{m/z} and intensity values at the same time.
Calling \verb{intensity()<-} will replace the full MS data (spectra variables
as well as peaks variables) of the associated Python variable.
\item \code{mz()}, \verb{mz()<-}: get or replace the \emph{m/z} values. \code{mz()} returns a
\code{NumericList} of length equal to the number of spectra with each element
being the \emph{m/z} values of the individual mass peaks per spectrum.
\verb{mz()<-} takes the same list-like structure as input parameter. Both the
number of spectra and the number of peaks must match the length of the
spectra and the number of existing mass peaks. To change the number of
peaks use the \verb{peaksData()<-} method instead that replaces the \emph{m/z} and
intensity values at the same time.
Calling \verb{mz()<-} will replace the full MS data (spectra variables
as well as peaks variables) of the associated Python variable.
\item \code{peaksData()}: extracts the peaks data matrices from the backend. Python
code is applied to the data structure in Python to
extract the \emph{m/z} and intensity values as a list of (numpy) arrays. These
are then translated into an R \code{list} of two-column \code{numeric} matrices.
Because Python does not allow to name columns of an array, an additional
loop in R is required to set the column names to \code{"mz"} and \code{"intensity"}.
\item \verb{peaksData()<-}: replaces the full peaks data (i.e., \emph{m/z} and intensity
values) for all spectra. Parameter \code{value} has to be a \code{list}-like
structure with each element being a \code{numeric} matrix with one column
(named \code{"mz"}) containing the spectrum's \emph{m/z} and one column (named
\code{"intensity"}) with the intensity values. This method will replace the
full data of the associated Python variable (i.e., both the spectra as
well as the peaks data).
\item \code{spectraData()}: extracts the spectra data from the backend. Which spectra
variables are translated and retrieved from the Python objects depends on
the backend's \code{spectraVariableMapping()}. All metadata names defined are
retrieved and added to the returned \code{DataFrame} (with eventually missing
\emph{core} spectra variables filled with \code{NA}).
\item \verb{spectraData()<-}: replaces the full spectra (+ peaks) data of the backend
with the values provided with the submitted \code{DataFrame}. The number of
rows of this \code{DataFrame} has to match the number of spectra of \code{object}
(i.e., being equal to \code{length(object)}) and the \code{DataFrame} must also
contain the spectras' \emph{m/z} and intensity values.
\item \code{spectraVariables()}: retrieves available spectra variables, which include
the names of all metadata attributes in the \code{matchms.Spectrum} objects
and the \emph{core} spectra variables \code{\link[Spectra:spectraData]{Spectra::coreSpectraVariables()}}.
\item \code{spectraVariableMapping()}: get the currently defined mapping for
\code{spectraVariables()} of the backend.
\item \verb{spectraVariableMapping<-}: replaces the \code{spectraVariableMapping} of the
backend (see \code{\link[=setSpectraVariableMapping]{setSpectraVariableMapping()}} for details and description
of the expected format).
\item \code{$}, \verb{$<-}: extract or add/replace values for a spectra variable from/in
the backend. Replacing or adding values for a spectra variable cause the
full data to be replaced. In detail, first the full data is retrieved from
Python, then the values are added/replaced and then the data is again
transferred to Python.
}
}

\section{Additional helper and utility functions}{

\itemize{
\item \code{reindex()}: update the internal \emph{index} to match \code{1:length(object)}.
This function is useful if the original data referenced by the backend was
subset or re-ordered by a different process (or a function in Python).
}
}

\examples{

## Loading an example MGF file provided by the SpectriPy package.
## As an alternative, the data could also be imported directly in Python
## using:
## import matchms
## from matchms.importing import load_from_mgf
## s_p = list(load_from_mgf(r.fl))
library(Spectra)
library(SpectriPy)
library(MsBackendMgf)

fl <- system.file("extdata", "mgf", "test.mgf", package = "SpectriPy")
s <- Spectra(fl, source = MsBackendMgf())
s

## Translating the MS data to Python and assigning it to a variable
## named "s_p" in the (*reticulate*'s) `py` Python environment. Assigning
## the variable to the Python environment has performance advantages, as
## any Python code applied to the MS data does not require any data
## conversions.
py_set_attr(py, "s_p", rspec_to_pyspec(s))

## Create a `MsBackendPy` representing an interface to the data in the
## "s_p" variable in Python:
be <- backendInitialize(MsBackendPy(), "s_p")
be

## Alternatively, by passing the full MS data with parameter `data`, the
## data is first converted to Python and the backend is initialized with
## that data. The `setBackend()` call from above internally uses this
## code to convert the data.
be <- backendInitialize(MsBackendPy(), "s_p3",
    data = spectraData(s, c(spectraVariables(s), "mz", "intensity")))

## Create a Spectra object which this backend:
s_2 <- Spectra(be)
s_2

## An easier way to change the data representation of a `Spectra` object
## from R to Python is to use the `Spectra`'s `setBackend()` method
## selecting a `MsBackendPy` as the target backend representation:
s_2 <- setBackend(s, MsBackendPy(), pythonVariableName = "s_p2")
s_2

## This moved the data from R to Python, storing it in a Python variable
## with the name `s_p2`. The resulting `s_2` is thus a `Spectra` object
## with all MS data however stored in Python.

## Note that by default only spectra variables that are part of
## `defaultSpectraVariableMapping()` are converted to Python
defaultSpectraVariableMapping()

## Thus, for example the precursor m/z is available in `s_2`, but other
## spectra variables from `s`, such as `"SMILES"` are not:
precursorMz(s)
precursorMz(s_2)

s$SMILES |> head()
## s_2$SMILES would throw an error.

## To also translate this spectra variable, it needs to be included and
## specified with the `spectraVariableMapping` parameter. The easiest
## approach is to use the `spectraVariableMapping()` function adding in
## addition to the default mapping for the Python library (`"matchms"`)
## also the mapping of additional spectra variables that should be converted:
s_2 <- setBackend(s, MsBackendPy(), pythonVariableName = "s_p2",
    spectraVariableMapping = spectraVariableMapping("matchms", c(SMILES = "smiles")))
s_2$SMILES |> head()

## Available spectra variables: these include, next to the *core* spectra
## variables, also the names of all metadata stored in the `matchms.Spectrum`
## objects.
spectraVariables(s_2)

## Get the full peaks data:
peaksData(s_2)

## Get the peaks from the first spectrum
peaksData(s_2)[[1L]]

## Get the full spectra data:
spectraData(s_2)

## Get the m/z values
mz(s_2)

## Plot the first spectrum
plotSpectra(s_2[1L])


########
## Using the spectrum_utils Python library

## Below we convert the data to a list of `MsmsSpectrum` object from the
## spectrum_utils library.
py_set_attr(py, "su_p", rspec_to_pyspec(s,
    spectraVariableMapping("spectrum_utils"), "spectrum_utils"))

## Create a MsBackendPy representing this data. Importantly, we need to
## specify the Python library using the `pythonLibrary` parameter and
## ideally also set the `spectraVariableMapping` to the one specific for
## that library.
be <- backendInitialize(MsBackendPy(), "su_p",
    spectraVariableMapping = spectraVariableMapping("spectrum_utils"),
    pythonLibrary = "spectrum_utils")
be

## Get the peaks data for the first 3 spectra
peaksData(be[1:3])

## Get the full spectraData
spectraData(be)

## Extract the precursor m/z
be$precursorMz
}
\author{
Johannes Rainer and the EuBIC hackathon team
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compareSpectriPy.R
\name{compareSpectriPy}
\alias{compareSpectriPy}
\alias{CosineGreedy}
\alias{CosineHungarian}
\alias{ModifiedCosine}
\alias{NeutralLossesCosine}
\alias{compareSpectriPy,Spectra,Spectra,CosineGreedy-method}
\alias{compareSpectriPy,Spectra,missing,CosineGreedy-method}
\title{Spectra similarity calculations using Python's matchms library}
\usage{
CosineGreedy(tolerance = 0.1, mz_power = 0, intensity_power = 1)

CosineHungarian(tolerance = 0.1, mz_power = 0, intensity_power = 1)

ModifiedCosine(tolerance = 0.1, mz_power = 0, intensity_power = 1)

NeutralLossesCosine(
  tolerance = 0.1,
  mz_power = 0,
  intensity_power = 1,
  ignore_peaks_above_precursor = TRUE
)

\S4method{compareSpectriPy}{Spectra,Spectra,CosineGreedy}(x, y, param, ...)

\S4method{compareSpectriPy}{Spectra,missing,CosineGreedy}(x, y, param, ...)
}
\arguments{
\item{tolerance}{\code{numeric(1)}: tolerated differences in the peaks' m/z. Peaks
with m/z differences \verb{<= tolerance} are considered matching.}

\item{mz_power}{\code{numeric(1)}: the power to raise m/z to in the cosine
function. The default is 0, in which case the peak intensity products will
not depend on the m/z ratios.}

\item{intensity_power}{\code{numeric(1)}: the power to raise intensity to in the
cosine function. The default is 1.}

\item{ignore_peaks_above_precursor}{For \code{NeutralLossesCosine()}:
\code{logical(1)}: if \code{TRUE} (the default), peaks with m/z values larger than
the precursor m/z are ignored.}

\item{x}{A \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object.}

\item{y}{A \code{\link[Spectra:Spectra]{Spectra::Spectra()}} object to compare against. If missing,
spectra similarities are calculated between all spectra in \code{x}.}

\item{param}{One of the parameter classes listed above (such as
\code{CosineGreedy}) defining the similarity scoring function in Python
and its parameters.}

\item{...}{ignored.}
}
\value{
\code{compareSpectriPy()} Returns a \code{numeric} matrix with the scores,
with the number of rows equal to \code{length(x)} and the number of columns
equal to \code{length(y)}.
}
\description{
The \code{compareSpectriPy()} function allows to calculate spectral similarity
scores using the \code{calculate_scores()} function of the Python
\href{https://matchms.readthedocs.io/en/latest/api/matchms.similarity.html}{matchms.similarity}.
module.

Selection and configuration of the algorithm can be performed with one of the
\emph{parameter} objects/functions:
\itemize{
\item \code{CosineGreedy()}: calculate the \emph{cosine similarity score} between
spectra. The score is calculated by finding the best possible matches
between peaks of two spectra. Two peaks are considered a potential match if
their m/z ratios lie within the given \code{tolerance}. The underlying peak
assignment problem is here solved in a \emph{greedy} way. This can perform
notably faster, but does occasionally deviate slightly from a fully correct
solution (as with the \code{CosineHungarian} algorithm). In practice this
will rarely affect similarity scores notably, in particular for smaller
tolerances. The algorithm can be configured with parameters \code{tolerance},
\code{mz_power} and \code{intensity_power} (see parameter description for more
details). See also
\href{https://matchms.readthedocs.io/en/latest/api/matchms.similarity.CosineGreedy.html}{matchms CosineGreedy} for more information.
\item \code{CosineHungarian()}: calculate the \emph{cosine similarity score} as with
\code{CosineGreedy}, but using the Hungarian algorithm to find the best
matching peaks between the compared spectra. The algorithm can be
configured with parameters \code{tolerance}, \code{mz_power} and \code{intensity_power}
(see parameter description for more details). See also
\href{https://matchms.readthedocs.io/en/latest/api/matchms.similarity.CosineHungarian.html}{matchms CosingHungarian} for more information.
\item \code{ModifiedCosine()}: The modified cosine score aims at quantifying the
similarity between two mass spectra. The score is calculated by finding
the best possible matches between peaks of two spectra. Two peaks are
considered a potential match if their m/z ratios lie within the given
\code{tolerance}, or if their m/z ratios lie within the tolerance once a
mass shift is applied. The mass shift is simply the difference in
precursor-m/z between the two spectra.
See also \href{https://matchms.readthedocs.io/en/latest/api/matchms.similarity.ModifiedCosine.html}{matchms ModifiedCosine} for more information.
\item \code{NeutralLossesCosine()}: The neutral losses cosine score aims at
quantifying the similarity between two mass spectra. The score is
calculated by finding the best possible matches between peaks of two
spectra. Two peaks are considered a potential match if their m/z ratios lie
within the given \code{tolerance} once a mass shift is applied. The mass shift
is the difference in precursor-m/z between the two spectra. See also
\href{https://matchms.readthedocs.io/en/latest/api/matchms.similarity.NeutralLossesCosine.html}{matchms NeutralLossesCosine} for more information.
\item \code{FingerprintSimilarity()}: Calculate similarity between molecules based
on their fingerprints. For this similarity measure to work, fingerprints
are expected to be derived by running \emph{add_fingerprint()}. See also
\href{https://matchms.readthedocs.io/en/latest/api/matchms.similarity.FingerprintSimilarity.html}{matchms FingerprintSimilarity} for more information.
}
}
\note{
Parameters and algorithms are named as originally defined in the \emph{matchms}
library (i.e. all parameters in \emph{snake_case} while \emph{CamelCase} is used
for the algorithms.
}
\examples{

library(Spectra)
## Create some example Spectra.
DF <- DataFrame(
    msLevel = c(2L, 2L, 2L),
    name = c("Caffeine", "Caffeine", "1-Methylhistidine"),
    precursorMz = c(195.0877, 195.0877, 170.0924)
)
DF$intensity <- list(
    c(340.0, 416, 2580, 412),
    c(388.0, 3270, 85, 54, 10111),
    c(3.407, 47.494, 3.094, 100.0, 13.240)
)
DF$mz <- list(
    c(135.0432, 138.0632, 163.0375, 195.0880),
    c(110.0710, 138.0655, 138.1057, 138.1742, 195.0864),
    c(109.2, 124.2, 124.5, 170.16, 170.52)
)
sps <- Spectra(DF)

## Calculate pairwise similarity beween all spectra within sps with
## matchms' CosineGreedy algorithm
## Note: the first compareSpectriPy will take longer because the Python
## environment needs to be set up.
res <- compareSpectriPy(sps, param = CosineGreedy())
res

## Next we calculate similarities for all spectra against the first one
res <- compareSpectriPy(sps, sps[1], param = CosineGreedy())

## Calculate pairwise similarity of all spectra in sps with matchms'
## ModifiedCosine algorithm
res <- compareSpectriPy(sps, param = ModifiedCosine())
res

## Note that the ModifiedCosine method requires the precursor m/z to be
## known for all input spectra. Thus, it is advisable to remove spectra
## without precursor m/z before using this algorithm.
sps <- sps[!is.na(precursorMz(sps))]
compareSpectriPy(sps, param = ModifiedCosine())
}
\seealso{
\code{\link[Spectra:compareSpectra]{Spectra::compareSpectra()}} in the \emph{Spectra} package for pure R
implementations of spectra similarity calculations.
}
\author{
Carolin Huber, Michael Witting, Johannes Rainer, Helge Hecht,
Marilyn De Graeve
}
